------------------------------------------------------------------------
--                  Generated by RSTK                                 --
--                      rs_chien.vhd                                  --
--                                                                    --
------------------------------------------------------------------------
--
--
LIBRARY IEEE ;
use IEEE.std_logic_1164.all ;
--
library IEEE ;
use     IEEE.std_logic_unsigned.all;
--
library IEEE ;
use     IEEE.NUMERIC_STD.all;
--
library gf_arith ;
use     gf_arith.gf_arith.all;
--
entity chien_srch is port( rst_n : IN std_logic ;
                           clk   : IN std_logic ;
                           start_search : IN std_logic ;
                           l0 : IN std_logic_vector(3 downto 0);
                           l1 : IN std_logic_vector(3 downto 0);
                           lroot0 : OUT std_logic_vector(3 downto 0);
                           n_roots : OUT std_logic_vector(3 downto 0);
                           search_done  : OUT std_logic
                       ) ;
end chien_srch ;

architecture rtl of chien_srch is 

--
-- Locals
--
    signal     root_0 : std_logic_vector(3 downto 0) ;

    signal     r0 : std_logic_vector(3 downto 0);
    signal     r1 : std_logic_vector(3 downto 0);


--  Search Control SM
    type chien_search_states is (
        csrch_s0,
        csrch_s1,
        csrch_s2,
        csrch_s3
    ) ;

    signal CSHRC_SM : chien_search_states ;

    constant GF_FIELD_CNT : integer := 15 ;

    signal load_coef  : std_logic ;
    signal enb_search : std_logic;
    signal done : std_logic;
--

    signal    root_cnt  : std_logic_vector(3 downto 0) ;
    signal    root_sel  : std_logic_vector(1 downto 0) ;
    signal    Num_Roots : std_logic_vector(3 downto 0) ;
    signal    root_calc : std_logic_vector(3 downto 0) ;

begin

    n_roots <= Num_Roots ;

    root_calc <= r0 xor 
                 r1 ;

    lroot0 <= root_0 ;

CHIEN_SEARCH_SM:process(rst_n, clk)
begin
    if (rst_n = '0') then
            enb_search <= '0' ;
            done <= '0' ;
            CSHRC_SM <= csrch_s0  ;
    elsif (clk'event and clk = '1') then
        case CSHRC_SM is
             when  csrch_s0 =>
                     if (start_search = '1') then
                         CSHRC_SM <= csrch_s1  ;
                     else 
                         CSHRC_SM <= csrch_s0  ;
                     end if ;

             when csrch_s1 =>
                     enb_search <= '1' ;
                     CSHRC_SM <= csrch_s2  ;

             when csrch_s2 =>
                     if (root_cnt < GF_FIELD_CNT) then
                             --root_cnt <= root_cnt + 1 ;
                             CSHRC_SM <= csrch_s2  ;
                     else
                             enb_search <= '0' ;
                             done <= '1' ;
                             CSHRC_SM <= csrch_s3  ;
                     end if ;

             when  csrch_s3 =>
                     CSHRC_SM <= csrch_s0  ;
                     done <= '0' ;

             when OTHERS =>
                     CSHRC_SM <= csrch_s0  ;
           end case ;
      end if ;
end process ;

    load_coef <= '1' when(CSHRC_SM = csrch_s0 and start_search ='1') else '0' ;
    search_done <= done ;

ROOT_FIND:process(rst_n, clk)
begin
    if (rst_n = '0') then
            r0 <= (OTHERS=> '0') ;
            r1 <= (OTHERS=> '0') ;
            root_0 <= (OTHERS=> '0') ;
            root_cnt <= (OTHERS=> '0') ;
            root_sel <= (OTHERS=> '0') ;
            Num_Roots <= (OTHERS=> '0') ;
    elsif (clk'event and clk = '1') then
        if (load_coef = '1' ) then
-- l0 Must be 1 for search to function correctly
            r0 <= l0 ;
            r1 <= l1 ;
            root_sel <= (OTHERS=> '0') ;
            Num_Roots <= (OTHERS=> '0') ;
        elsif (enb_search = '1') then
             if (root_calc = 0) then
                 case root_sel is
                     when "00" =>
                          root_0 <= root_cnt ;

                     when OTHERS =>

                  end case ;
                    root_sel <= root_sel + 1 ;
                    Num_Roots <= Num_Roots + 1 ;
               end if ;

            r1 <= gfmul_1(r1) ;
            root_cnt <= root_cnt + 1 ;
        end if ;
    end if;
end process ;

END ; 
