------------------------------------------------------------------------
--                  Generated by RSTK                                 --
--                      rs_encode.vhd                                 --
--                                                                    --
------------------------------------------------------------------------
--
--
LIBRARY IEEE ;
use IEEE.std_logic_1164.all ;
--
library IEEE ;
use 	IEEE.std_logic_unsigned.all;
--
library IEEE ;
use 	IEEE.NUMERIC_STD.all;
--
--
library gf_arith ;
use 	gf_arith.gf_arith.all;
--
--
entity rs_encode is
  generic ( N  : integer := 3 ;
            K  : integer := 3
		   ); 
    port( rst_n        : IN std_logic ;
          clk          : IN std_logic ;
          start        : IN std_logic ;
          d_in         : IN std_logic_vector(3 downto 0) ;
          enc_dout     : OUT std_logic_vector(3 downto 0);
          enc_dov      : OUT std_logic ;
          enc_synv     : OUT std_logic ;
          enc_done     : OUT std_logic
   ) ;
end rs_encode ;


architecture rtl of rs_encode is
--
--
--



    signal	din_r : std_logic_vector(3 downto 0);

    signal	g0 : std_logic_vector(3 downto 0);
    signal	g1 : std_logic_vector(3 downto 0);


    signal	g0_r : std_logic_vector(3 downto 0);
    signal	g1_r : std_logic_vector(3 downto 0);
    signal	g2   : std_logic_vector(3 downto 0)  ;


    signal	feedback : std_logic_vector(3 downto 0)  ;

    signal	xfer_cnt: std_logic_vector(3 downto 0) ;

    signal  init         : std_logic ;
    signal  done         : std_logic ;
    signal  enb_enc      : std_logic ;
    signal  enc_feedback : std_logic ;
    signal  syn_valid    : std_logic ;

    type    rs_encode_state is (
                rs_enc_s0,
                rs_enc_s1,
                rs_enc_s2,
                rs_enc_s3
            ) ;

    signal    RSENC_SM  : rs_encode_state ;
--
--
--
begin

     g0 <= gfmul_1(feedback) after 1 ns ;
     g1 <= gfmul_4(feedback) after 1 ns ;
     g2 <= g1_r  after 1 ns ;

     feedback <= (g2 xor din_r) when (enc_feedback = '1') else (OTHERS=>'0') after 1 ns ;

     enc_dout <= din_r  when (enc_feedback = '1') else g2 after 1 ns ;

dpipe:process (rst_n, clk)
begin
    if (rst_n = '0') then 
        din_r <= (OTHERS => '0') ;
    else if (clk'event and clk = '1') then
         din_r <= d_in after 1 ns ;
         end if ;
    end if ;
end process ;

rs_encoder:process (rst_n, clk)
begin
    if (rst_n = '0') then
	    g0_r <= (OTHERS => '0') ;
	    g1_r <= (OTHERS => '0') ;
    elsif (clk'event and clk = '1') then
        if (start = '1') then
	        g0_r <= (OTHERS => '0') ;
	        g1_r <= (OTHERS => '0') ;
        elsif (enb_enc = '1') then
	        g0_r <= g0 ;
	        g1_r <= g0_r xor g1 ;
		end if ;
    end if ;
end process ;

rs_encode_sm:process (rst_n, clk)
begin
    if (rst_n = '0') then
        RSENC_SM <= rs_enc_s0 ;
        xfer_cnt <= (OTHERS=>'0');
        enc_feedback <= '0' ;
        enb_enc <= '0' ;
        init <= '0' ;
    else
        if (clk'event and clk ='1') then
            case RSENC_SM is
                when rs_enc_s0 => 
                    if (start = '1') then
                        RSENC_SM <= rs_enc_s1 ;
                        init <= '1' ;
                        enb_enc <= '1' ;
                        enc_feedback <= '1' ;
                    else
                        RSENC_SM <= rs_enc_s0 ;
                    end if ;

                when rs_enc_s1 =>
                    if (xfer_cnt < K-1) then
                        RSENC_SM <= rs_enc_s1 ;
                        xfer_cnt <= xfer_cnt + 1 ;
                        init <= '0' ;
                    else
                        RSENC_SM <= rs_enc_s2 ;
                        xfer_cnt <= xfer_cnt + 1 ;
                        enc_feedback <= '0' ;
                        init <= '0' ;
                    end if ;

                when rs_enc_s2 => 
                    if (xfer_cnt < N) then
                        RSENC_SM <= rs_enc_s2 ;
                        xfer_cnt <= xfer_cnt + 1 ;
                        else
                        RSENC_SM <= rs_enc_s0 ;
                        enb_enc <= '0' ;
                        xfer_cnt <= (OTHERS => '0') ;
                    end if ;

                when OTHERS =>
                     RSENC_SM <= rs_enc_s0 ;

             end case ;
        end if ;
    end if ;

end process ;

    enc_dov <= enc_feedback after 1 ns ;

encoder_syn_valid:process (rst_n, clk) 
begin
    if (rst_n = '0') then 
        syn_valid <= '0' ;
    else if (clk'event and clk = '1') then
        if (RSENC_SM = rs_enc_s1 and xfer_cnt = K-1) then
            syn_valid <= '1' after 1 ns ;
        elsif (done = '1') then
            syn_valid <= '0' after 1 ns ;
        end if ;
	    end if ;
    end if ;
end process ;

    enc_synv <= syn_valid ; 

encoder_done:process (rst_n, clk)
begin
    if (rst_n = '0') then 
        done <= '0' ;
	else if (clk'event and clk = '1') then
        if (RSENC_SM = rs_enc_s2 and xfer_cnt = N-2) then
            done <= '1' ;
        else
            done <= '0' ;
        end if ;
        end if ;
    end if ;
end process ;

    enc_done <= done after 1 ns ;

END ;
