------------------------------------------------------------------------------------
--
--
-- File: sim_spram.vhd
--
--
------------------------------------------------------------------------------------
--
library ieee;
use ieee.std_logic_1164.all;
--
library ieee;
use ieee.std_logic_unsigned.all;
--
library ieee;
use std.textio.all;
--
library arithmetic ;
use arithmetic.std_logic_arith.all ;
--
library io_lib ;
use io_lib.io_utils.all;
--
entity spram_block is
  generic ( RAM_NUM_WORDS : integer ;
            RAM_WIDTH_A   : integer ;
            RAM_WIDTH_D   : integer ;
			  RAM_FILE      : string 
		   );
  port (clk  : in std_logic;
        we   : in std_logic;
        addr : in std_logic_vector(RAM_WIDTH_A-1 downto 0);
        di   : in std_logic_vector(RAM_WIDTH_D-1 downto 0);
        do   : out std_logic_vector(RAM_WIDTH_D-1 downto 0)
		);
end spram_block;
--
architecture syn of spram_block is

  attribute ram_style : string ;

  type ram_type is array (0 to RAM_NUM_WORDS-1)
        of std_logic_vector (RAM_WIDTH_D-1 downto 0);

  signal RAM : ram_type;
  attribute ram_style of RAM : signal is "block";

  signal read_adr : std_logic_vector(RAM_WIDTH_A-1 downto 0) ;

  signal ram_a    : integer := 0 ;
--  signal r_data   : std_logic_vector(RAM_WITDH-1 downto 0) ;

  signal mem_init : integer := 0 ;

begin

--process (clk)
--  variable outline  : line;
--begin
--    if (clk'event and clk = '1') then
--      if (we = '1') then
--        RAM(conv_integer(addr)) <= di;
--            write_string(outline, "what???");
--      end if;
--      read_adr <= addr ;
--    end if;
--end process;
--
----   do <= RAM(conv_integer(read_adr)) ;
--
-- Xilinx RAM simulation model
-- For use with VHDL simulator.
--
-- Note the use of variables in the procedure during initialization.
-- This is required since the <= operator is used for concurrent assignment
-- and the variable operator := is scheduled for immediate assignment.
-- If the variable is needed outside of the scope of the process
-- it is assigned using the <= operator.
-- 
process(clk)
  variable val      : integer;
  file     myfile   : text IS  RAM_FILE;
  variable inline   : line;
  variable outline  : line;
  variable mem_init_intl : integer := 0 ;
  variable ram_a_intl    : integer := 0 ;
  variable r_data_intl   : std_logic_vector(RAM_WIDTH_D-1 downto 0) ;
begin
   if (mem_init_intl = 0 and not endfile(myfile)) then
            readline(myfile, inline);

            write(outline, inline.ALL);

            read_based(inline, val);
            write_string(outline, " hex: ");
            write(outline, val, base => hex);
            writeline(output, outline);

			r_data_intl := conv_std_logic_vector(val,RAM_WIDTH_D) ;

			RAM(ram_a_intl) <= r_data_intl ;

            ram_a_intl := ram_a_intl + 1 ;

			ram_a <= ram_a_intl ;
            if (ram_a_intl = RAM_NUM_WORDS) then 
			    mem_init_intl := 1 ;
			end if ;
			mem_init <= mem_init_intl ;
--
--  Memory model is defined here after the initialization
--
    elsif (clk'event and clk = '1') then
      if (we = '1') then
        RAM(conv_integer(addr)) <= di;
       end if;
      read_adr <= addr after 1 ns ;
    end if;

end process;

   do <= RAM(conv_integer(read_adr)) after 1 ns ;

end syn;

