------------------------------------------------------------------------------------------
--																			 		    --
--			Reed-Solomon Encoder/Decoder Test Bench      								--
--																						--
------------------------------------------------------------------------------------------
--
--
LIBRARY IEEE ;
use IEEE.std_logic_1164.all ;
--
library IEEE ;
use 	IEEE.std_logic_unsigned.all;
--
library IEEE ;
use 	IEEE.NUMERIC_STD.all;
--

entity reed_solomon_tb is end ;

use work.chip_reg_access.all ;

architecture rtl of reed_solomon_tb is
component rs_encode is 
  generic ( N  : integer := 15 ;
            K  : integer := 13
		   ); 
    port( rst_n        : IN std_logic ;
	      clk          : IN std_logic ;
	      start        : IN std_logic ;
	      d_in         : IN std_logic_vector(3 downto 0) ;
	      enc_dout     : OUT std_logic_vector(3 downto 0) ;
		  enc_dov      : OUT std_logic ;
		  enc_synv     : OUT std_logic ;
		  enc_done     : OUT std_logic 
	    ) ;
end component ;

component dec_syn_calc is 
    port( rst_n   : IN std_logic ;
          init    : IN std_logic ;
          clk     : IN std_logic ;
          enb_dec : IN std_logic ;
          d_in : IN std_logic_vector(3 downto 0) ;
          s1   : OUT std_logic_vector(3 downto 0) ;
          s2   : OUT std_logic_vector(3 downto 0) ;
          syn_zero : OUT std_logic 
         ) ;
end component ;

component berl_mas  is 
    port( rst_n : IN std_logic;
          clk : IN std_logic;
          start_berl : IN std_logic;
          s0: IN std_logic_vector(3 downto 0);
          s1: IN std_logic_vector(3 downto 0);
          l0: OUT std_logic_vector(3 downto 0);
          l1: OUT std_logic_vector(3 downto 0);
          start_loc_search : OUT std_logic 
          ) ;
end component ;

component chien_srch is 
    port( rst_n : IN std_logic ;
          clk   : IN std_logic ;
          start_search : IN std_logic ;
          l0 : IN std_logic_vector(3 downto 0);
          l1 : IN std_logic_vector(3 downto 0);
          lroot0 : OUT std_logic_vector(3 downto 0);
          n_roots : OUT std_logic_vector(3 downto 0);
          search_done  : OUT std_logic
         ) ;
end component ;

component forney is 
    port ( rst_n : IN std_logic;
           clk : IN std_logic;
           start_forney : IN std_logic;
           root_search_done : IN std_logic;
           s0 : IN std_logic_vector(3 downto 0);
           s1 : IN std_logic_vector(3 downto 0);
           l0 : IN std_logic_vector(3 downto 0);
           l1 : IN std_logic_vector(3 downto 0);
           lroot0 : IN std_logic_vector(3 downto 0);
           n_roots : IN std_logic_vector(3 downto 0);
           corr_err_val0 : OUT std_logic_vector(3 downto 0);
           omega_dlam_rdy : OUT std_logic;
           forney_done : OUT std_logic
         ) ;
end component ;

	signal		start		   : std_logic := '0' ;
	signal		d_in		   : std_logic_vector(3 downto 0) ;
	signal		enc_dout	   : std_logic_vector(3 downto 0) ;
	signal      enc_dov        : std_logic ;
	signal      enc_synv       : std_logic ;
	signal      enc_done       : std_logic ;

	signal      enb_dec  : std_logic ;
	signal		s0	     : std_logic_vector(3 downto 0) ;
	signal		s1	     : std_logic_vector(3 downto 0) ;
	signal      syn_zero : std_logic ;

	signal		l0	     : std_logic_vector(3 downto 0) ;
	signal		l1	     : std_logic_vector(3 downto 0) ;
	signal      start_loc_search : std_logic ;

    signal      lroot0   : std_logic_vector(3 downto 0) ;
    signal      n_roots  : std_logic_vector(3 downto 0) ;
	signal      search_done : std_logic ;

	signal      corr_err_val0  : std_logic_vector(3 downto 0) ;
	signal      omega_dlam_rdy : std_logic ;
	signal      forney_done : std_logic ;
--
--
	signal	cpu_adr_bus	: std_logic_vector(31 downto 0):= "11111111111111111111111111111111" ;

	signal clk  : std_logic := '0' ;
	signal l_zero 	: std_logic := '0'  ;



	signal rst_n	: std_logic := '0' ;

	signal i : std_logic_vector(10 downto 0) := "00000000000" ;
	signal j : std_logic_vector(31 downto 0) := "00000000000000000000000000000000" ;
	signal p : std_logic_vector(7 downto 0) := "00000000" ;


	signal clk_count 		: integer ;
	signal clk_cnt_en       : std_logic ;
	signal enb_word_cnt		: std_logic  ;
	signal Word_Cnt 		: integer ;
	signal CodeWord_Din		: std_logic_vector(3 downto 0) ;
	signal Xfer_Cnt 		: std_logic_vector(7 downto 0) ;
	signal enb_xfer_cnt     : std_logic := '0' ;
    type   TestDataArray is array(0 to 15) of std_logic_vector(3 downto 0) ;
    signal TestData : TestDataArray ;

begin




	clk <= not(clk) after 10 ns ;

rs_encoder: rs_encode 
         generic map ( N  => 15,
                       K  => 13
		        ) 
         port map
               ( rst_n => rst_n,
	             clk => clk,
	             start => start,
	             d_in => d_in,
	             enc_dout => enc_dout,
				 enc_dov  => enc_dov,
				 enc_synv => enc_synv,
			     enc_done => enc_done
	            ) ;

rx_syn_dec :dec_syn_calc 
        port map 
		      ( rst_n => rst_n,
                init => start,
                clk => clk,
                enb_dec => enb_dec,
                d_in => CodeWord_Din,
                s1 => s0,
                s2 => s1,
                syn_zero => syn_zero
               ) ;

    CodeWord_Din <= enc_dout when (Word_Cnt < 12) else 
	                X"4" when (Word_Cnt = 12) else 
					enc_dout ;

rx_berl : berl_mas
        port map ( rst_n => rst_n,
                   clk => clk,
                   start_berl => enc_done,
                   s0 => s0, 
                   s1 => s1,
                   l0 => l0,
                   l1 => l1,
                   start_loc_search => start_loc_search
                 ) ;


rx_chien : chien_srch 
    port map( rst_n => rst_n,
              clk => clk,
              start_search => start_loc_search,
              l0 => l0,
              l1 => l1,
              lroot0 => lroot0,
              n_roots => n_roots,
              search_done => search_done
          ) ;

rx_forney : forney  
    port map ( rst_n => rst_n,
               clk => clk,
               start_forney => search_done,
               root_search_done => '0',
               s0 => s0,
               s1 => s1,
               l0 => l0,
               l1 => l1,
               lroot0 => lroot0,
               n_roots => n_roots,
               corr_err_val0 => corr_err_val0,
               omega_dlam_rdy => omega_dlam_rdy,
               forney_done => forney_done
            ) ;
--
-- Reset Control
--
init:process 
		begin
			if ( rst_n = '0' ) then
				i <= "00000000000" ;
				for i in 0 to 6 loop
					wait until clk ='1' ;
				end loop ;
				rst_n <= '1' ;
			end if ;
			if ( rst_n = '1' ) then

				clk_wait(10, clk) ;
                start <= '1' after 1 ns ;
				enb_xfer_cnt <= '1' ;
				clk_wait(1, clk) ;
                start <= '0' after 1 ns ;
				wait until enc_done = '1' ;
				clk_wait(1, clk) ;
				enb_xfer_cnt <= '0' ;
				clk_wait(1000, clk) ;

			end if ;
	end process ;

--	d_in <= X"00" when (Word_Cnt < 248) else X"01" after 1 ns ;
	d_in <= Xfer_Cnt(3 downto 0) when (Word_Cnt < 13) else
					 (OTHERS => '0') after 1 ns ;
--
--
simulation_cycle_cnt: process (rst_n, clk) 
begin
	if ( clk'event and clk='1' ) then
	   	if (rst_n = '0') then
			clk_count <= 0 ;
		else
			clk_count <= clk_count + 1 ;
		end if ;
end if ;
end process ;

	clk_cnt_en <= '1' when clk_count > 6 else '0' ;

--
-- 
--
process (rst_n, clk) 
begin
   	if ( rst_n = '0' ) then
		enb_word_cnt <= '0' ;
	else
		if ( clk'event and clk='1' ) then
			if (start = '1') then
				enb_word_cnt <= '1' after 1 ns ;
			end if ;
			if (enc_done = '1') then
				enb_word_cnt <= '0' after 1 ns ;
			end if ;

		end if ;
	end if ;
end process ;

process (rst_n, clk) 
begin
   	if ( rst_n = '0' ) then
		Word_Cnt <= 0 ;
	else
		if ( clk'event and clk='1' ) then
			if (enb_word_cnt = '1' and enc_done = '0') then
			    Word_Cnt <= Word_Cnt + 1 after 1 ns ;
			elsif (enb_word_cnt = '1' and enc_done = '1') then
		        Word_Cnt <= 0  after 1 ns;
			end if ;
		end if ;
	end if ;
end process ;

process (rst_n, clk) 
begin
   	if ( rst_n = '0' ) then
		Xfer_Cnt <= (OTHERS => '0') ;
	else
		if ( clk'event and clk='1' ) then
			if (enb_xfer_cnt = '1') then
			    Xfer_Cnt <= Xfer_Cnt + 1 after 1 ns ;
			elsif (enb_word_cnt = '0') then
		        Xfer_Cnt <= (OTHERS => '0')  after 1 ns;
			end if ;
		end if ;
	end if ;
end process ;
--
-- Start decode process
--
process (rst_n, clk) 
begin
   	if ( rst_n = '0' ) then
		enb_dec <= '0' ;
	else
		if ( clk'event and clk='1' ) then
			if (start = '1') then
			    enb_dec <= '1' after 1 ns ;
			elsif (enc_done = '1') then
			    enb_dec <= '0' after 1 ns ;
			end if ;
		end if ;
	end if ;
end process ;

--end of testbench 
end ;
