/*******************************************************************************/
/*                                                                             */
/*            gen_berl.c                                                       */
/*                                                                             */
/*    Generate RS Receiver Berlenkamp-Massey                                   */
/*                                                                             */
/*                                                                             */
/*                                                                             */
/*                                                                             */ 
/*                                                                             */
/*                                                                             */
/*                                                                             */
/*    Copyright (C) 2003  John S. Willy (jswilly@pacbell.net)                  */ 
/*                                                                             */
/*    This program is free software; you can redistribute it and/or modify     */
/*    it under the terms of the GNU General Public License as published by     */
/*    the Free Software Foundation; either version 2 of the License, or        */
/*    (at your option) any later version.                                      */
/*                                                                             */
/*    This program is distributed in the hope that it will be useful,          */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*    GNU General Public License for more details.                             */
/*                                                                             */
/*    You should have received a copy of the GNU General Public License        */
/*    along with this program; if not, write to the Free Software              */
/*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA*/
/*******************************************************************************/

#include <stdio.h>
#include "rsccg_proto.h"

#define VHDL     0 
#define VLOG     1 
#define C        2 

extern unsigned char poly_table[] ;   // RS code table 
extern unsigned char log_poly_table[] ; 
extern unsigned int p_wsp ;           // RS Generator Polynomial

/* locals  */

void gen_rs_berl(FILE *stream, int base, int gf_m, int deg, int chk_sym, int hdl_type, int gf_field)
{

    int i, vlen, t;
    unsigned char b_vector[16] ; // Binary bit vector
    unsigned char zero_vector[16] ; // Zeros vector for 0 initialization in component
    vlen = vector_len(2*chk_sym) ;
    t = 2*chk_sym ;
  //printf("VLEN = %d\n", vlen) ;
  //conv2_bvec_str(2*chk_sym, vlen, b_vector) ;
    for (i=0; i<16; i++)
     zero_vector[i] = '0' ;
   switch (hdl_type) {
       case C:
           fprintf(stream, "//--------------------------------------------------------------------//\n") ;
           fprintf(stream, "//                  Generated by RSTK                                 //\n") ;  
           fprintf(stream, "//                      rs_syndec.c                                   //\n") ;  
           fprintf(stream, "//                                                                    //\n") ;  
           fprintf(stream, "//--------------------------------------------------------------------//\n") ;
           fprintf(stream, "#include <stdio.h>\n") ;
           fprintf(stream, "#include \"proto_types.h\"\n") ;
           fprintf(stream, "/**********************************************************/\n") ;
           fprintf(stream, "/* Berlenkamp-Massey Error Locator Algorithm              */\n") ;
           fprintf(stream, "/**********************************************************/\n") ;
           fprintf(stream, "void berl_mas( int chk_sym, char *s, char *Lam, int dbg )\n") ;
           fprintf(stream, "{\n") ;
           fprintf(stream, "int i,j,t,r,v ;\n") ;
           fprintf(stream, "int L ;\n") ;
           fprintf(stream, "unsigned char S_x[%d] ;\n", chk_sym+2) ;
           fprintf(stream, "unsigned char lambda[%d] ;\n", chk_sym+2) ;
           fprintf(stream, "unsigned char T_x[%d] ;\n", chk_sym+2) ;
           fprintf(stream, "unsigned char B_x[%d] ;\n", chk_sym+2) ;
           fprintf(stream, "unsigned char delta_R ;\n") ;

           fprintf(stream, "unsigned char poly_x1[2] = {0,1}  ;\n") ;

           fprintf(stream, "unsigned char temp[%d] ;\n", chk_sym+2) ;
           fprintf(stream, "int deg ;\n") ;
           fprintf(stream, "unsigned char inv_delta_R ;\n") ;

           fprintf(stream, "  /* Clear work space */\n") ;
           fprintf(stream, "  for (i=0; i<%d; i++) {\n", chk_sym+2) ;
           fprintf(stream, "    S_x[i] = 0 ;\n") ;
           fprintf(stream, "    lambda[i] = 0 ;\n") ;
           fprintf(stream, "    T_x[i] = 0 ;\n") ;
           fprintf(stream, "    B_x[i] = 0 ;\n") ;
           fprintf(stream, "    temp[i] = 0 ;\n") ;
           fprintf(stream, "  }\n") ;

           fprintf(stream, "  /* Set to initial conditions */\n") ;
           fprintf(stream, "  t=0;\n") ;
           fprintf(stream, "  r = 0 ;\n") ;
           fprintf(stream, "  delta_R = 0 ;\n") ;
           fprintf(stream, "  L = 0 ;\n") ;
           fprintf(stream, "  lambda[0] = 1 ;\n") ;
           fprintf(stream, "  B_x[0] = 1 ;\n") ;
           fprintf(stream, "  v = 2*chk_sym ; \n") ;  
           fprintf(stream, "  /* Copy syndromes to new workspace syndrome list */\n") ;
           fprintf(stream, "  //for (i=0; i<v+1; i++) {\n") ;
           fprintf(stream, "    //S_x[i+1] = *s++ ;\n") ;
           fprintf(stream, "  //}\n") ;

           fprintf(stream, "  for (i=0; i<v; i++) {\n") ;
           fprintf(stream, "    S_x[i] = *s++ ;\n") ;
           fprintf(stream, "  }\n") ;

           fprintf(stream, "  for ( t=0; t<v; t++) {\n") ;
           fprintf(stream, "    if (dbg == 1) printf(\" ------------Iteration----------- %%d  \\n\", t) ;\n") ;
           fprintf(stream, "    //r = r + 1 ;\n") ;

           fprintf(stream, "    delta_R = 0 ;\n") ;
           fprintf(stream, "      if (dbg == 1) {\n") ;
           fprintf(stream, "        printf(\"L(x) = \") ;\n") ;
           fprintf(stream, "        for (i=0; i<= v; i++)  \n") ;
           fprintf(stream, "           printf(\"%%2x \", lambda[i]) ;\n") ;
           fprintf(stream, "        printf(\"\\n\") ;\n") ;
           fprintf(stream, "      }\n") ;
           fprintf(stream, "      if (dbg == 1) { \n") ;
           fprintf(stream, "        printf(\"x*B(x) = \") ;\n") ;
           fprintf(stream, "        for (i=0; i<= v; i++)\n") ;  
           fprintf(stream, "           printf(\"%%2x \", B_x[i]) ;\n") ;
           fprintf(stream, "        printf(\"\\n\") ;\n") ; 
           fprintf(stream, "      }\n") ;

           fprintf(stream, "    for (j=0; j<=L; j++) {\n") ;
           fprintf(stream, "       //delta_R = gfmul_ab(lambda[j],S_x[r-j]) ^ delta_R ; \n") ;   
           fprintf(stream, "       delta_R = gfmul_ab(lambda[j],S_x[r-j]) ^ delta_R ; \n") ;   
           fprintf(stream, "       if (dbg == 1) printf(\"delta_R = %%2x\\n\",delta_R) ;\n") ;
           fprintf(stream, "    }\n") ;

           fprintf(stream, "    r = r + 1 ;\n") ;
           fprintf(stream, "    if ( delta_R != 0 ) {\n") ; 
           fprintf(stream, "      /*    B(x) by delta R*x   */\n") ;
           fprintf(stream, "      poly_x1[1] = delta_R ;\n") ;
           fprintf(stream, "      gf_poly_mul(poly_x1, 1,  B_x, v, temp, &deg) ;\n") ;
     
           fprintf(stream, "      if (dbg == 1) {\n") ;
           fprintf(stream, "        printf(\"x*B(x) = \") ;\n") ;
           fprintf(stream, "        for (i=0; i<= v; i++)\n") ;  
           fprintf(stream, "           printf(\"%%2x \", temp[i]) ;\n") ;
           fprintf(stream, "        printf(\"\\n\") ;\n") ;
           fprintf(stream, "      }\n") ;

           fprintf(stream, "      /* Add lambda(x) to delta_R*x*B(x) */\n") ;
           fprintf(stream, "      gf_poly_add(lambda, v, temp, v, T_x, &deg) ;\n") ; 

           fprintf(stream, "      if (dbg == 1) {\n") ;
           fprintf(stream, "        printf(\"T(x) = Lambda + x*B(x) = \") ;\n") ;
           fprintf(stream, "        for (i=0; i<= v; i++) \n") ; 
           fprintf(stream, "           printf(\"%%2x \", T_x[i]) ;\n") ;
           fprintf(stream, "        printf(\"\\n\") ;\n") ;
           fprintf(stream, "      }\n") ;

           fprintf(stream, "      if (2*L <= r - 1) {\n") ;
           fprintf(stream, "        if (dbg == 1) printf(\" ++ 2L<= r-1 ++ \\n\") ;\n") ;
           fprintf(stream, "        inv_delta_R = gf_inverse(delta_R) ;\n") ;  
           fprintf(stream, "        if (dbg == 1) printf(\"Inverse Delta R = %%2x \\n\", inv_delta_R) ;\n") ;
           fprintf(stream, "        gf_poly_conmul( inv_delta_R, lambda, v,  temp, v, &deg) ;\n") ;
     

           fprintf(stream, "        if (dbg == 1) printf(\"B(x) =\") ;\n") ;
           fprintf(stream, "        for (i=0; i<= v; i++) {\n") ;   
           fprintf(stream, "          B_x[i] = temp[i] ;\n") ;
           fprintf(stream, "            if (dbg == 1) printf(\"%%2x \", B_x[i]) ;\n") ;
           fprintf(stream, "        }\n") ;

           fprintf(stream, "        for (i=0; i<= v; i++)\n") ;   
           fprintf(stream, "          lambda[i] = T_x[i] ;\n") ;

           fprintf(stream, "        if (dbg == 1) {\n") ;
           fprintf(stream, "          printf(\"\\nL(x) = \") ;\n") ;
           fprintf(stream, "          for (i=0; i<= v; i++) \n") ; 
           fprintf(stream, "             printf(\"%%2x \", lambda[i]) ;\n") ;
           fprintf(stream, "          printf(\"\\n\") ;\n") ;
           fprintf(stream, "        }\n") ;
           fprintf(stream, "        if (dbg == 1) printf(\"\\n\") ;\n") ;

           fprintf(stream, "          L = r - L ;\n") ;
           fprintf(stream, "      }\n") ; 
           fprintf(stream, "      else {\n") ;
           fprintf(stream, "        if (dbg == 1) printf(\" ** 2L > r-1 ** \\n\") ;\n") ;
           fprintf(stream, "        for (i=0; i<= v; i++)\n") ;   
           fprintf(stream, "          lambda[i] = T_x[i] ;\n") ;

           fprintf(stream, "        poly_x1[1] = 1 ;\n") ;
           fprintf(stream, "        gf_poly_mul(poly_x1, 1, B_x, v, temp, &deg) ;\n") ;
        
           fprintf(stream, "        for (i=0; i<= v; i++)\n") ;   
           fprintf(stream, "          B_x[i] = temp[i] ;\n") ;

           fprintf(stream, "        if (dbg == 1) { \n") ;
           fprintf(stream, "          printf(\"L(x) =\") ;\n") ;
           fprintf(stream, "          for (i=0; i<= v; i++) \n") ;   
           fprintf(stream, "             printf(\"%%2x \", lambda[i]) ;\n") ;
           fprintf(stream, "          printf(\"B(x) =\") ;\n") ;
           fprintf(stream, "          for (i=0; i<= v; i++)\n") ;    
           fprintf(stream, "             printf(\"%%2x \", B_x[i]) ;\n") ;
           fprintf(stream, "          printf(\"\\n\") ;\n") ;
           fprintf(stream, "        }\n") ;
           fprintf(stream, "      } \n") ;
           fprintf(stream, "    }\n") ;
           fprintf(stream, "    else {\n") ;
           fprintf(stream, "        poly_x1[1] = 1 ;\n") ;
           fprintf(stream, "        gf_poly_mul(poly_x1, 1, B_x, v, temp, &deg) ;\n") ;
        
           fprintf(stream, "        for (i=0; i<= v; i++)\n") ;   
           fprintf(stream, "          B_x[i] = temp[i] ;\n") ;
           fprintf(stream, "    }\n") ;

           fprintf(stream, "  } \n") ;
           fprintf(stream, "  for (i=0; i<= v; i++) \n") ;  
           fprintf(stream, "    *Lam++ = lambda[i] ;\n") ;
           fprintf(stream, "  printf(\"L = %%d\\n\", L ) ;\n") ;
  
           fprintf(stream, "}\n") ;
           break;

       case VLOG : 
           fprintf(stream, "//--------------------------------------------------------------------//\n") ;
           fprintf(stream, "//                  Generated by RSTK                                 //\n") ;  
           fprintf(stream, "//                      rs_berl.v                                     //\n") ;  
           fprintf(stream, "//                                                                    //\n") ;  
           fprintf(stream, "//--------------------------------------------------------------------//\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "module berl_mas(  rst_n,\n") ;
           fprintf(stream, "       clk,\n") ;
           fprintf(stream, "       start_berl,\n") ;
             for(i=0; i < 2*chk_sym; i++) 
               fprintf(stream, "       s%d,\n", i) ;
             for(i=0; i < 2*chk_sym; i++) 
               fprintf(stream, "       l%d, \n", i) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "       start_loc_search \n") ;
           fprintf(stream, "     ) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "  input    rst_n ;\n") ;
           fprintf(stream, "  input    clk ;\n") ;
           fprintf(stream, "  input    start_berl ;\n") ;
           fprintf(stream, "\n") ;
             for(i=0; i < 2*chk_sym; i++)
               fprintf(stream, "  input[%d:0]  s%d ;\n", gf_m-1, i) ;
           fprintf(stream, "\n") ;
             for(i=0; i < 2*chk_sym; i++)
               fprintf(stream, "  output[%d:0]  l%d ;\n", gf_m-1, i) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "  output          start_loc_search ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "`include \"../gf_arith.v\"\n") ;
           fprintf(stream, "\n") ;		   
           fprintf(stream, "// Locals\n") ;
           fprintf(stream, "    parameter   V = %d ;\n", 2*chk_sym) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "// Delta_R\n") ;
           fprintf(stream, "    reg[%d:0]  delta_r ;\n", gf_m-1) ;
           fprintf(stream, "    wire[%d:0]  inv_delta_r ;\n", gf_m-1) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "// L Count\n") ;
           fprintf(stream, "    reg[%d:0]  L ;\n", vlen) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "// r Count\n") ;
           fprintf(stream, "    reg[%d:0]  r ;\n", vlen) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "// j Count - used for Delta R calculation\n") ;
           fprintf(stream, "    reg[%d:0]  j ;\n", vlen) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "// Polynomial Length Count \n") ; 
           fprintf(stream, "    reg[%d:0]  Poly_Len_Cnt;\n", vlen) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "// Syndrome Delta R calculator\n") ;
           fprintf(stream, "   wire[%d:0]    syn_sel ;\n", vlen) ;
           fprintf(stream, "   wire[%d:0]    lam_sel ;\n", vlen) ;
           fprintf(stream, "   wire[%d:0]    delta_syn ;\n", gf_m-1) ;
           fprintf(stream, "   wire[%d:0]    delta_lam ;\n", gf_m-1) ;
           fprintf(stream, "   wire[%d:0]    delta_r_calc ;\n", gf_m-1) ;
           fprintf(stream, "   wire[%d:0]    delta_r_mul ;\n", gf_m-1) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "// Unit Control State Machine\n") ;
           fprintf(stream, "   reg[3:0] Berl_SM ;\n") ;
           fprintf(stream, "   reg      loc_srch ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "// Berl_SM States\n") ;
           fprintf(stream, "   parameter Berl_s0 = 0  ;\n") ;
           fprintf(stream, "   parameter Berl_s1 = 1  ;\n") ;
           fprintf(stream, "   parameter Berl_s2 = 2  ;\n") ;
           fprintf(stream, "   parameter Berl_s3 = 3  ;\n") ;
           fprintf(stream, "   parameter Berl_s4 = 4  ;\n") ;
           fprintf(stream, "   parameter Berl_s5 = 5  ;\n") ;
           fprintf(stream, "   parameter Berl_s6 = 6  ;\n") ;
           fprintf(stream, "   parameter Berl_s7 = 7  ;\n") ;
           fprintf(stream, "   parameter Berl_s8 = 8  ;\n") ;
           fprintf(stream, "   parameter Berl_s9 = 9  ;\n") ;
           fprintf(stream, "   parameter Berl_s10 = 10  ;\n") ;
           fprintf(stream, "   parameter Berl_s11 = 11  ;\n") ;
           fprintf(stream, "   parameter Berl_s12 = 12  ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "// B(x) register control modes\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "   wire[%d:0] B_x_Q ;\n", (2*chk_sym*gf_m-1)+gf_m) ;
           fprintf(stream, "   wire[2:0]  B_x_Op ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "// L(x) register control modes\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "   wire[%d:0] L_x_Q ;\n", (2*chk_sym*gf_m-1)+gf_m) ;
           fprintf(stream, "   wire[2:0]  L_x_Op ;\n") ;
             for(i=0; i <= 2*chk_sym; i++)
               fprintf(stream, "   assign l%d = L_x_Q[%d:%d] ;\n",  i, ((i+1)*gf_m)-1, i*gf_m) ;
           fprintf(stream, "\n") ; 
           fprintf(stream, "//\n") ;
           fprintf(stream, "// T(x) register control modes\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "   wire[%d:0] T_x_Q ;\n",(2*chk_sym*gf_m-1)+gf_m) ;
           fprintf(stream, "   wire[2:0]  T_x_Op ;\n") ;
           fprintf(stream, "\n") ; 
           fprintf(stream, "//\n") ;
           fprintf(stream, "// R(x) register control modes\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "   wire[%d:0] R_0_Q ;\n", (2*chk_sym*gf_m-1)+gf_m) ;
           fprintf(stream, "   wire[%d:0] R_1_Q ;\n", (2*chk_sym*gf_m-1)+gf_m) ;
           fprintf(stream, "   wire[%d:0] R_2_Q ;\n", (2*chk_sym*gf_m-1)+gf_m) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "   wire[%d:0]  R_2_din ;\n", gf_m-1) ;
           fprintf(stream, "   wire[%d:0] R_0_din ;\n",(2*chk_sym*gf_m-1)+gf_m) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "   wire[2:0]  R_0_Op ;\n") ;
           fprintf(stream, "   wire[2:0]  R_1_Op ;\n") ;
           fprintf(stream, "   wire[2:0]  R_2_Op ;\n") ;
           fprintf(stream, "   wire[%d:0]  R_1_add ;\n", gf_m-1) ;
           fprintf(stream, "   wire[%d:0]  R_0_mul ;\n", gf_m-1) ;
           fprintf(stream, "   wire[%d:0]  R_1_mul ;\n", gf_m-1) ;
           fprintf(stream, "\n") ;   
           fprintf(stream, "//\n") ;
           fprintf(stream, "// Register Control Op Codes\n") ;
           fprintf(stream, "   parameter x_nop       = 0 ;\n") ;
           fprintf(stream, "   parameter x_load      = 1 ;\n") ;
           fprintf(stream, "   parameter x_shft_l_z  = 2 ;\n") ;
           fprintf(stream, "   parameter x_shft_l    = 3 ;\n") ;
           fprintf(stream, "   parameter x_shft_r_z  = 4 ;\n") ;
           fprintf(stream, "   parameter x_shft_r    = 5 ;\n") ;
           fprintf(stream, "   parameter x_init_zero = 6 ;\n") ;
           fprintf(stream, "   parameter x_init_one  = 7 ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "//\n") ;
             for(i=0; i < chk_sym+2; i++)
               fprintf(stream, "   wire [%d:0] l_%d ;\n", gf_m-1, i) ;
           fprintf(stream, "\n") ;
             for(i=0; i < chk_sym+2; i++)
               fprintf(stream, "   assign l_%d = L_x_Q[%d:%d] ;\n", i, ((i+1)*gf_m)-1, i*gf_m) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "\n") ;
             for(i=0; i < chk_sym+2; i++)
               fprintf(stream, "   parameter syn_s%d    = %d ;\n", i, i) ;
           fprintf(stream, "\n") ;
             for(i=0; i < chk_sym+2; i++)
               fprintf(stream, "   parameter lam_l%d    = %d ;\n", i ,i) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "   assign start_loc_search = loc_srch ;\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "// GF Inverse ROM Table\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "// Generates the inverse of a^n (1/a^n)\n") ;
           fprintf(stream, "// \n") ;
           fprintf(stream, "//////////////////////////////////////////////////////////////////////////\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "//     ROM Tables for Berlekamp-Massey Algorithm\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "///////////////////////////////////////////////////////////////////////////\n") ;
           fprintf(stream, "// --- INSERT ROM HERE ---") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "//////////////////////////////////////////////////////////////////////////////////////////\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "//  Control State Machine\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "always @(negedge rst_n or posedge clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n == 1'b0)\n") ;
           fprintf(stream, "        begin\n") ;
           fprintf(stream, "            delta_r <= 8'h00 ;\n") ;
           fprintf(stream, "            r <= %d'b0 ;\n", vlen+1) ;
           fprintf(stream, "            j <= %d'b0 ;\n", vlen+1) ;
           fprintf(stream, "            L <= %d'b0 ;\n", vlen+1) ;
           fprintf(stream, "            delta_r <= %d'b0 ;\n", gf_m) ;
           fprintf(stream, "            Poly_Len_Cnt <= %d'b0 ;\n", vlen+1) ;
           fprintf(stream, "            Berl_SM <= Berl_s0 ;\n") ;
           fprintf(stream, "        end\n") ;
           fprintf(stream, "    else\n") ;
           fprintf(stream, "        case (Berl_SM)\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "            Berl_s0: begin\n") ;
           fprintf(stream, "                if (start_berl == 1'b1)\n") ;
           fprintf(stream, "                    begin\n") ;
           fprintf(stream, "                        r <= %d'b0 ;\n", vlen+1) ;
           fprintf(stream, "                        L <= %d'b0 ;\n", vlen+1) ;
           fprintf(stream, "                        Berl_SM <= #1 Berl_s1 ;\n") ;
           fprintf(stream, "                    end\n") ;
           fprintf(stream, "                end \n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "            Berl_s1: begin\n") ;
           fprintf(stream, "                         delta_r <= %d'b0 ;\n", gf_m) ;
           fprintf(stream, "                         j <= %d'b0 ;\n", vlen+1) ;
           fprintf(stream, "                         Berl_SM <= #1 Berl_s2 ;\n") ;
           fprintf(stream, "                end \n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "            Berl_s2: begin                 // Calculate delta r\n") ;
           fprintf(stream, "                         if (j <= L )\n") ;
           fprintf(stream, "                             begin\n") ;
           fprintf(stream, "                                 delta_r <= delta_r_calc ;\n") ;
           fprintf(stream, "                                 j <= j + 1 ;\n") ;
           fprintf(stream, "                                 Berl_SM <= Berl_s2 ;\n") ;
           fprintf(stream, "                             end\n") ;
           fprintf(stream, "                         else\n") ;
           fprintf(stream, "                             begin\n") ;
           fprintf(stream, "                                 r <= r + 1 ;\n") ;
           fprintf(stream, "                                 Berl_SM <= Berl_s3 ;\n") ;
           fprintf(stream, "                             end \n") ;
           fprintf(stream, "                end \n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "            Berl_s3: begin\n") ;
           fprintf(stream, "                         if (delta_r == %d'b0)\n", gf_m) ;
           fprintf(stream, "                             begin\n") ;
           fprintf(stream, "                                 Berl_SM <= Berl_s11 ;\n") ;
           fprintf(stream, "                             end\n") ;
           fprintf(stream, "                         else\n") ;
           fprintf(stream, "                             begin\n") ;
           fprintf(stream, "                                 Berl_SM <= Berl_s4 ;\n") ;
           fprintf(stream, "                                 Poly_Len_Cnt <= %d'b0 ;\n", vlen+1) ;
           fprintf(stream, "                             end \n") ;
           fprintf(stream, "                end\n") ; 
           fprintf(stream, "\n") ;
           fprintf(stream, "            Berl_s4: begin\n") ;
           fprintf(stream, "                         Berl_SM <= Berl_s5 ;\n") ;
           fprintf(stream, "                     end\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "            Berl_s5: begin                    // Calculate delta_r*x*B(x)\n") ;
           fprintf(stream, "                         if (Poly_Len_Cnt < V)\n") ;
           fprintf(stream, "                             begin\n") ;
           fprintf(stream, "                                 Berl_SM <= Berl_s5 ;\n") ;
           fprintf(stream, "                                 Poly_Len_Cnt <= Poly_Len_Cnt + 1 ;\n") ;
           fprintf(stream, "                             end\n") ;
           fprintf(stream, "                         else\n") ;
           fprintf(stream, "                             begin\n") ;
           fprintf(stream, "                                 Berl_SM <= Berl_s6 ;\n") ;
           fprintf(stream, "                                 Poly_Len_Cnt <= %d'b0 ;\n", vlen+1) ;
           fprintf(stream, "                             end\n") ; 
           fprintf(stream, "                end\n") ; 
           fprintf(stream, "\n") ;
           fprintf(stream, "            Berl_s6: begin                   // Calculate T(x) = L(x) + delta_r*x*B(x)\n") ;
           fprintf(stream, "                         if (Poly_Len_Cnt < V)\n") ;
           fprintf(stream, "                             begin\n") ;
           fprintf(stream, "                                 Berl_SM <= Berl_s6 ;\n") ;
           fprintf(stream, "                                 Poly_Len_Cnt <= Poly_Len_Cnt + 1 ;\n") ;
           fprintf(stream, "                             end\n") ;
           fprintf(stream, "                         else\n") ;
           fprintf(stream, "                             Berl_SM <= Berl_s7 ;\n") ;
           fprintf(stream, "                end\n") ; 
           fprintf(stream, "\n") ;
           fprintf(stream, "            Berl_s7: begin\n") ;
           fprintf(stream, "                         if (2*L <= r-1)\n") ; 
           fprintf(stream, "                             begin\n") ;
           fprintf(stream, "                                 Berl_SM <= Berl_s8 ;\n") ;
           fprintf(stream, "                                 Poly_Len_Cnt <= %d'b0 ;\n", vlen+1) ;
           fprintf(stream, "                             end\n") ;
           fprintf(stream, "                         else\n") ;
           fprintf(stream, "                             begin\n") ;
           fprintf(stream, "                                 Berl_SM <= Berl_s10 ;\n") ;
           fprintf(stream, "                             end\n") ;
           fprintf(stream, "                end\n") ; 
           fprintf(stream, "\n") ;
           fprintf(stream, "            Berl_s8: begin                   // Calculate (delta_r^-1)*L(x)\n") ;
           fprintf(stream, "                         if (Poly_Len_Cnt < V)\n") ;
           fprintf(stream, "                             begin\n") ;
           fprintf(stream, "                                 Berl_SM <= Berl_s8 ;\n") ;
           fprintf(stream, "                                 Poly_Len_Cnt <= Poly_Len_Cnt + 1 ;\n") ;
           fprintf(stream, "                              end\n") ;
           fprintf(stream, "                         else\n") ;
           fprintf(stream, "                             Berl_SM <= Berl_s9 ;\n") ;
           fprintf(stream, "                end \n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "            Berl_s9: begin\n") ;
           fprintf(stream, "                         L <= r - L ;\n") ;
           fprintf(stream, "                         Berl_SM <= Berl_s11 ;\n") ;
           fprintf(stream, "                end\n") ; 
           fprintf(stream, "\n") ;
           fprintf(stream, "            Berl_s10 : begin\n") ;
           fprintf(stream, "                         Berl_SM <= Berl_s11 ;\n") ;
           fprintf(stream, "                end\n") ; 
           fprintf(stream, "\n") ;
           fprintf(stream, "           Berl_s11: begin\n") ;
           fprintf(stream, "                         if (r == V)\n") ;
           fprintf(stream, "                             Berl_SM <= Berl_s12 ;\n") ;
           fprintf(stream, "                         else\n") ;
           fprintf(stream, "                             begin\n") ;
           fprintf(stream, "                                 Berl_SM <= Berl_s1 ;\n") ;
           fprintf(stream, "                                 j <= %d'b0 ;\n", vlen+1) ;
           fprintf(stream, "                             end\n") ;
           fprintf(stream, "                end \n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "            Berl_s12: begin\n") ;
           fprintf(stream, "                         Berl_SM <= Berl_s0 ;\n") ;
           fprintf(stream, "                end\n") ; 
           fprintf(stream, "\n") ;
           fprintf(stream, "            default: begin\n") ;
           fprintf(stream, "                end\n") ;
           fprintf(stream, "        endcase\n") ;
           fprintf(stream, "end\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    assign syn_sel = r - j ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    assign delta_syn = (syn_sel == syn_s0) ? s0 :\n") ;
             for (i=1; i < chk_sym; i++)
               fprintf(stream, "                       (syn_sel == syn_s%d) ? s%d :\n", i, i) ;
               fprintf(stream, "                       (syn_sel == syn_s%d) ? s%d : %d'b0 ;\n", i, i, gf_m) ;
		   fprintf(stream, "\n") ;
           fprintf(stream, "    assign lam_sel = j ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    assign delta_lam = (lam_sel == lam_l0) ? l_0 :\n") ;
             for (i=1; i < chk_sym+1; i++)
               fprintf(stream, "                       (lam_sel == lam_l%d) ? l_%d :\n", i, i) ;
               fprintf(stream, "                       (lam_sel == lam_l%d) ? l_%d : %d'b0 ;\n", i, i, gf_m) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    assign delta_r_mul = gfmul_ab(delta_syn, delta_lam)  ;\n") ;
           fprintf(stream, "    assign delta_r_calc = delta_r_mul ^ delta_r ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    assign #1 B_x_Op =  (Berl_SM == Berl_s0 && start_berl == 1'b1) ? x_init_one :\n") ;
           fprintf(stream, "                        (Berl_SM == Berl_s7 && (2*L <= r-1)) ? x_nop :\n") ;
           fprintf(stream, "                        (Berl_SM == Berl_s7 && (2*L > r-1)) ? x_shft_l_z :\n") ;
           fprintf(stream, "                        (Berl_SM == Berl_s9) ? x_load :\n") ;
           fprintf(stream, "                        //(Berl_SM == Berl_s10) ? x_load :\n") ;
           fprintf(stream, "                        x_nop ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    assign #1 L_x_Op =  (Berl_SM == Berl_s0 && start_berl == 1'b1) ? x_init_one :\n") ;
           fprintf(stream, "                        (Berl_SM == Berl_s7 && (2*L <= r-1)) ? x_nop :\n") ;
           fprintf(stream, "                        (Berl_SM == Berl_s7 && (2*L > r-1)) ? x_load :\n") ;
           fprintf(stream, "                        (Berl_SM == Berl_s9) ? x_load :\n") ;
           fprintf(stream, "                        (Berl_SM == Berl_s10) ? x_load :\n") ;
           fprintf(stream, "                        x_nop ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    assign #1 T_x_Op =  (Berl_SM == Berl_s0 && start_berl == 1'b1) ? x_init_zero :\n") ;
           fprintf(stream, "                        (Berl_SM == Berl_s6 && Poly_Len_Cnt == V) ? x_load :\n") ;
           fprintf(stream, "                        x_nop ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    assign #1 R_0_Op =  (Berl_SM == Berl_s0 && start_berl == 1'b1) ? x_init_zero :\n") ;
           fprintf(stream, "                        (Berl_SM == Berl_s3 && delta_r != %d'b0) ? x_load :\n", gf_m) ;
           fprintf(stream, "                        (Berl_SM == Berl_s4 ) ? x_shft_l :\n") ;
           fprintf(stream, "                        (Berl_SM == Berl_s5 && Poly_Len_Cnt < V ) ? x_shft_l_z :\n") ;
           fprintf(stream, "                        (Berl_SM == Berl_s5 && Poly_Len_Cnt == V ) ? x_load :\n") ;
           fprintf(stream, "                        (Berl_SM == Berl_s6 ) ? x_shft_l_z :\n") ;
           fprintf(stream, "                        x_nop ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    assign #1 R_1_Op =  (Berl_SM == Berl_s0 && start_berl == 1'b1) ? x_init_zero :\n") ;
           fprintf(stream, "                        (Berl_SM == Berl_s3 && delta_r != %d'b0 ) ? x_load :\n", gf_m) ;
           fprintf(stream, "                        (Berl_SM == Berl_s6 ) ? x_shft_l_z :\n") ;
           fprintf(stream, "                        (Berl_SM == Berl_s7 && (2*L <= r-1)) ? x_load :\n") ;
           fprintf(stream, "                        (Berl_SM == Berl_s8  && Poly_Len_Cnt < V) ? x_shft_l_z :\n") ;
           fprintf(stream, "                        x_nop ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    assign #1 R_2_Op =  (Berl_SM == Berl_s0 && start_berl == 1'b1) ? x_init_zero :\n") ;
           fprintf(stream, "                        (Berl_SM == Berl_s4 ) ? x_shft_l :\n") ;
           fprintf(stream, "                        (Berl_SM == Berl_s5 ) ? x_shft_l :\n") ;
           fprintf(stream, "                        (Berl_SM == Berl_s6 ) ? x_shft_l :\n") ;
           fprintf(stream, "                        (Berl_SM == Berl_s8 && Poly_Len_Cnt < V ) ? x_shft_l :\n") ;
           fprintf(stream, "                        x_nop ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "// \n") ;
           fprintf(stream, "//  Register R_0\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "poly_reg     R_0(.rst_n(rst_n),\n") ;
           fprintf(stream, "                 .clk(clk),\n") ;
           fprintf(stream, "                 .op_code(R_0_Op),\n") ;
           fprintf(stream, "                 .sd_in(%d'b0),\n", gf_m) ;
           fprintf(stream, "                 .ld_in(R_0_din),\n") ;
           fprintf(stream, "                 .d_out(R_0_Q)\n") ;
           fprintf(stream, "                ) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    assign R_0_din = (Berl_SM == Berl_s5 && Poly_Len_Cnt == V) ? R_2_Q : B_x_Q ;\n") ;
           fprintf(stream, "    assign R_0_mul = gfmul_ab(delta_r, R_0_Q[%d:%d]) ; \n", 2*chk_sym*gf_m-1, (2*chk_sym-1)*gf_m) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "// \n") ;
           fprintf(stream, "//  Register R_1\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "poly_reg     R_1(.rst_n(rst_n),\n") ;
           fprintf(stream, "                 .clk(clk),\n") ;
           fprintf(stream, "                 .op_code(R_1_Op),\n") ;
           fprintf(stream, "                 .sd_in(%d'b0),\n", gf_m) ;
           fprintf(stream, "                 .ld_in(L_x_Q),\n") ;
           fprintf(stream, "                 .d_out(R_1_Q)\n") ;
           fprintf(stream, "                ) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    assign R_1_add = R_1_Q[%d:%d] ^ R_0_Q[%d:%d] ;\n", 2*chk_sym*gf_m-1, (2*chk_sym-1)*gf_m, 2*chk_sym*gf_m-1, (2*chk_sym-1)*gf_m) ;
           fprintf(stream, "    assign R_1_mul = gfmul_ab(inv_delta_r, R_1_Q[%d:%d]) ;\n", 2*chk_sym*gf_m-1, (2*chk_sym-1)*gf_m) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "//\n") ; 
           fprintf(stream, "//  Register R_2\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "poly_reg     R_2(.rst_n(rst_n),\n") ;
           fprintf(stream, "                 .clk(clk),\n") ;
           fprintf(stream, "                 .op_code(R_2_Op),\n") ;
           fprintf(stream, "                 .sd_in(R_2_din),\n") ;
           fprintf(stream, "                 .ld_in(%d'b0),\n", (2*chk_sym*gf_m)+gf_m) ;
           fprintf(stream, "                 .d_out(R_2_Q)\n") ;
           fprintf(stream, "                ) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    assign R_2_din = (Berl_SM == Berl_s5) ? R_0_mul :\n") ;
           fprintf(stream, "                     (Berl_SM == Berl_s6) ? R_1_add :\n") ;
           fprintf(stream, "                     (Berl_SM == Berl_s8) ? R_1_mul :\n") ;
           fprintf(stream, "                     %d'h0 ;\n", gf_m) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "//\n") ; 
           fprintf(stream, "//        B(x)\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "poly_reg     B_x(.rst_n(rst_n),\n") ;
           fprintf(stream, "                 .clk(clk),\n") ;
           fprintf(stream, "                 .op_code(B_x_Op),\n") ;
           fprintf(stream, "                 .sd_in(%d'b0),\n", gf_m) ;
           fprintf(stream, "                 .ld_in(R_2_Q),\n") ;
           fprintf(stream, "                 .d_out(B_x_Q)\n") ;
           fprintf(stream, "                ) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "// \n") ;
           fprintf(stream, "//        L(x)\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "poly_reg     L_x(.rst_n(rst_n),\n") ;
           fprintf(stream, "                 .clk(clk),\n") ;
           fprintf(stream, "                 .op_code(L_x_Op),\n") ;
           fprintf(stream, "                 .sd_in(%d'h0),\n", gf_m) ;
           fprintf(stream, "                 .ld_in(T_x_Q),\n") ;
           fprintf(stream, "                 .d_out(L_x_Q)\n") ;
           fprintf(stream, "                ) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "//\n") ; 
           fprintf(stream, "//        T(x)\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "poly_reg     T_x(.rst_n(rst_n),\n") ;
           fprintf(stream, "                 .clk(clk),\n") ;
           fprintf(stream, "                 .op_code(T_x_Op),\n") ;
           fprintf(stream, "                 .sd_in(%d'b0),\n", gf_m) ;
           fprintf(stream, "                 .ld_in(R_2_Q),\n") ;
           fprintf(stream, "                 .d_out(T_x_Q)\n") ;
           fprintf(stream, "                ) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "always @(negedge rst_n or posedge clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n == 1'b0)\n") ;
           fprintf(stream, "        begin\n") ;
           fprintf(stream, "            loc_srch <= 1'b0 ;\n") ;
           fprintf(stream, "        end\n") ;
           fprintf(stream, "    else if (Berl_SM == Berl_s12)\n") ;
           fprintf(stream, "            loc_srch <= 1'b1 ;\n") ;
           fprintf(stream, "         else\n") ;
           fprintf(stream, "            loc_srch <= 1'b0 ;\n") ;
           fprintf(stream, "end\n") ;
           fprintf(stream, "endmodule \n") ;
       break;

       case VHDL :
           fprintf(stream, "------------------------------------------------------------------------\n") ;
           fprintf(stream, "--                  Generated by RSTK                                 --\n") ;  
           fprintf(stream, "--                      rs_encode.vhd                                 --\n") ;  
           fprintf(stream, "--                                                                    --\n") ;  
           fprintf(stream, "------------------------------------------------------------------------\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "LIBRARY IEEE ;\n") ;
           fprintf(stream, "use IEEE.std_logic_1164.all ;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "library IEEE ;\n") ;
           fprintf(stream, "use   IEEE.std_logic_unsigned.all;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "library IEEE ;\n") ;
           fprintf(stream, "use   IEEE.NUMERIC_STD.all;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "library gf_arith ;\n") ;
           fprintf(stream, "use   gf_arith.gf_arith.all;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "library io_lib ;\n") ;
           fprintf(stream, "use   io_lib.io_utils.all;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--library mem_lib ;\n") ;
           fprintf(stream, "--use   mem_lib.mem_lib.all;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "entity berl_mas  is port( rst_n : IN std_logic;\n") ;
           fprintf(stream, "                      clk : IN std_logic;\n") ;
           fprintf(stream, "                      start_berl : IN std_logic;\n") ;
             for(i=0; i < deg; i++)
               fprintf(stream, "                      s%d: IN std_logic_vector(%d downto 0);\n", i, gf_m-1) ;
             for(i=0; i < deg; i++)
           fprintf(stream, "                      l%d: OUT std_logic_vector(%d downto 0);\n", i, gf_m-1) ;
           fprintf(stream, "                      start_loc_search : OUT std_logic \n") ;
           fprintf(stream, "                      ) ;\n") ;
           fprintf(stream, "end berl_mas ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "architecture rtl of berl_mas is\n") ;    
           fprintf(stream, "-- Locals\n") ;
           fprintf(stream, "--    constant  N : integer := 94 ;\n") ;
           fprintf(stream, "--    constant  K : integer := 88 ;\n") ;
           fprintf(stream, "-- V=2*v\n") ;
           fprintf(stream, "    constant  V : std_logic_vector := \"") ;
             conv2_bvec_str(2*chk_sym, vlen, b_vector) ;
             pr_vec_str(stream, vlen, b_vector) ;
           fprintf(stream, "\" ;\n") ;            

           fprintf(stream, "\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "-- Delta_R\n") ;
           fprintf(stream, "    signal delta_r : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "    signal inv_delta_r : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "-- L Count\n") ;
           fprintf(stream, "    signal L   : std_logic_vector(%d downto 0);\n", vlen) ;
           fprintf(stream, "    signal Lx2 : std_logic_vector(%d downto 0);  -- 2 * L\n", vlen) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "-- r Count\n") ;
           fprintf(stream, "    signal r        : std_logic_vector(%d downto 0);\n", vlen) ;
           fprintf(stream, "    signal rminus_1 : std_logic_vector(%d downto 0);\n", vlen) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "-- j Count - used for Delta R calculation\n") ;
           fprintf(stream, "    signal j : std_logic_vector(%d downto 0);\n", vlen) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "-- Polynomial Lenght Count\n") ;  
           fprintf(stream, "    signal Poly_Len_Cnt: std_logic_vector(%d downto 0);\n", vlen) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "-- Syndrome Delta R calcultor\n") ;
           fprintf(stream, "    signal syn_sel : std_logic_vector(%d downto 0);\n", vlen) ;
           fprintf(stream, "    signal lam_sel : std_logic_vector(%d downto 0);\n", vlen) ;
           fprintf(stream, "    signal delta_syn : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "    signal delta_lam : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "    signal delta_r_calc : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "    signal delta_r_mul : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "-- Berl_SM States\n") ;
           fprintf(stream, "   type berl_sm_states is (\n") ;
           fprintf(stream, "        Berl_s0,\n") ;
           fprintf(stream, "        Berl_s1,\n") ;
           fprintf(stream, "        Berl_s2,\n") ;
           fprintf(stream, "        Berl_s3,\n") ;
           fprintf(stream, "        Berl_s4,\n") ;
           fprintf(stream, "        Berl_s5,\n") ;
           fprintf(stream, "        Berl_s6,\n") ;
           fprintf(stream, "        Berl_s7,\n") ;
           fprintf(stream, "        Berl_s8,\n") ;
           fprintf(stream, "        Berl_s9,\n") ;
           fprintf(stream, "        Berl_s10,\n") ;
           fprintf(stream, "        Berl_s11,\n") ;
           fprintf(stream, "        Berl_s12 \n") ; 
           fprintf(stream, "  ) ;\n") ;
           fprintf(stream, "-- Unit Control State Machine\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    signal Berl_SM : berl_sm_states ;\n") ;
           fprintf(stream, "    signal loc_srch : std_logic ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "-- B(x) register control modes\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "    signal B_x_Q : std_logic_vector(%d downto 0);\n", (2*chk_sym*gf_m-1)+gf_m) ;
           fprintf(stream, "    signal B_x_Op : std_logic_vector(2 downto 0);\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "-- L(x) register control modes\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "    signal L_x_Q : std_logic_vector(%d downto 0);\n", (2*chk_sym*gf_m-1)+gf_m) ;
           fprintf(stream, "    signal L_x_Op : std_logic_vector(2 downto 0);\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "-- T(x) register control modes\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "    signal T_x_Q : std_logic_vector(%d downto 0);\n", (2*chk_sym*gf_m-1)+gf_m) ;
           fprintf(stream, "    signal T_x_Op: std_logic_vector (2 downto 0);\n") ;
           fprintf(stream, "\n") ; 
           fprintf(stream, "--\n") ;
           fprintf(stream, "-- R(x) register control modes\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "    signal R_0_Q : std_logic_vector(%d downto 0);\n", (2*chk_sym*gf_m-1)+gf_m) ;
           fprintf(stream, "    signal R_1_Q : std_logic_vector(%d downto 0);\n", (2*chk_sym*gf_m-1)+gf_m) ;
           fprintf(stream, "    signal R_2_Q : std_logic_vector(%d downto 0);\n", (2*chk_sym*gf_m-1)+gf_m) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    signal R_2_din : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "    signal R_0_din : std_logic_vector(%d downto 0);\n", (2*chk_sym*gf_m-1)+gf_m) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    signal R_0_Op : std_logic_vector(2 downto 0);\n") ;
           fprintf(stream, "    signal R_1_Op : std_logic_vector(2 downto 0);\n") ;
           fprintf(stream, "    signal R_2_Op : std_logic_vector(2 downto 0);\n") ;
           fprintf(stream, "    signal R_1_add : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "    signal R_0_mul : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "    signal R_1_mul : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "\n") ;   
           fprintf(stream, "--\n") ;
           fprintf(stream, "--  Register control op modes\n") ; 
           fprintf(stream, "--\n") ; 
           fprintf(stream, "    constant x_nop       : std_logic_vector(2 downto 0) := \"000\" ;\n") ;
           fprintf(stream, "    constant x_load      : std_logic_vector(2 downto 0) := \"001\" ;\n") ;
           fprintf(stream, "    constant x_shft_l_z  : std_logic_vector(2 downto 0) := \"010\" ;\n") ;
           fprintf(stream, "    constant x_shft_l    : std_logic_vector(2 downto 0) := \"011\" ;\n") ;
           fprintf(stream, "    constant x_shft_r_z  : std_logic_vector(2 downto 0) := \"100\" ;\n") ;
           fprintf(stream, "    constant x_shft_r    : std_logic_vector(2 downto 0) := \"101\" ;\n") ;
           fprintf(stream, "    constant x_init_zero : std_logic_vector(2 downto 0) := \"110\" ;\n") ;
           fprintf(stream, "    constant x_init_one  : std_logic_vector(2 downto 0) := \"111\" ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--\n") ;
             for(i=0; i < chk_sym+2; i++)
               fprintf(stream, "   signal l_%d : std_logic_vector(%d downto 0);\n", i, gf_m-1) ;
             for(i=0; i < chk_sym+1; i++) {
               fprintf(stream, "   constant syn_s%d : std_logic_vector(%d downto 0) := \"", i, vlen) ;
           conv2_bvec_str(i, vlen, b_vector) ;
               pr_vec_str(stream, vlen, b_vector) ;
               fprintf(stream, "\" ;\n") ;               
       }
             for(i=0; i < chk_sym+2; i++) {
               fprintf(stream, "   constant lam_l%d : std_logic_vector(%d downto 0) := \"" , i, vlen) ;
           conv2_bvec_str(i, vlen, b_vector) ;
               pr_vec_str(stream, vlen, b_vector) ;
               fprintf(stream, "\" ;\n") ;               
           }
           fprintf(stream, "\n") ;
           fprintf(stream, "   component poly_reg is port (\n") ;
           fprintf(stream, "                         rst_n   : IN std_logic ;\n") ;
           fprintf(stream, "                         clk     : IN std_logic ;\n") ;
           fprintf(stream, "                         op_code : IN std_logic_vector(2 downto 0) ;\n") ;
           fprintf(stream, "                         sd_in   : IN std_logic_vector(%d downto 0) ;\n", gf_m-1) ;
           fprintf(stream, "                         ld_in   : IN std_logic_vector(%d downto 0) ;\n", (2*chk_sym*gf_m-1)+gf_m) ;
           fprintf(stream, "                         d_out   : OUT std_logic_vector(%d downto 0)\n", (2*chk_sym*gf_m-1)+gf_m) ;
           fprintf(stream, "                       );\n") ;
           fprintf(stream, "   end component ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "-- GF Inverse ROM Table\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "-- outputs the inverse of a^n (1/a^n) from input of a^n\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "-- Address => a^n\n") ;
           fprintf(stream, "-- Data Out => 1/a^n\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--\n") ; 
           fprintf(stream, "--------------------------------------------------------------------------\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--     ROM Tables for Berlekamp-Massey Algorithm\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--     This table is loaded into RAM during initialization\n") ;
           fprintf(stream, "--------------------------------------------------------------------------\n") ;
           fprintf(stream, "component spram_block is\n") ;
           fprintf(stream, "  generic ( RAM_NUM_WORDS : integer ;\n") ;
           fprintf(stream, "            RAM_WIDTH_A   : integer ;\n") ;
           fprintf(stream, "            RAM_WIDTH_D   : integer ;\n") ;
           fprintf(stream, "      RAM_FILE      : string\n") ; 
           fprintf(stream, "       );\n") ;
           fprintf(stream, "  port (clk  : in std_logic;\n") ;
           fprintf(stream, "        we   : in std_logic;\n") ;
           fprintf(stream, "        addr : in std_logic_vector(RAM_WIDTH_A-1 downto 0);\n") ;
           fprintf(stream, "        di   : in std_logic_vector(RAM_WIDTH_D-1 downto 0);\n") ;
           fprintf(stream, "        do   : out std_logic_vector(RAM_WIDTH_D-1 downto 0)\n") ;
           fprintf(stream, "    );\n") ;
           fprintf(stream, "end component;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    start_loc_search <= loc_srch ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "  berl_rom: spram_block\n") ;
           fprintf(stream, "       generic map( RAM_NUM_WORDS => %d,\n", gf_field) ;
           fprintf(stream, "                  RAM_WIDTH_A => %d,\n", gf_m) ;
           fprintf(stream, "          RAM_WIDTH_D => %d,\n", gf_m) ;
           fprintf(stream, "          RAM_FILE => \"gf_inv_rom.hex\")\n") ;
           fprintf(stream, "       port map(\n") ;
           fprintf(stream, "        clk => clk,\n") ;
           fprintf(stream, "        we => '0',\n") ;
           fprintf(stream, "        addr => delta_r,\n") ;
           fprintf(stream, "        di => \"") ;
		   for (i=0; i< gf_m ; i++)
               fprintf(stream, "0") ;
		   fprintf(stream, "\",\n") ;
           fprintf(stream, "        do => inv_delta_r\n") ;
           fprintf(stream, "    ) ;\n") ;
           fprintf(stream, "------------------------------------------------------------------------------------------\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--  Control State Machine\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "process(rst_n, clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n <= '0')then\n") ;
           fprintf(stream, "            delta_r <= (OTHERS => '0') ;\n") ;
           fprintf(stream, "            r <= (OTHERS => '0') ;\n") ;
           fprintf(stream, "            j <= (OTHERS => '0') ;\n") ;
           fprintf(stream, "            L <= (OTHERS => '0') ;\n") ;
           fprintf(stream, "            delta_r <= (OTHERS => '0') ;\n") ;
           fprintf(stream, "            Poly_Len_Cnt <= (OTHERS => '0') ;\n") ;
           fprintf(stream, "            Berl_SM <= Berl_s0 ;\n") ;
           fprintf(stream, "    elsif (clk'event and clk = '1')then\n") ;
           fprintf(stream, "        case Berl_SM is\n") ;
           fprintf(stream, "            when Berl_s0 =>\n") ;
           fprintf(stream, "                if (start_berl = '1') then\n") ;
           fprintf(stream, "                        r <= (OTHERS => '0') ;\n") ;
           fprintf(stream, "                        L <= (OTHERS => '0') ;\n") ;
           fprintf(stream, "                        Berl_SM <=  Berl_s1 ;\n") ;
           fprintf(stream, "                end if ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "            when Berl_s1 =>\n") ;
           fprintf(stream, "                     delta_r <= (OTHERS => '0') ;\n") ;
           fprintf(stream, "                     j <= (OTHERS => '0') ;\n") ;
           fprintf(stream, "                     Berl_SM <= Berl_s2 ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "            when Berl_s2 =>                 -- Calculate delta r\n") ;
           fprintf(stream, "                         if (j <= L ) then\n") ;
           fprintf(stream, "                             delta_r <= delta_r_calc ;\n") ;
           fprintf(stream, "                             j <= j + 1 ;\n") ;
           fprintf(stream, "                             Berl_SM <= Berl_s2 ;\n") ;
           fprintf(stream, "                         else\n") ;
           fprintf(stream, "                             r <= r + 1 ;\n") ;
           fprintf(stream, "                             Berl_SM <= Berl_s3 ;\n") ;
           fprintf(stream, "                         end if ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "            when Berl_s3 =>\n") ;
           fprintf(stream, "                         if (delta_r <= 0) then\n") ;
           fprintf(stream, "                             Berl_SM <= Berl_s11 ;\n") ;
           fprintf(stream, "                         else\n") ;
           fprintf(stream, "                             Berl_SM <= Berl_s4 ;\n") ;
           fprintf(stream, "                             Poly_Len_Cnt <= (OTHERS => '0') ;\n") ;
           fprintf(stream, "                         end if ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "            when Berl_s4 =>\n") ;
           fprintf(stream, "                         Berl_SM <= Berl_s5 ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "            when Berl_s5 =>                    -- Calculate delta_r*x*B(x)\n") ;
           fprintf(stream, "                         if (Poly_Len_Cnt < V) then\n") ;
           fprintf(stream, "                             Berl_SM <= Berl_s5 ;\n") ;
           fprintf(stream, "                             Poly_Len_Cnt <= Poly_Len_Cnt + 1 ;\n") ;
           fprintf(stream, "                         else\n") ;
           fprintf(stream, "                             Berl_SM <= Berl_s6 ;\n") ;
           fprintf(stream, "                             Poly_Len_Cnt <= (OTHERS => '0') ;\n") ;
           fprintf(stream, "                         end if ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "            when Berl_s6 =>                  -- Calculate T(x) = L(x) + delta_r*x*B(x)\n") ;
           fprintf(stream, "                         if (Poly_Len_Cnt < V) then\n") ;
           fprintf(stream, "                             Berl_SM <= Berl_s6 ;\n") ;
           fprintf(stream, "                             Poly_Len_Cnt <= Poly_Len_Cnt + 1 ;\n") ;
           fprintf(stream, "                         else\n") ;
           fprintf(stream, "                             Berl_SM <= Berl_s7 ;\n") ;
           fprintf(stream, "                         end if ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "            when Berl_s7 =>\n") ;
           fprintf(stream, "                         if (Lx2 <= rminus_1) then\n") ;
           fprintf(stream, "                             Berl_SM <= Berl_s8 ;\n") ;
           fprintf(stream, "                             Poly_Len_Cnt <= (OTHERS => '0') ;\n") ;
           fprintf(stream, "                         else\n") ;
           fprintf(stream, "                             Berl_SM <= Berl_s10 ;\n") ;
           fprintf(stream, "                         end if ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "            when Berl_s8 =>                   -- Calculate (delta_r^-1)*L(x)\n") ;
           fprintf(stream, "                         if (Poly_Len_Cnt < V) then\n") ;
           fprintf(stream, "                             Berl_SM <= Berl_s8 ;\n") ;
           fprintf(stream, "                             Poly_Len_Cnt <= Poly_Len_Cnt + 1 ;\n") ;
           fprintf(stream, "                         else\n") ;
           fprintf(stream, "                             Berl_SM <= Berl_s9 ;\n") ;
           fprintf(stream, "               end if ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "            when Berl_s9 =>\n") ;
           fprintf(stream, "                         L <= r - L ;\n") ;
           fprintf(stream, "                         Berl_SM <= Berl_s11 ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "            when Berl_s10 =>\n") ;
           fprintf(stream, "                         Berl_SM <= Berl_s11 ;\n") ;
           fprintf(stream, "\n") ; 
           fprintf(stream, "            when Berl_s11 =>\n") ;
           fprintf(stream, "                         if (r = V) then\n") ;
           fprintf(stream, "                             Berl_SM <= Berl_s12 ;\n") ;
           fprintf(stream, "                         else\n") ;
           fprintf(stream, "                             Berl_SM <= Berl_s1 ;\n") ;
           fprintf(stream, "                             j <= (OTHERS => '0') ;\n") ;
           fprintf(stream, "                         end if ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "            when Berl_s12 =>\n") ;
           fprintf(stream, "                         Berl_SM <= Berl_s0 ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "            when OTHERS =>\n") ;
           fprintf(stream, "                         Berl_SM <= Berl_s0 ;\n") ;
           fprintf(stream, "        end case ;\n") ;
           fprintf(stream, "  end if ;\n") ;
           fprintf(stream, "end process ;\n") ;
           fprintf(stream, "-- Syndrome selection\n") ;
           fprintf(stream, "     syn_sel <= r - j ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "   with syn_sel select\n") ;
           fprintf(stream, "         delta_syn <=  s0 when syn_s0,\n") ;
             for(i=1; i < 2*chk_sym; i++)
               fprintf(stream, "                       s%d when syn_s%d,\n", i, i) ;
           fprintf(stream, "              (OTHERS => '0')  when others ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "-- Lambda selection\n") ;
           fprintf(stream, "     lam_sel <= j ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "   with lam_sel select\n") ;
           fprintf(stream, "         delta_lam <= l_0 when lam_l0,\n") ;
             for(i=1; i < 2*chk_sym; i++)
               fprintf(stream, "                      l_%d when lam_l%d,\n", i, i) ;
           fprintf(stream, "              (OTHERS => '0')  when others ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "     delta_r_mul <= gfmul_ab(delta_syn, delta_lam)  ;\n") ;
           fprintf(stream, "     delta_r_calc <= delta_r_mul xor delta_r ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "      Lx2(%d downto 0) <= L(%d downto 0) & '0' ;\n", vlen, vlen-1) ;
           fprintf(stream, "      rminus_1 <= r - 1 ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "      B_x_Op <=  x_init_one when (Berl_SM = Berl_s0 and start_berl = '1') else\n") ;
           fprintf(stream, "                      x_nop when (Berl_SM = Berl_s7 and (Lx2 <= rminus_1)) else\n") ; 
           fprintf(stream, "                 x_shft_l_z when (Berl_SM = Berl_s7 and (Lx2 > rminus_1)) else\n") ;
           fprintf(stream, "                 x_shft_l_z when (Berl_SM = Berl_s3 and delta_r = 0 ) else\n") ;
           fprintf(stream, "                     x_load when (Berl_SM = Berl_s9) else\n") ;
           fprintf(stream, "                     x_nop ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "      L_x_Op <=  x_init_one when (Berl_SM = Berl_s0 and start_berl = '1') else\n") ;
           fprintf(stream, "                      x_nop when (Berl_SM = Berl_s7 and (Lx2 <= rminus_1)) else\n") ;
           fprintf(stream, "                     x_load when (Berl_SM = Berl_s7 and (Lx2 > rminus_1)) else\n") ; 
           fprintf(stream, "                     x_load when (Berl_SM = Berl_s9) else\n") ;
           fprintf(stream, "                     x_load when (Berl_SM = Berl_s10) else\n") ;
           fprintf(stream, "                      x_nop ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "      T_x_Op <= x_init_zero when (Berl_SM = Berl_s0 and start_berl = '1') else\n") ;
           fprintf(stream, "                     x_load when (Berl_SM = Berl_s6 and Poly_Len_Cnt = V) else\n") ;
           fprintf(stream, "                      x_nop ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "      R_0_Op <= x_init_zero when (Berl_SM = Berl_s0 and start_berl = '1') else\n") ; 
           fprintf(stream, "                     x_load when (Berl_SM = Berl_s3 and delta_r /= 0 ) else\n") ;
           fprintf(stream, "                   x_shft_l when (Berl_SM = Berl_s4 ) else\n") ;
           fprintf(stream, "                 x_shft_l_z when (Berl_SM = Berl_s5 and Poly_Len_Cnt < V ) else\n") ;
           fprintf(stream, "                     x_load when (Berl_SM = Berl_s5 and Poly_Len_Cnt = V ) else\n") ;
           fprintf(stream, "                 x_shft_l_z when (Berl_SM = Berl_s6 ) else\n") ;
           fprintf(stream, "                     x_nop ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "      R_1_Op <= x_init_zero when (Berl_SM = Berl_s0 and start_berl = '1') else\n") ;
           fprintf(stream, "                     x_load when (Berl_SM = Berl_s3 and delta_r /= 0 ) else\n") ;
           fprintf(stream, "                 x_shft_l_z when (Berl_SM = Berl_s6 ) else\n") ;
           fprintf(stream, "                     x_load when (Berl_SM = Berl_s7 and (Lx2 <= rminus_1)) else\n") ;
           fprintf(stream, "                 x_shft_l_z when (Berl_SM = Berl_s8  and Poly_Len_Cnt < V) else\n") ;
           fprintf(stream, "                     x_nop ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "      R_2_Op <= x_init_zero when (Berl_SM = Berl_s0 and start_berl = '1') else\n") ;
           fprintf(stream, "                   x_shft_l when (Berl_SM = Berl_s4 ) else\n") ;
           fprintf(stream, "                   x_shft_l when (Berl_SM = Berl_s5 ) else\n") ;
           fprintf(stream, "                   x_shft_l when (Berl_SM = Berl_s6 ) else\n") ;
           fprintf(stream, "                   x_shft_l when (Berl_SM = Berl_s8 and Poly_Len_Cnt < V ) else\n") ;
           fprintf(stream, "                   x_nop ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--\n") ; 
           fprintf(stream, "--  Register R_0\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "    R_0 : poly_reg port map(rst_n => rst_n,\n") ;
           fprintf(stream, "                    clk => clk,\n") ;
           fprintf(stream, "                    op_code => R_0_Op,\n") ;
           fprintf(stream, "                    sd_in => \"") ;
             for(i=0; i < gf_m; i++) 
               fprintf(stream, "%c", zero_vector[i]) ;
               fprintf(stream, "\",\n") ;
           fprintf(stream, "                    ld_in => R_0_din,\n") ;
           fprintf(stream, "                    d_out => R_0_Q\n") ;
           fprintf(stream, "                    ) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "     R_0_din <=  R_2_Q when (Berl_SM = Berl_s5 and Poly_Len_Cnt = V) else B_x_Q ;\n") ;
           fprintf(stream, "     R_0_mul <= gfmul_ab(delta_r, R_0_Q(%d downto %d)) ;\n", 2*chk_sym*gf_m-1, (2*chk_sym-1)*gf_m) ; 
           fprintf(stream, "\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--\n") ; 
           fprintf(stream, "--  Register R_1\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "    R_1 :  poly_reg port map(rst_n => rst_n,\n") ;
           fprintf(stream, "                    clk => clk,\n") ;
           fprintf(stream, "                    op_code => R_1_Op,\n") ;
           fprintf(stream, "                    sd_in => \"") ;
             for(i=0; i < gf_m; i++) 
               fprintf(stream, "%c", zero_vector[i]) ;
               fprintf(stream, "\",\n") ;
           fprintf(stream, "                    ld_in => L_x_Q,\n") ;
           fprintf(stream, "                    d_out => R_1_Q\n") ;
           fprintf(stream, "                   ) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    R_1_add <= R_1_Q(%d downto %d) xor R_0_Q(%d downto %d) ;\n", 2*chk_sym*gf_m-1, (2*chk_sym-1)*gf_m, 2*chk_sym*gf_m-1, (2*chk_sym-1)*gf_m) ;
           fprintf(stream, "    R_1_mul <= gfmul_ab(inv_delta_r(%d downto 0), R_1_Q(%d downto %d)) ;\n", gf_m-1, 2*chk_sym*gf_m-1, (2*chk_sym-1)*gf_m) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--\n") ; 
           fprintf(stream, "--  Register R_2\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "    R2 : poly_reg port map(rst_n => rst_n,\n") ;
           fprintf(stream, "                  clk => clk,\n") ;
           fprintf(stream, "                  op_code => R_2_Op,\n") ;
           fprintf(stream, "                  sd_in => R_2_din,\n") ;
           fprintf(stream, "                  ld_in => \"") ;
             for(i=0; i < (2*chk_sym*gf_m)+gf_m; i++) 
               fprintf(stream, "%c", '0') ;
               fprintf(stream, "\",\n") ;
           fprintf(stream, "                  d_out => R_2_Q\n") ;
           fprintf(stream, "                 ) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "     R_2_din <= R_0_mul when (Berl_SM = Berl_s5) else\n") ;
           fprintf(stream, "                R_1_add when (Berl_SM = Berl_s6) else\n") ;
           fprintf(stream, "                R_1_mul when (Berl_SM = Berl_s8) else\n") ;
           fprintf(stream, "                (OTHERS => '0') ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "--\n") ; 
           fprintf(stream, "--        B(x)\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "    B_X : poly_reg port map(rst_n => rst_n,\n") ;
           fprintf(stream, "                   clk => clk,\n") ;
           fprintf(stream, "                   op_code => B_x_Op,\n") ;
           fprintf(stream, "                   sd_in => \"") ;
             for(i=0; i < gf_m; i++) 
               fprintf(stream, "%c", zero_vector[i]) ;
               fprintf(stream, "\",\n") ;
           fprintf(stream, "                   ld_in => R_2_Q,\n") ;
           fprintf(stream, "                   d_out => B_x_Q\n") ;
           fprintf(stream, "                  ) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "--\n") ; 
           fprintf(stream, "--        L(x)\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "    L_x : poly_reg port map( rst_n => rst_n,\n") ;
           fprintf(stream, "                   clk => clk,\n") ;
           fprintf(stream, "                   op_code => L_x_Op,\n") ;
           fprintf(stream, "                   sd_in => \"") ;
             for(i=0; i < gf_m; i++) 
               fprintf(stream, "%c", zero_vector[i]) ;
               fprintf(stream, "\",\n") ;
           fprintf(stream, "                   ld_in => T_x_Q,\n") ;
           fprintf(stream, "                   d_out => L_x_Q\n") ;
           fprintf(stream, "                  ) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    l0 <= L_x_Q(%d downto 0);\n", gf_m-1) ;
             for(i=1; i < 2*chk_sym; i++)
               fprintf(stream, "    l%d <= L_x_Q(%d downto %d);\n", i, ((i+1)*gf_m)-1, gf_m*i) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    l_0 <= L_x_Q(%d downto 0);\n", gf_m-1) ;
             for(i=1; i < 2*chk_sym; i++)
               fprintf(stream, "    l_%d <= L_x_Q(%d downto %d);\n", i, ((i+1)*gf_m)-1, gf_m*i) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "--\n") ; 
           fprintf(stream, "--        T(x)\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "    T_x : poly_reg port map(rst_n => rst_n,\n") ;
           fprintf(stream, "                   clk => clk,\n") ;
           fprintf(stream, "                   op_code => T_x_Op,\n") ;
           fprintf(stream, "                   sd_in => \"") ;
             for(i=0; i < gf_m; i++) 
               fprintf(stream, "%c", zero_vector[i]) ;
               fprintf(stream, "\",\n") ;
           fprintf(stream, "                   ld_in => R_2_Q,\n") ;
           fprintf(stream, "                   d_out => T_x_Q\n") ;
           fprintf(stream, "                 ) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "process(rst_n, clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n = '0') then\n") ;
           fprintf(stream, "            loc_srch <= '0' ;\n") ;
           fprintf(stream, "  elsif (clk'event and clk = '1') then\n") ;
           fprintf(stream, "         if (Berl_SM = Berl_s12) then\n") ;
           fprintf(stream, "            loc_srch <= '1' ;\n") ;
           fprintf(stream, "         else\n") ;
           fprintf(stream, "            loc_srch <= '0' ;\n") ;
           fprintf(stream, "     end if ;\n") ;
           fprintf(stream, "  end if ;\n") ;
           fprintf(stream, "end process ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "END ;\n") ;
       break ;
     default :
       break ;
   }
}
//
// 
void gen_rs_polyreg(FILE *stream, int base, int gf_m, int deg, int chk_sym, int hdl_type)
{

    int i ;

   switch (hdl_type) {
       case C:
           fprintf(stream, "//--------------------------------------------------------------------//\n") ;
           fprintf(stream, "//                  Generated by RSTK                                 //\n") ;  
           fprintf(stream, "//                      rs_polymath.c                                   //\n") ;  
           fprintf(stream, "//                                                                    //\n") ;  
           fprintf(stream, "//--------------------------------------------------------------------//\n") ;
       gen_rs_cutil(stream) ;
       break ;
       case VLOG:
           fprintf(stream, "//--------------------------------------------------------------------//\n") ;
           fprintf(stream, "//                  Generated by RSTK                                 //\n") ;  
           fprintf(stream, "//                      rs_polyreg.v                                   //\n") ;  
           fprintf(stream, "//                                                                    //\n") ;  
           fprintf(stream, "//--------------------------------------------------------------------//\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "`timescale 1ns/100ps\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "module poly_reg( rst_n,\n") ;
           fprintf(stream, "                 clk,\n") ;
           fprintf(stream, "                 op_code,\n") ;
           fprintf(stream, "                 sd_in,\n") ;
           fprintf(stream, "                 ld_in,\n") ;
           fprintf(stream, "                 d_out\n") ;
           fprintf(stream, "                );\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    input         rst_n ;\n") ;
           fprintf(stream, "    input         clk ;\n") ;
           fprintf(stream, "    input[2:0]    op_code;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    input[%d:0]    sd_in ;\n", gf_m-1) ;
           fprintf(stream, "    input[%d:0]   ld_in ;\n", (2*chk_sym*gf_m-1)+gf_m) ;
           fprintf(stream, "    output[%d:0]  d_out ;\n", (2*chk_sym*gf_m-1)+gf_m) ;
           fprintf(stream, "\n") ;    
           fprintf(stream, "// \n") ;
           fprintf(stream, "// Registers\n") ; 
           fprintf(stream, "//\n") ;
             for (i=0; i<chk_sym+2; i++)
               fprintf(stream, "    reg[%d:0]      r_%d ;\n", gf_m-1, i) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    wire[%d:0]    reg_out ;\n", (2*chk_sym*gf_m-1)+gf_m) ;    
           fprintf(stream, "\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "//  Register control modes\n") ; 
           fprintf(stream, "//\n") ; 
           fprintf(stream, "    parameter    nop       = 0 ;\n") ;
           fprintf(stream, "    parameter    load      = 1 ;\n") ;
           fprintf(stream, "    parameter    shift_l_z = 2 ;\n") ;
           fprintf(stream, "    parameter    shift_l   = 3 ;\n") ;
           fprintf(stream, "    parameter    shift_r_z = 4 ;\n") ;
           fprintf(stream, "    parameter    shift_r   = 5 ;\n") ;
           fprintf(stream, "    parameter    init_zero = 6 ;\n") ;
           fprintf(stream, "    parameter    init_one  = 7 ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "always @(negedge rst_n or posedge clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n == 1'b0)\n") ;
           fprintf(stream, "        begin\n") ;
             for (i=0; i<chk_sym+2; i++)
               fprintf(stream, "            r_%d <= %d'b0 ;\n", i, gf_m) ;
           fprintf(stream, "        end\n") ;
           fprintf(stream, "    else\n") ; 
           fprintf(stream, "        case (op_code)\n") ;
           fprintf(stream, "            nop:\n") ;
           fprintf(stream, "                begin\n") ;
           fprintf(stream, "                end\n") ;
           fprintf(stream, "            load:\n") ;
           fprintf(stream, "                begin\n") ;
             for (i=0; i<chk_sym+2; i++)
               fprintf(stream, "                    r_%d <= ld_in[%d:%d] ;\n", i, (i+1)*gf_m-1, i*gf_m ) ;
           fprintf(stream, "                end\n") ;
           fprintf(stream, "            shift_l_z:\n") ;
           fprintf(stream, "                begin\n") ;
           fprintf(stream, "                    r_0 <= %d'b0 ;\n", gf_m) ;
             for (i=0; i<chk_sym+1; i++)
               fprintf(stream, "                    r_%d <= r_%d ;\n", i+1, i) ;
           fprintf(stream, "                end\n") ;
           fprintf(stream, "            shift_l:\n") ;
           fprintf(stream, "                begin\n") ;
           fprintf(stream, "                    r_0 <= sd_in[%d:0] ;\n", gf_m-1) ;
             for (i=0; i<chk_sym+1; i++)
               fprintf(stream, "                    r_%d <= r_%d ;\n", i+1, i) ;
           fprintf(stream, "                end\n") ;
           fprintf(stream, "            shift_r_z:\n") ;
           fprintf(stream, "                begin\n") ;
             for (i=0; i<chk_sym+1; i++)
               fprintf(stream, "                    r_%d <= r_%d ;\n", i, i+1) ;
           fprintf(stream, "                    r_%d <= %d'b0 ;\n", i, gf_m) ;
           fprintf(stream, "                end\n") ;
           fprintf(stream, "            shift_r:\n") ;
           fprintf(stream, "                begin\n") ;
             for (i=0; i<chk_sym+1; i++)
               fprintf(stream, "                    r_%d <= r_%d ;\n", i, i+1) ;
           fprintf(stream, "                    r_%d <= sd_in[%d:0] ;\n", i, gf_m-1) ;
           fprintf(stream, "                end\n") ;
           fprintf(stream, "            init_zero:\n") ;
           fprintf(stream, "                begin\n") ;
             for (i=0; i<chk_sym+2; i++)
               fprintf(stream, "                    r_%d <= %d'b0 ;\n", i, gf_m) ;
           fprintf(stream, "                end\n") ;
           fprintf(stream, "            init_one:\n") ;
           fprintf(stream, "                begin\n") ;
           fprintf(stream, "                    r_0[0] <= 1'b1 ;\n") ;
           fprintf(stream, "                    r_0[%d:1] <= %d'b0 ;\n", gf_m-1, gf_m-1) ;
             for (i=1; i<chk_sym+2; i++)
               fprintf(stream, "                    r_%d <= %d'b0 ;\n", i, gf_m) ;
           fprintf(stream, "                end\n") ;
           fprintf(stream, "            default:\n") ;
           fprintf(stream, "                begin\n") ;
           fprintf(stream, "                end\n") ;
           fprintf(stream, "        endcase\n") ;
           fprintf(stream, "end\n") ; 
           fprintf(stream, "        assign reg_out[%d:0] = {r_%d,\n", (2*chk_sym*gf_m-1)+gf_m, chk_sym+1) ;
             for (i= chk_sym; i > 0; i--)
               fprintf(stream, "                                r_%d,\n", i) ;
           fprintf(stream, "                                r_%d } ;\n", i) ;
           fprintf(stream, "        assign d_out = reg_out ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "endmodule \n") ;
           fprintf(stream, "\n") ;
       break ;

       case VHDL:
           fprintf(stream, "------------------------------------------------------------------------\n") ;
           fprintf(stream, "--                  Generated by RSTK                                 --\n") ;  
           fprintf(stream, "--                      rs_polyreg.vhd                                --\n") ;  
           fprintf(stream, "--                                                                    --\n") ;  
           fprintf(stream, "------------------------------------------------------------------------\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "LIBRARY IEEE ;\n") ;
           fprintf(stream, "use IEEE.std_logic_1164.all ;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "library IEEE ;\n") ;
           fprintf(stream, "use   IEEE.std_logic_unsigned.all;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "library IEEE ;\n") ;
           fprintf(stream, "use   IEEE.NUMERIC_STD.all;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "entity poly_reg is port (rst_n   : IN std_logic ;\n") ;
           fprintf(stream, "                         clk     : IN std_logic ;\n") ;
           fprintf(stream, "                         op_code : IN std_logic_vector(2 downto 0) ;\n") ;
           fprintf(stream, "                         sd_in   : IN std_logic_vector(%d downto 0) ;\n", gf_m-1) ;
           fprintf(stream, "                         ld_in   : IN std_logic_vector(%d downto 0) ;\n", (2*chk_sym*gf_m-1)+gf_m) ;
           fprintf(stream, "                         d_out   : OUT std_logic_vector(%d downto 0)\n", (2*chk_sym*gf_m-1)+gf_m) ;
           fprintf(stream, "                        );\n") ;
           fprintf(stream, "end poly_reg ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "architecture rtl of poly_reg is\n") ;
           fprintf(stream, "--\n") ; 
           fprintf(stream, "-- Registers\n") ; 
           fprintf(stream, "--\n") ;
             for (i=0; i < chk_sym+2; i++) 
               fprintf(stream, "    signal      r_%d : std_logic_vector(%d downto 0) ;\n", i, gf_m-1) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    signal    reg_out : std_logic_vector(%d downto 0) ;\n", (2*chk_sym*gf_m-1)+gf_m) ;   
           fprintf(stream, "\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--  Register control modes \n") ;
           fprintf(stream, "--\n") ; 
           fprintf(stream, "    constant nop       : std_logic_vector(2 downto 0) := \"000\" ;\n") ;
           fprintf(stream, "    constant load      : std_logic_vector(2 downto 0) := \"001\" ;\n") ;
           fprintf(stream, "    constant shift_l_z : std_logic_vector(2 downto 0) := \"010\" ;\n") ;
           fprintf(stream, "    constant shift_l   : std_logic_vector(2 downto 0) := \"011\" ;\n") ;
           fprintf(stream, "    constant shift_r_z : std_logic_vector(2 downto 0) := \"100\" ;\n") ;
           fprintf(stream, "    constant shift_r   : std_logic_vector(2 downto 0) := \"101\" ;\n") ;
           fprintf(stream, "    constant init_zero : std_logic_vector(2 downto 0) := \"110\" ;\n") ;
           fprintf(stream, "    constant init_one  : std_logic_vector(2 downto 0) := \"111\" ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "process(rst_n, clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n = '0') then\n") ;
             for (i=0; i < chk_sym+2; i++) 
               fprintf(stream, "            r_%d <= (OTHERS => '0') ;\n", i) ;
           fprintf(stream, "    elsif (clk'event and clk = '1') then\n") ;
           fprintf(stream, "        case op_code is\n") ;
           fprintf(stream, "            when nop =>\n") ;
           fprintf(stream, "\n") ; 
           fprintf(stream, "            when load =>\n") ;
             for (i=0; i < chk_sym+2; i++)
               fprintf(stream, "                    r_%d <= ld_in(%d downto %d) ;\n", i, (i+1)*gf_m-1, i*gf_m) ;
           fprintf(stream, "            when shift_l_z =>\n") ;
           fprintf(stream, "                    r_0 <= (OTHERS => '0') ;\n") ;
             for (i=0; i < chk_sym+1; i++)
               fprintf(stream, "                    r_%d <= r_%d ;\n", i+1, i) ;
           fprintf(stream, "            when shift_l =>\n") ;
           fprintf(stream, "                    r_0 <= sd_in(%d downto 0) ;\n", gf_m-1) ;
             for (i=0; i < chk_sym+1; i++)
               fprintf(stream, "                    r_%d <= r_%d ;\n", i+1, i) ;
           fprintf(stream, "            when shift_r_z =>\n") ;
             for (i=0; i < chk_sym+1; i++)
               fprintf(stream, "                    r_%d <= r_%d ;\n", i, i+1) ;
           fprintf(stream, "                    r_%d <= (OTHERS => '0') ;\n", i) ;
           fprintf(stream, "            when shift_r =>\n") ;
             for (i=0; i < chk_sym+1; i++)
               fprintf(stream, "                    r_%d <= r_%d ;\n", i, i+1) ;
           fprintf(stream, "                    r_%d <= sd_in(%d downto 0) ;\n", i, gf_m-1) ;
           fprintf(stream, "            when init_zero =>\n") ;
             for (i=0; i < chk_sym+2; i++)
               fprintf(stream, "                    r_%d <= (OTHERS => '0') ;\n", i) ;
           fprintf(stream, "            when init_one =>\n") ;
           fprintf(stream, "                    r_0(0) <= '1' ;\n") ;
           fprintf(stream, "                    r_0(%d downto 1) <= (OTHERS => '0')  ;\n", gf_m-1) ;
             for (i=1; i < chk_sym+2; i++)
               fprintf(stream, "                    r_%d <= (OTHERS => '0') ;\n", i) ;
           fprintf(stream, "            when OTHERS =>\n") ;
           fprintf(stream, "        end case ;\n") ;
           fprintf(stream, "  end if ;\n") ;
           fprintf(stream, "end process ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "        reg_out(%d downto 0) <= r_%d \n", (2*chk_sym*gf_m-1)+gf_m,chk_sym+1) ;
         for (i= chk_sym; i > 0; i--)
           fprintf(stream, "                              & r_%d \n", i) ;
           fprintf(stream, "                              & r_%d ; \n", i) ;
           fprintf(stream, "        d_out <= reg_out ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "END ;\n") ;
       break ;

     default :
       break ;
   }
}

