/*******************************************************************************/
/*                                                                             */
/*            gen_chien.c                                                      */
/*                                                                             */
/*    Generate RS Receiver Chien Search                                        */
/*                                                                             */
/*                                                                             */
/*                                                                             */
/*                                                                             */
/*                                                                             */
/*                                                                             */
/*                                                                             */
/*    Copyright (C) 2003  John S. Willy (jswilly@pacbell.net)                  */ 
/*                                                                             */
/*    This program is free software; you can redistribute it and/or modify     */
/*    it under the terms of the GNU General Public License as published by     */
/*    the Free Software Foundation; either version 2 of the License, or        */
/*    (at your option) any later version.                                      */
/*                                                                             */
/*    This program is distributed in the hope that it will be useful,          */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*    GNU General Public License for more details.                             */
/*                                                                             */
/*    You should have received a copy of the GNU General Public License        */
/*    along with this program; if not, write to the Free Software              */
/*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA*/
/*******************************************************************************/

#include <stdio.h>
#include "rsccg_proto.h"

#define VHDL     0 
#define VLOG     1 
#define C        2 

extern unsigned char poly_table[] ;   // RS code table 
extern unsigned char log_poly_table[] ; 
extern unsigned int p_wsp ;           // RS Generator Polynomial

/* locals  */

void gen_rs_chien(FILE *stream, int base, int gf_m, int deg, int chk_sym, int hdl_type)
{
   int i , t, vlen, gf_feild_size;
   unsigned char b_vector[16] ; // Binary bit vector

   vlen = vector_len(2*chk_sym) ;
   t = 2*chk_sym ;
   gf_feild_size = 1 ;
   for(i=0; i<gf_m; i++) gf_feild_size = 2*gf_feild_size ;

   switch (hdl_type) {
       case C:
           fprintf(stream, "//--------------------------------------------------------------------//\n") ;
           fprintf(stream, "//                  Generated by RSTK                                 //\n") ;  
           fprintf(stream, "//                      rs_chien.c                                    //\n") ;  
           fprintf(stream, "//                                                                    //\n") ;  
           fprintf(stream, "//--------------------------------------------------------------------//\n") ;
           fprintf(stream, "#include <stdio.h>\n") ;
           fprintf(stream, "#include \"proto_types.h\"\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "void chien_search_hw( v, lam, r, lam_roots )\n") ;
           fprintf(stream, "int v ;\n") ;
           fprintf(stream, "unsigned char *lam ;\n") ;
           fprintf(stream, "int *r ;\n") ;
           fprintf(stream, "unsigned char *lam_roots ;\n") ;
           fprintf(stream, "{\n") ;
           fprintf(stream, "unsigned char root_calc ;\n") ;
           fprintf(stream, "int i, t[%d], n_roots  ;\n", chk_sym+1) ;
           fprintf(stream, "unsigned char c0,\n") ;
           for (i=1; i<chk_sym; i++)
               fprintf(stream, "              c%d,\n", i) ;
           fprintf(stream, "              c%d ;\n", i) ;
           fprintf(stream, "\n") ;   
           fprintf(stream, "   n_roots = 0 ;\n") ;
           fprintf(stream, "   for (i=0; i<v; i++) {\n") ;
           fprintf(stream, "     t[i] = *lam++ ;\n") ;
           fprintf(stream, "   }\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "\n") ;
           for (i=0; i<=chk_sym; i++)
               fprintf(stream, "   c%d = t[%d] ;\n", i, i) ; 
           fprintf(stream, "\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "   for (i=0; i<%d ; i++) {\n", gf_feild_size) ;
           fprintf(stream, "     root_calc = c0 ^ \n") ;
           for (i=1; i< chk_sym; i++)
               fprintf(stream, "                 c%d ^\n", i) ;
           fprintf(stream, "                 c%d ;\n", i) ;
           fprintf(stream, "     if (root_calc == 0) {\n") ; 
           fprintf(stream, "        *lam_roots++ = i ;\n") ;
           fprintf(stream, "        printf(\"root = %%2d root_calc = %%2x\\n \", i, root_calc) ;\n") ;
           fprintf(stream, "        n_roots =  n_roots + 1 ;\n") ;
           fprintf(stream, "     }\n") ;
           for (i=1; i<= chk_sym; i++)
               fprintf(stream, "     c%d = gfmul_%d(c%d) ;\n", i, i, i) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "   }\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "   *r = n_roots ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "}\n") ;


           break;

       case VLOG : 
           fprintf(stream, "//--------------------------------------------------------------------//\n") ;
           fprintf(stream, "//                  Generated by RSTK                                 //\n") ;  
           fprintf(stream, "//                      rs_chien.v                                   //\n") ;  
           fprintf(stream, "//                                                                    //\n") ;  
           fprintf(stream, "//--------------------------------------------------------------------//\n") ;

           fprintf(stream, "module chien_srch( rst_n,\n") ;
           fprintf(stream, "           clk,\n") ;
           fprintf(stream, "           start_search,\n") ;
           for (i=0; i<2*chk_sym; i++)
               fprintf(stream, "           l%d,\n", i) ;
           for (i=0; i<chk_sym; i++)
               fprintf(stream, "           lroot%d,\n", i) ; 
           fprintf(stream, "           n_roots,\n") ;
           fprintf(stream, "           search_done \n") ;
           fprintf(stream, "         ) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    input        rst_n ;\n") ;
           fprintf(stream, "    input        clk ;\n") ;
           fprintf(stream, "    input        init ;\n") ;
           fprintf(stream, "    input        start_search ;\n") ;
           fprintf(stream, "\n") ;
           for (i=0; i<2*chk_sym; i++)
               fprintf(stream, "    input[%d:0]    l%d ;\n", gf_m-1, i) ;
           fprintf(stream, "\n") ;
           for (i=0; i<chk_sym; i++)
           fprintf(stream, "    output[%d:0]    lroot%d ;\n", gf_m-1, i) ;
           fprintf(stream, "    output[%d:0]    n_roots ;\n", gf_m-1) ;
           fprintf(stream, "    output      search_done ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "`include \"../gf_arith.v\"\n") ;		   
           fprintf(stream, "\n") ;
           fprintf(stream, "// Locals\n") ;
           fprintf(stream, "\n") ;
           for (i=0; i<chk_sym; i++)
               fprintf(stream, "    reg[%d:0]    root_%d ;\n", gf_m-1, i) ;
           fprintf(stream, "\n") ;
           for (i=0; i<2*chk_sym; i++)
               fprintf(stream, "    reg[%d:0]    r%d ;\n", gf_m-1, i) ;
           fprintf(stream, "\n") ;
           for (i=0; i<chk_sym; i++)
               fprintf(stream, "    assign lroot%d = root_%d ;\n", i, i) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "//  Search Control SM\n") ;
           fprintf(stream, "    reg[1:0]    Src_SM ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    parameter   Src_s0 = 0 ;\n") ;
           fprintf(stream, "    parameter   Src_s1 = 1 ;\n") ;
           fprintf(stream, "    parameter   Src_s2 = 2 ;\n") ;
           fprintf(stream, "    parameter   Src_s3 = 3 ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    parameter   GF_FIELD_CNT = %d ;\n", gf_feild_size-1) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    wire        load_coef ;\n") ;
           fprintf(stream, "    reg         enb_search ;\n") ;
           fprintf(stream, "    reg         done ;\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "\n") ;    
           fprintf(stream, "    reg[%d:0]    root_cnt ;\n", gf_m-1) ;
           fprintf(stream, "    reg[%d:0]    root_sel ;\n", vlen) ;
           fprintf(stream, "    reg[%d:0]    Num_Roots ;\n", gf_m-1) ;
           fprintf(stream, "    assign      n_roots = Num_Roots ;\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "    wire[%d:0]    root_calc ;\n", gf_m-1) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    assign      root_calc = r0 ^ \n") ;
           for (i=1; i<2*chk_sym-1; i++)
               fprintf(stream, "                            r%d ^ \n", i) ;
               fprintf(stream, "                            r%d ;\n", i) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "always @(negedge rst_n or posedge clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n == 1'b0)\n") ;
           fprintf(stream, "        begin\n") ;
           fprintf(stream, "            enb_search <= 1'b0 ;\n") ;
           fprintf(stream, "            done <= 1'b0 ;\n") ;
           fprintf(stream, "            Src_SM <= Src_s0  ;\n") ;
           fprintf(stream, "        end\n") ;
           fprintf(stream, "     else\n") ;
           fprintf(stream, "         case (Src_SM)\n") ;
           fprintf(stream, "             Src_s0:\n") ;
           fprintf(stream, "                 begin\n") ;
           fprintf(stream, "                     if (start_search == 1'b1)\n") ;
           fprintf(stream, "                         Src_SM <= Src_s1  ;\n") ;
           fprintf(stream, "                     else \n") ;
           fprintf(stream, "                         Src_SM <= Src_s0  ;\n") ;
           fprintf(stream, "                 end\n") ;
           fprintf(stream, "             Src_s1:\n") ;
           fprintf(stream, "                 begin\n") ;
           fprintf(stream, "                     enb_search <= 1'b1 ;\n") ;
           fprintf(stream, "                     Src_SM <= Src_s2  ;\n") ;
           fprintf(stream, "                 end\n") ;
           fprintf(stream, "             Src_s2:\n") ;
           fprintf(stream, "                 begin\n") ;
           fprintf(stream, "                     if (root_cnt < GF_FIELD_CNT)\n") ;
           fprintf(stream, "                         begin\n") ;
           fprintf(stream, "                             //root_cnt <= root_cnt + 1 ;\n") ;
           fprintf(stream, "                             Src_SM <= Src_s2  ;\n") ;
           fprintf(stream, "                         end\n") ; 
           fprintf(stream, "                     else\n") ;
           fprintf(stream, "                         begin\n") ;
           fprintf(stream, "                             enb_search <= 1'b0 ;\n") ;
           fprintf(stream, "                             done <= 1'b1 ;\n") ;
           fprintf(stream, "                             Src_SM <= Src_s3  ;\n") ;
           fprintf(stream, "                         end\n") ; 
           fprintf(stream, "                 end\n") ;
           fprintf(stream, "             Src_s3:\n") ;
           fprintf(stream, "                 begin\n") ;
           fprintf(stream, "                     Src_SM <= Src_s0  ;\n") ;
           fprintf(stream, "                     done <= 1'b0 ;\n") ;
           fprintf(stream, "                 end\n") ;
           fprintf(stream, "         endcase\n") ;
           fprintf(stream, "end\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    assign load_coef = (Src_SM == Src_s0 && start_search == 1'b1) ? 1'b1 : 1'b0 ;\n") ;
           fprintf(stream, "    assign search_done = done ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "always @(negedge rst_n or posedge clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n == 1'b0)\n") ;
           fprintf(stream, "        begin\n") ;
           for (i=0; i<2*chk_sym; i++)
               fprintf(stream, "            r%d <= %d'b0 ;\n", i, gf_m) ;
           for (i=0; i<chk_sym; i++)
               fprintf(stream, "            root_%d <= %d'b0 ;\n", i, gf_m) ;
           fprintf(stream, "            root_cnt <= %d'b0 ;\n", vlen) ;
           fprintf(stream, "            root_sel <= %d'b0 ;\n", vlen) ;
           fprintf(stream, "            Num_Roots <= %d'b0 ;\n", vlen) ;
           fprintf(stream, "        end\n") ;
           fprintf(stream, "    else if (load_coef == 1'b1 )\n") ;
           fprintf(stream, "        begin\n") ;
           fprintf(stream, "   // l0 Must be 1 for search to function correctly\n") ;
           for (i=0; i<2*chk_sym; i++)
               fprintf(stream, "            r%d <= l%d ;\n", i, i) ;
           fprintf(stream, "            root_sel <= 0 ;\n") ;
           fprintf(stream, "            Num_Roots <= 0 ;\n") ;
           fprintf(stream, "        end\n") ;
           fprintf(stream, "    else if (enb_search == 1'b1)\n") ; 
           fprintf(stream, "        begin\n") ;
           fprintf(stream, "           if (root_calc == 0)\n") ;
           fprintf(stream, "                begin\n") ;
           fprintf(stream, "                    case (root_sel)\n") ;
           for (i=0; i<chk_sym; i++) {
               fprintf(stream, "                        %d : begin\n", i) ;
               fprintf(stream, "                               root_%d <= root_cnt ;\n", i) ;
               fprintf(stream, "                            end\n") ;
            }
           fprintf(stream, "                    endcase\n") ; 
           fprintf(stream, "                    root_sel <= root_sel + 1 ;\n") ;
           fprintf(stream, "                    Num_Roots <= Num_Roots + 1 ;\n") ;
           fprintf(stream, "                end\n") ;
           for (i=1; i<=chk_sym; i++)
           fprintf(stream, "            r%d <= gfmul_%d(r%d) ;\n", i, i, i) ;
           fprintf(stream, "            root_cnt <= root_cnt + 1 ;\n") ;
           fprintf(stream, "        end\n") ;
           fprintf(stream, "end\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "endmodule \n") ;
           break;

       case VHDL :
           fprintf(stream, "------------------------------------------------------------------------\n") ;
           fprintf(stream, "--                  Generated by RSTK                                 --\n") ;  
           fprintf(stream, "--                      rs_chien.vhd                                  --\n") ;  
           fprintf(stream, "--                                                                    --\n") ;  
           fprintf(stream, "------------------------------------------------------------------------\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "LIBRARY IEEE ;\n") ;
           fprintf(stream, "use IEEE.std_logic_1164.all ;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "library IEEE ;\n") ;
           fprintf(stream, "use     IEEE.std_logic_unsigned.all;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "library IEEE ;\n") ;
           fprintf(stream, "use     IEEE.NUMERIC_STD.all;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "library gf_arith ;\n") ;
           fprintf(stream, "use     gf_arith.gf_arith.all;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "entity chien_srch is port( rst_n : IN std_logic ;\n") ;
           fprintf(stream, "                           clk   : IN std_logic ;\n") ;
           fprintf(stream, "                           start_search : IN std_logic ;\n") ;
           for (i=0; i<2*chk_sym; i++)
               fprintf(stream, "                           l%d : IN std_logic_vector(%d downto 0);\n", i, gf_m-1) ;
           for (i=0; i<chk_sym; i++)
               fprintf(stream, "                           lroot%d : OUT std_logic_vector(%d downto 0);\n", i, gf_m-1) ; 
           fprintf(stream, "                           n_roots : OUT std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "                           search_done  : OUT std_logic\n") ; 
           fprintf(stream, "                       ) ;\n") ;
           fprintf(stream, "end chien_srch ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "architecture rtl of chien_srch is \n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "-- Locals\n") ;
           fprintf(stream, "--\n") ;
           for (i=0; i<chk_sym; i++)
               fprintf(stream, "    signal     root_%d : std_logic_vector(%d downto 0) ;\n", i, gf_m-1) ;
           fprintf(stream, "\n") ;
           for (i=0; i<2*chk_sym; i++)
               fprintf(stream, "    signal     r%d : std_logic_vector(%d downto 0);\n", i, gf_m-1) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "--  Search Control SM\n") ;
           fprintf(stream, "    type chien_search_states is (\n") ;
           fprintf(stream, "        csrch_s0,\n") ;
           fprintf(stream, "        csrch_s1,\n") ;
           fprintf(stream, "        csrch_s2,\n") ;
           fprintf(stream, "        csrch_s3\n") ;
           fprintf(stream, "    ) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    signal CSHRC_SM : chien_search_states ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    constant GF_FIELD_CNT : integer := %d ;\n", gf_feild_size-1 ) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    signal load_coef  : std_logic ;\n") ;
           fprintf(stream, "    signal enb_search : std_logic;\n") ;
           fprintf(stream, "    signal done : std_logic;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "\n") ;    
           fprintf(stream, "    signal    root_cnt  : std_logic_vector(%d downto 0) ;\n", gf_m-1) ;
           fprintf(stream, "    signal    root_sel  : std_logic_vector(%d downto 0) ;\n", vlen) ;
           fprintf(stream, "    signal    Num_Roots : std_logic_vector(%d downto 0) ;\n", gf_m-1) ;
           fprintf(stream, "    signal    root_calc : std_logic_vector(%d downto 0) ;\n", gf_m-1) ;
           fprintf(stream, "\n") ; 
           fprintf(stream, "begin\n") ;
           fprintf(stream, "\n") ; 
           fprintf(stream, "    n_roots <= Num_Roots ;\n") ;
           fprintf(stream, "\n") ; 
           fprintf(stream, "    root_calc <= r0 xor \n") ;
           for (i=1; i<chk_sym-1; i++)
               fprintf(stream, "                 r%d xor\n", i) ;
               fprintf(stream, "                 r%d ;\n", i) ;
           fprintf(stream, "\n") ; 
           for (i=0; i<chk_sym; i++)
               fprintf(stream, "    lroot%d <= root_%d ;\n", i, i) ;
           fprintf(stream, "\n") ; 
           fprintf(stream, "CHIEN_SEARCH_SM:process(rst_n, clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n = '0') then\n") ;
           fprintf(stream, "            enb_search <= '0' ;\n") ;
           fprintf(stream, "            done <= '0' ;\n") ;
           fprintf(stream, "            CSHRC_SM <= csrch_s0  ;\n") ;
           fprintf(stream, "    elsif (clk'event and clk = '1') then\n") ;
           fprintf(stream, "        case CSHRC_SM is\n") ;
           fprintf(stream, "             when  csrch_s0 =>\n") ;
           fprintf(stream, "                     if (start_search = '1') then\n") ;
           fprintf(stream, "                         CSHRC_SM <= csrch_s1  ;\n") ;
           fprintf(stream, "                     else \n") ;
           fprintf(stream, "                         CSHRC_SM <= csrch_s0  ;\n") ;
           fprintf(stream, "                     end if ;\n") ;
           fprintf(stream, "\n") ; 
           fprintf(stream, "             when csrch_s1 =>\n") ;
           fprintf(stream, "                     enb_search <= '1' ;\n") ;
           fprintf(stream, "                     CSHRC_SM <= csrch_s2  ;\n") ;
           fprintf(stream, "\n") ; 
           fprintf(stream, "             when csrch_s2 =>\n") ;
           fprintf(stream, "                     if (root_cnt < GF_FIELD_CNT) then\n") ;
           fprintf(stream, "                             --root_cnt <= root_cnt + 1 ;\n") ;
           fprintf(stream, "                             CSHRC_SM <= csrch_s2  ;\n") ;
           fprintf(stream, "                     else\n") ;
           fprintf(stream, "                             enb_search <= '0' ;\n") ;
           fprintf(stream, "                             done <= '1' ;\n") ;
           fprintf(stream, "                             CSHRC_SM <= csrch_s3  ;\n") ;
           fprintf(stream, "                     end if ;\n") ;
           fprintf(stream, "\n") ; 
           fprintf(stream, "             when  csrch_s3 =>\n") ;
           fprintf(stream, "                     CSHRC_SM <= csrch_s0  ;\n") ;
           fprintf(stream, "                     done <= '0' ;\n") ;
           fprintf(stream, "\n") ; 
           fprintf(stream, "             when OTHERS =>\n") ;
           fprintf(stream, "                     CSHRC_SM <= csrch_s0  ;\n") ;
           fprintf(stream, "           end case ;\n") ;
           fprintf(stream, "      end if ;\n") ;
           fprintf(stream, "end process ;\n") ;
           fprintf(stream, "\n") ; 
           fprintf(stream, "    load_coef <= '1' when(CSHRC_SM = csrch_s0 and start_search ='1') else '0' ;\n") ;
           fprintf(stream, "    search_done <= done ;\n") ;
           fprintf(stream, "\n") ; 
           fprintf(stream, "ROOT_FIND:process(rst_n, clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n = '0') then\n") ;
           for (i=0; i<2*chk_sym; i++)
               fprintf(stream, "            r%d <= (OTHERS=> '0') ;\n", i) ;
           for (i=0; i<chk_sym; i++)
               fprintf(stream, "            root_%d <= (OTHERS=> '0') ;\n", i) ;
           fprintf(stream, "            root_cnt <= (OTHERS=> '0') ;\n") ;
           fprintf(stream, "            root_sel <= (OTHERS=> '0') ;\n") ;
           fprintf(stream, "            Num_Roots <= (OTHERS=> '0') ;\n") ;
           fprintf(stream, "    elsif (clk'event and clk = '1') then\n") ;
           fprintf(stream, "        if (load_coef = '1' ) then\n") ;
           fprintf(stream, "-- l0 Must be 1 for search to function correctly\n") ;
           for (i=0; i<2*chk_sym; i++)
               fprintf(stream, "            r%d <= l%d ;\n", i, i) ;
           fprintf(stream, "            root_sel <= (OTHERS=> '0') ;\n") ;
           fprintf(stream, "            Num_Roots <= (OTHERS=> '0') ;\n") ;
           fprintf(stream, "        elsif (enb_search = '1') then\n") ;
           fprintf(stream, "             if (root_calc = 0) then\n") ;
           fprintf(stream, "                 case root_sel is\n") ;
           for (i=0; i<chk_sym; i++) {
           fprintf(stream, "                     when \"") ;
               conv2_bvec_str(i, vlen, b_vector) ;
               pr_vec_str(stream, vlen, b_vector) ;
               fprintf(stream, "\" =>\n") ;
               fprintf(stream, "                          root_%d <= root_cnt ;\n", i) ;
               fprintf(stream, "\n") ; 
           }
           fprintf(stream, "                     when OTHERS =>\n") ; 
           fprintf(stream, "\n") ;                           
           fprintf(stream, "                  end case ;\n") ;
           fprintf(stream, "                    root_sel <= root_sel + 1 ;\n") ;
           fprintf(stream, "                    Num_Roots <= Num_Roots + 1 ;\n") ;
           fprintf(stream, "               end if ;\n") ;
           fprintf(stream, "\n") ; 
           for (i=1; i<=chk_sym; i++)
               fprintf(stream, "            r%d <= gfmul_%d(r%d) ;\n", i, i, i) ;
           fprintf(stream, "            root_cnt <= root_cnt + 1 ;\n") ;
           fprintf(stream, "        end if ;\n") ;
           fprintf(stream, "    end if;\n") ;
           fprintf(stream, "end process ;\n") ;
           fprintf(stream, "\n") ; 
           fprintf(stream, "END ; \n") ;
           break ;
           default :
                break ;
    }
}
 
