/*******************************************************************************/
/*                                                                             */
/*            gen_rs_enc.c                                                     */
/*                                                                             */
/*    Generate RS Encoder     from RS generator                                */
/*    polynomial                                                               */
/*                                                                             */
/*                                                                             */
/*                                                                             */
/*                                                                             */
/*                                                                             */
/*    Copyright (C) 2003  John S. Willy (jswilly@pacbell.net)                  */ 
/*                                                                             */
/*    This program is free software; you can redistribute it and/or modify     */
/*    it under the terms of the GNU General Public License as published by     */
/*    the Free Software Foundation; either version 2 of the License, or        */
/*    (at your option) any later version.                                      */
/*                                                                             */
/*    This program is distributed in the hope that it will be useful,          */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*    GNU General Public License for more details.                             */
/*                                                                             */
/*    You should have received a copy of the GNU General Public License        */
/*    along with this program; if not, write to the Free Software              */
/*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA*/
/*******************************************************************************/
 
#include <stdio.h>

#define VHDL     0 
#define VLOG     1 
#define C        2 

extern unsigned char log_poly_table[] ; 

/* locals  */

void gen_rs_encoder(FILE *stream, unsigned int *p_wsp, int gf_2n, int deg, int hdl_type)
{
   int coeff, a, i, k ;

   switch (hdl_type) {
       case C:
           fprintf(stream, "//--------------------------------------------------------------------//\n") ;
           fprintf(stream, "//                  Generated by RSTK                                 //\n") ;  
           fprintf(stream, "//                      rs_encode.c                                   //\n") ;  
           fprintf(stream, "//                                                                    //\n") ;  
           fprintf(stream, "//--------------------------------------------------------------------//\n") ;
           fprintf(stream, "#include <stdio.h>\n") ;
           fprintf(stream, "#include \"proto_types.h\"\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "extern unsigned char msg_in[] ;\n") ;
           fprintf(stream, "extern unsigned char enc_msg_out[] ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "void rs_encode(int n, int k, char dbg)\n") ;
           fprintf(stream, "{\n") ;
           fprintf(stream, "unsigned char\n") ;
		   for(i=0; i < deg; i++) 
			   fprintf(stream, "       g%d,\n", i) ;
		   fprintf(stream, "       g%d,\n", i) ;

		   for(i=0; i < deg-1; i++) 
			   fprintf(stream, "       g%d_r,\n", i) ;
			   fprintf(stream, "       g%d_r ;\n", i) ;
           fprintf(stream, "unsigned char feedback ;\n") ;
           fprintf(stream, "char d_in ;\n") ;
           fprintf(stream, "int i ;\n") ;
           fprintf(stream, "\n") ;
             for(i=0; i <= deg; i++) 
               fprintf(stream, "    g%d = 0 ;\n", i) ;

           fprintf(stream, "\n") ;
             for(i=0; i < deg; i++)
	       fprintf(stream, "    g%d_r = 0 ;\n", i) ;

           fprintf(stream, "\n") ;
           fprintf(stream, "    feedback = 0 ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "  for (i=0; i<=k-1; i++) {\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    d_in = msg_in[i] ;\n") ;
           fprintf(stream, "    feedback = g%d_r ^ d_in ;\n", deg-1) ;
           fprintf(stream, "\n") ;
		   for(i=0; i< deg; i++) {
			   coeff = p_wsp[i] ;
			   a = (unsigned int)log_poly_table[coeff] ;
               fprintf(stream, "    g%d = gfmul_%d(feedback) ;\n", i, a) ;
		   }

           fprintf(stream, "\n") ;
           fprintf(stream, "\n") ;
		   for(i=0; i < deg-1; i++) {
			   k = deg-1-i ;
               fprintf(stream, "    g%d_r = g%d_r ^ g%d ;\n", k, k-1,k) ;
		   }
           fprintf(stream, "    g0_r = g0 ;\n") ;

           fprintf(stream, "\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    enc_msg_out[i] = d_in   ;\n") ;
           fprintf(stream, "  }\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "     feedback = 0 ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "  for (i=k; i<= n; i++) {\n") ;
           fprintf(stream, "     enc_msg_out[i] = g%d_r   ;\n", deg-1) ;
		   for(i=0; i < deg-1; i++) {
			   k = deg-1-i ;
               fprintf(stream, "     g%d_r = g%d_r ;\n", k, k-1) ;
		   }
           fprintf(stream, "     g0_r = feedback ;\n") ;

           fprintf(stream, "  }\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "}\n") ;
		   break;

       case VLOG : 
           fprintf(stream, "//--------------------------------------------------------------------//\n") ;
           fprintf(stream, "//                  Generated by RSTK                                 //\n") ;  
           fprintf(stream, "//                      rs_encode.v                                   //\n") ;  
           fprintf(stream, "//                                                                    //\n") ;  
           fprintf(stream, "//--------------------------------------------------------------------//\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "`timescale 1ns/100ps\n");

           fprintf(stream, "\n") ;
           fprintf(stream, "module rs_encode (\n") ;
           fprintf(stream, "                   rst_n,\n") ;
           fprintf(stream, "                   clk,\n") ;
           fprintf(stream, "                   start,\n") ;
           fprintf(stream, "                   d_in,\n") ;
           fprintf(stream, "                   enc_dout,\n") ;
           fprintf(stream, "                   enc_dov,\n") ;
           fprintf(stream, "                   enc_synv,\n") ;
           fprintf(stream, "                   enc_done  \n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "	    ) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    input        rst_n ;\n") ;
           fprintf(stream, "    input        clk ;\n") ;
           fprintf(stream, "    input        start ;\n") ;
           fprintf(stream, "    input[%d:0]  d_in ;\n", gf_2n-1) ;
           fprintf(stream, "    output[%d:0] enc_dout ;\n", gf_2n-1) ;
           fprintf(stream, "    output       enc_dov ;\n") ;
           fprintf(stream, "    output       enc_synv ;\n") ;
           fprintf(stream, "    output       enc_done ;\n") ; 
           fprintf(stream, "\n") ; 
           fprintf(stream, "`include \"../gf_arith.v\"\n") ;
           fprintf(stream, "// Locals\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "     parameter N = 255 ;  // N \n") ;
           fprintf(stream, "     parameter K = 249 ;  // K \n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    reg[%d:0]    din_r ;\n", gf_2n-1) ;
		   for(i=0; i < deg; i++)
               fprintf(stream, "    wire[%d:0]  	g%d ;\n", gf_2n-1, i) ;
           fprintf(stream, "\n") ;
		   for(i=0; i < deg; i++)
               fprintf(stream, "    reg[%d:0]  	g%d_r ;\n", gf_2n-1, i) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    wire[%d:0]  	g%d ;\n", gf_2n-1, deg) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    wire[%d:0]    feedback ;\n", gf_2n-1) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    reg[%d:0]  	xfer_cnt ;\n", gf_2n-1) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    reg    init ;\n") ;
           fprintf(stream, "    reg    done ;\n") ;
           fprintf(stream, "    reg    enb_enc ;\n") ;
           fprintf(stream, "    reg    enc_feedback ;\n") ;
           fprintf(stream, "    reg    syn_valid ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    parameter  rs_enc_s0 = 0,\n") ;
           fprintf(stream, "               rs_enc_s1 = 1,\n") ;
           fprintf(stream, "               rs_enc_s2 = 2,\n") ;
           fprintf(stream, "               rs_enc_s3 = 3 ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    reg[1:0]   RSENC_SM ;\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "//\n") ;
		   for(i=0; i< deg; i++) {
			   coeff = p_wsp[i] ;
			   a = (unsigned int)log_poly_table[coeff] ;
               fprintf(stream, "     assign #1 g%d = gfmul_%d(feedback) ;\n", i, a) ;
		   }
           fprintf(stream, "     assign #1 g%d = g%d_r;\n", deg, deg-1) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "     assign #1 feedback = (enc_feedback == 1'b1) ? (g%d ^ din_r) : %d'b0 ;\n", deg, gf_2n) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "     assign #1 enc_dout = (enc_feedback == 1'b1) ? din_r : g%d ;\n", deg) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "always @(negedge rst_n or posedge clk)\n") ;
           fprintf(stream, "begin:dpipe\n");
           fprintf(stream, "    if (rst_n == 1'b0) \n") ;
           fprintf(stream, "	    din_r <= 0 ;\n") ;
           fprintf(stream, "	else\n") ;
           fprintf(stream, "        din_r <= d_in ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "end \n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "always @(negedge rst_n or posedge clk)\n") ;
           fprintf(stream, "begin:rs_encoder\n") ;
           fprintf(stream, "    if (rst_n == 1'b0) begin\n") ;
		   for(i=0; i < deg; i++)
               fprintf(stream, "	    g%d_r <= 0 ;\n", i) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "		end\n") ;
           fprintf(stream, "    else if (start == 1'b1)\n") ;
           fprintf(stream, "	    begin\n") ;
		   for(i=0; i < deg; i++)
               fprintf(stream, "	        g%d_r <= 0 ;\n", i) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "	    end\n") ;
           fprintf(stream, "    else if (enb_enc == 1'b1)\n") ;
           fprintf(stream, "	    begin\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "	        g0_r <= g0 ;\n") ;
		   for(i=0; i < deg-1; i++)
               fprintf(stream, "	        g%d_r <= g%d_r ^ g%d ;\n", i+1, i, i+1) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "		end \n") ;
           fprintf(stream, "end\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "always @(negedge rst_n or posedge clk)\n") ;
           fprintf(stream, "begin:rs_encode_sm\n") ;
           fprintf(stream, "    if (rst_n == 1'b0)\n") ;
           fprintf(stream, "        begin\n") ;
           fprintf(stream, "            RSENC_SM <= rs_enc_s0 ;\n") ;
           fprintf(stream, "            xfer_cnt <= 0 ;\n") ;
           fprintf(stream, "            enc_feedback <= 1'b0 ;\n") ;
           fprintf(stream, "            enb_enc <= 1'b0 ;\n") ;
           fprintf(stream, "            init <= 1'b0 ;\n") ;
           fprintf(stream, "        end\n") ;
           fprintf(stream, "    else\n") ;
           fprintf(stream, "        case (RSENC_SM)\n") ; 
           fprintf(stream, "            rs_enc_s0: \n") ;
           fprintf(stream, "                if (start == 1'b1)\n") ;
           fprintf(stream, "                    begin\n") ;
           fprintf(stream, "                        RSENC_SM <= rs_enc_s1 ;\n") ;
           fprintf(stream, "                        init <= 1'b1 ;\n") ;
           fprintf(stream, "                        enb_enc <= 1'b1 ;\n") ;
           fprintf(stream, "                        enc_feedback <= 1'b1 ;\n") ;
           fprintf(stream, "                        end\n") ;
           fprintf(stream, "                    else\n") ;
           fprintf(stream, "                        RSENC_SM <= rs_enc_s0 ;\n") ;
           fprintf(stream, "\n") ; 
           fprintf(stream, "            rs_enc_s1:\n") ; 
           fprintf(stream, "                if (xfer_cnt < K-1) \n") ;
           fprintf(stream, "                    begin\n") ;
           fprintf(stream, "                        RSENC_SM <= rs_enc_s1 ;\n") ;
           fprintf(stream, "                        xfer_cnt <= xfer_cnt + 1 ;\n") ;
           fprintf(stream, "                        init <= 1'b0 ;\n") ;
           fprintf(stream, "                    end\n") ;
           fprintf(stream, "                else\n") ;
           fprintf(stream, "                    begin\n") ;
           fprintf(stream, "                        RSENC_SM <= rs_enc_s2 ;\n") ;
           fprintf(stream, "                        xfer_cnt <= xfer_cnt + 1 ;\n") ;
           fprintf(stream, "                        enc_feedback <= 1'b0 ;\n") ;
           fprintf(stream, "                        init <= 1'b0 ;\n") ;
           fprintf(stream, "                    end \n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "            rs_enc_s2:\n") ;
           fprintf(stream, "                if (xfer_cnt < N)\n") ;
           fprintf(stream, "                    begin\n") ;   
           fprintf(stream, "                        RSENC_SM <= rs_enc_s2 ;\n") ;
           fprintf(stream, "                        xfer_cnt <= xfer_cnt + 1 ;\n") ;
           fprintf(stream, "                    end\n") ;
           fprintf(stream, "                else\n") ;
           fprintf(stream, "                    begin\n") ;
           fprintf(stream, "                        RSENC_SM <= rs_enc_s0 ;\n") ;
           fprintf(stream, "                        enb_enc <= 1'b0 ;\n") ;
           fprintf(stream, "                        xfer_cnt <= 0 ;\n") ;
           fprintf(stream, "                    end\n") ;  
           fprintf(stream, "\n") ;
           fprintf(stream, "            default:\n") ;
           fprintf(stream, "                RSENC_SM <= rs_enc_s0 ;\n") ;
           fprintf(stream, "         endcase\n") ;  
           fprintf(stream, "\n") ;
           fprintf(stream, "end\n") ;  
           fprintf(stream, "\n") ;
           fprintf(stream, "    assign #1 enc_dov = enc_feedback ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "always @(negedge rst_n or posedge clk)\n") ;
           fprintf(stream, "begin:enc_syn_valid\n") ;
           fprintf(stream, "    if (rst_n == 1'b0)\n") ;
           fprintf(stream, "        syn_valid <= 1'b0 ;\n") ;
           fprintf(stream, "    else if (RSENC_SM == rs_enc_s1 && xfer_cnt == K-1)\n") ;
           fprintf(stream, "             syn_valid <= 1'b1 ;\n") ;
           fprintf(stream, "    else if (done == 1'b1)\n") ;
           fprintf(stream, "	         syn_valid <= 1'b0 ;	\n") ;	
           fprintf(stream, "\n") ;
           fprintf(stream, "end\n") ;  
           fprintf(stream, "\n") ;
           fprintf(stream, "    assign #1 enc_synv = syn_valid ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "always @(negedge rst_n or posedge clk)\n") ;
           fprintf(stream, "begin:encoder_done\n") ;
           fprintf(stream, "    if (rst_n == 1'b0)\n") ;
           fprintf(stream, "        done <= 1'b0 ;\n") ;
           fprintf(stream, "	else if (RSENC_SM == rs_enc_s2 && xfer_cnt == N-2)\n") ;
           fprintf(stream, "             done <= 1'b1 ;\n") ;
           fprintf(stream, "         else\n") ;
           fprintf(stream, "             done <= 1'b0 ;\n") ;		
           fprintf(stream, "\n") ;
           fprintf(stream, "end\n") ; 
           fprintf(stream, "\n") ;
           fprintf(stream, "    assign #1 enc_done = done ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "endmodule\n") ;
		   break;

       case VHDL :
           fprintf(stream, "------------------------------------------------------------------------\n") ;
           fprintf(stream, "--                  Generated by RSTK                                 --\n") ;  
           fprintf(stream, "--                      rs_encode.vhd                                 --\n") ;  
           fprintf(stream, "--                                                                    --\n") ;  
           fprintf(stream, "------------------------------------------------------------------------\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "LIBRARY IEEE ;\n") ;
           fprintf(stream, "use IEEE.std_logic_1164.all ;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "library IEEE ;\n") ;
           fprintf(stream, "use 	IEEE.std_logic_unsigned.all;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "library IEEE ;\n") ;
           fprintf(stream, "use 	IEEE.NUMERIC_STD.all;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "library gf_arith ;\n") ;
           fprintf(stream, "use 	gf_arith.gf_arith.all;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "entity rs_encode is\n") ;
           fprintf(stream, "  generic ( N  : integer := %d ;\n", gf_2n-1) ;
           fprintf(stream, "            K  : integer := %d\n", gf_2n-1) ;
           fprintf(stream, "		   ); \n") ;		   
           fprintf(stream, "    port( rst_n        : IN std_logic ;\n") ;
           fprintf(stream, "          clk          : IN std_logic ;\n") ;
           fprintf(stream, "          start        : IN std_logic ;\n") ;
           fprintf(stream, "          d_in         : IN std_logic_vector(%d downto 0) ;\n", gf_2n-1) ;
           fprintf(stream, "          enc_dout     : OUT std_logic_vector(%d downto 0);\n", gf_2n-1) ;
           fprintf(stream, "          enc_dov      : OUT std_logic ;\n" ) ;
           fprintf(stream, "          enc_synv     : OUT std_logic ;\n" ) ;
           fprintf(stream, "          enc_done     : OUT std_logic\n" ) ; 
           fprintf(stream, "   ) ;\n") ;
           fprintf(stream, "end rs_encode ;\n") ;
           fprintf(stream, "\n\n") ;
           fprintf(stream, "architecture rtl of rs_encode is\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "\n\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    signal	din_r : std_logic_vector(%d downto 0);\n", gf_2n-1) ; 
           fprintf(stream, "\n") ;
		   for(i=0; i < deg; i++) 
			   fprintf(stream, "    signal	g%d : std_logic_vector(%d downto 0);\n", i, gf_2n-1) ;
        
           fprintf(stream, "\n") ;
           fprintf(stream, "\n") ;
		   for(i=0; i < deg; i++) 
			   fprintf(stream, "    signal	g%d_r : std_logic_vector(%d downto 0);\n", i, gf_2n-1) ;
		   fprintf(stream, "    signal	g%d   : std_logic_vector(%d downto 0)  ;\n", deg, gf_2n-1) ;
 
           fprintf(stream, "\n") ;
 
           fprintf(stream, "\n") ;
           fprintf(stream, "    signal	feedback : std_logic_vector(%d downto 0)  ;\n", gf_2n-1) ; 
           fprintf(stream, "\n") ;
           fprintf(stream, "    signal	xfer_cnt: std_logic_vector(%d downto 0) ;\n", gf_2n-1) ; 
           fprintf(stream, "\n") ;
           fprintf(stream, "    signal  init         : std_logic ;\n") ; 
           fprintf(stream, "    signal  done         : std_logic ;\n") ; 
           fprintf(stream, "    signal  enb_enc      : std_logic ;\n") ; 
           fprintf(stream, "    signal  enc_feedback : std_logic ;\n") ; 
           fprintf(stream, "    signal  syn_valid    : std_logic ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    type    rs_encode_state is (\n") ; 
           fprintf(stream, "                rs_enc_s0,\n") ; 
           fprintf(stream, "                rs_enc_s1,\n") ; 
           fprintf(stream, "                rs_enc_s2,\n") ; 
           fprintf(stream, "                rs_enc_s3\n") ; 
           fprintf(stream, "            ) ;\n") ; 
           fprintf(stream, "\n") ;
           fprintf(stream, "    signal    RSENC_SM  : rs_encode_state ;\n") ; 
           fprintf(stream, "--\n") ; 
           fprintf(stream, "--\n") ; 
           fprintf(stream, "--\n") ; 
           fprintf(stream, "begin\n") ; 
           fprintf(stream, "\n") ;
		   for(i=0; i< deg; i++) {
			   coeff = p_wsp[i] ;
			   a = (unsigned int)log_poly_table[coeff] ;
               fprintf(stream, "     g%d <= gfmul_%d(feedback) after 1 ns ;\n", i, a) ;
		   }
           fprintf(stream, "     g%d <= g%d_r  after 1 ns ;\n", deg, deg-1) ;
 
           fprintf(stream, "\n") ;
           fprintf(stream, "     feedback <= (g%d xor din_r) when (enc_feedback = '1') else (OTHERS=>'0') after 1 ns ;\n", deg) ; 
 
           fprintf(stream, "\n") ;
           fprintf(stream, "     enc_dout <= din_r  when (enc_feedback = '1') else g%d after 1 ns ;\n", deg) ; 
           fprintf(stream, "\n") ;
           fprintf(stream, "dpipe:process (rst_n, clk)\n") ; 
           fprintf(stream, "begin\n") ; 
           fprintf(stream, "    if (rst_n = '0') then \n") ; 
           fprintf(stream, "        din_r <= (OTHERS => '0') ;\n") ; 
           fprintf(stream, "    else if (clk'event and clk = '1') then\n") ; 
           fprintf(stream, "         din_r <= d_in after 1 ns ;\n") ; 
           fprintf(stream, "         end if ;\n") ; 
           fprintf(stream, "    end if ;\n") ; 
           fprintf(stream, "end process ;\n") ; 
           fprintf(stream, "\n") ;
           fprintf(stream, "rs_encoder:process (rst_n, clk)\n") ; 
           fprintf(stream, "begin\n") ; 
           fprintf(stream, "    if (rst_n = '0') then\n") ; 
		   for(i=0; i< deg; i++) 
               fprintf(stream, "	    g%d_r <= (OTHERS => '0') ;\n", i) ;
 
           fprintf(stream, "    elsif (clk'event and clk = '1') then\n") ; 
           fprintf(stream, "        if (start = '1') then\n") ; 
		   for(i=0; i< deg; i++) 
               fprintf(stream, "	        g%d_r <= (OTHERS => '0') ;\n", i) ; 
 
           fprintf(stream, "        elsif (enb_enc = '1') then\n") ;
           fprintf(stream, "	        g0_r <= g0 ;\n") ;
		   for(i=0; i< deg-1; i++) 
               fprintf(stream, "	        g%d_r <= g%d_r xor g%d ;\n", i+1, i, i+1) ;			   
 
           fprintf(stream, "		end if ;\n") ; 
           fprintf(stream, "    end if ;\n") ; 
           fprintf(stream, "end process ;\n") ; 
           fprintf(stream, "\n") ;
           fprintf(stream, "rs_encode_sm:process (rst_n, clk)\n") ; 
           fprintf(stream, "begin\n") ; 
           fprintf(stream, "    if (rst_n = '0') then\n") ; 
           fprintf(stream, "        RSENC_SM <= rs_enc_s0 ;\n") ; 
           fprintf(stream, "        xfer_cnt <= (OTHERS=>'0');\n") ; 
           fprintf(stream, "        enc_feedback <= '0' ;\n") ; 
           fprintf(stream, "        enb_enc <= '0' ;\n") ; 
           fprintf(stream, "        init <= '0' ;\n") ; 
           fprintf(stream, "    else\n") ; 
           fprintf(stream, "        if (clk'event and clk ='1') then\n") ; 
           fprintf(stream, "            case RSENC_SM is\n") ; 
           fprintf(stream, "                when rs_enc_s0 => \n") ; 
           fprintf(stream, "                    if (start = '1') then\n") ; 
           fprintf(stream, "                        RSENC_SM <= rs_enc_s1 ;\n") ; 
           fprintf(stream, "                        init <= '1' ;\n") ; 
           fprintf(stream, "                        enb_enc <= '1' ;\n") ; 
           fprintf(stream, "                        enc_feedback <= '1' ;\n") ; 
           fprintf(stream, "                    else\n") ; 
           fprintf(stream, "                        RSENC_SM <= rs_enc_s0 ;\n") ; 
           fprintf(stream, "                    end if ;\n") ; 
           fprintf(stream, "\n") ;
           fprintf(stream, "                when rs_enc_s1 =>\n") ;  
           fprintf(stream, "                    if (xfer_cnt < K-1) then\n") ; 
           fprintf(stream, "                        RSENC_SM <= rs_enc_s1 ;\n") ; 
           fprintf(stream, "                        xfer_cnt <= xfer_cnt + 1 ;\n") ; 
           fprintf(stream, "                        init <= '0' ;\n") ; 
           fprintf(stream, "                    else\n") ;
           fprintf(stream, "                        RSENC_SM <= rs_enc_s2 ;\n") ; 
           fprintf(stream, "                        xfer_cnt <= xfer_cnt + 1 ;\n") ; 
           fprintf(stream, "                        enc_feedback <= '0' ;\n") ; 
           fprintf(stream, "                        init <= '0' ;\n") ; 
           fprintf(stream, "                    end if ;\n") ; 
           fprintf(stream, "\n") ;
           fprintf(stream, "                when rs_enc_s2 => \n") ; 
           fprintf(stream, "                    if (xfer_cnt < N) then\n") ; 
           fprintf(stream, "                        RSENC_SM <= rs_enc_s2 ;\n") ;
           fprintf(stream, "                        xfer_cnt <= xfer_cnt + 1 ;\n") ;
           fprintf(stream, "                        else\n") ;
           fprintf(stream, "                        RSENC_SM <= rs_enc_s0 ;\n") ;
           fprintf(stream, "                        enb_enc <= '0' ;\n") ;
           fprintf(stream, "                        xfer_cnt <= (OTHERS => '0') ;\n") ;
           fprintf(stream, "                    end if ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "                when OTHERS =>\n") ;
           fprintf(stream, "                     RSENC_SM <= rs_enc_s0 ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "             end case ;\n") ;
           fprintf(stream, "        end if ;\n") ;
           fprintf(stream, "    end if ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "end process ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    enc_dov <= enc_feedback after 1 ns ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "encoder_syn_valid:process (rst_n, clk) \n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n = '0') then \n") ;
           fprintf(stream, "        syn_valid <= '0' ;\n") ;
           fprintf(stream, "    else if (clk'event and clk = '1') then\n") ;
           fprintf(stream, "        if (RSENC_SM = rs_enc_s1 and xfer_cnt = K-1) then\n") ;
           fprintf(stream, "            syn_valid <= '1' after 1 ns ;\n") ;
           fprintf(stream, "        elsif (done = '1') then\n") ;
           fprintf(stream, "            syn_valid <= '0' after 1 ns ;\n") ;		
           fprintf(stream, "        end if ;\n") ;
           fprintf(stream, "	    end if ;\n") ;
           fprintf(stream, "    end if ;\n") ;
           fprintf(stream, "end process ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    enc_synv <= syn_valid ; \n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "encoder_done:process (rst_n, clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n = '0') then \n") ;
           fprintf(stream, "        done <= '0' ;\n") ;
           fprintf(stream, "	else if (clk'event and clk = '1') then\n") ;
           fprintf(stream, "        if (RSENC_SM = rs_enc_s2 and xfer_cnt = N-2) then\n") ;
           fprintf(stream, "            done <= '1' ;\n") ;
           fprintf(stream, "        else\n") ;
           fprintf(stream, "            done <= '0' ;\n") ;		
           fprintf(stream, "        end if ;\n") ;
           fprintf(stream, "        end if ;\n") ;
           fprintf(stream, "    end if ;\n") ;
           fprintf(stream, "end process ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    enc_done <= done after 1 ns ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "END ;\n") ;
		   break ;
	   default :
		   break ;
   }
}

