/*******************************************************************************/
/*                                                                             */
/*            gen_forney.c                                                     */
/*                                                                             */
/*    Generate RS Receiver Syndrome Check                                      */
/*                                                                             */
/*                                                                             */
/*                                                                             */
/*                                                                             */
/*                                                                             */
/*                                                                             */
/*                                                                             */
/*    Copyright (C) 2003  John S. Willy (jswilly@pacbell.net)                  */ 
/*                                                                             */
/*    This program is free software; you can redistribute it and/or modify     */
/*    it under the terms of the GNU General Public License as published by     */
/*    the Free Software Foundation; either version 2 of the License, or        */
/*    (at your option) any later version.                                      */
/*                                                                             */
/*    This program is distributed in the hope that it will be useful,          */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*    GNU General Public License for more details.                             */
/*                                                                             */
/*    You should have received a copy of the GNU General Public License        */
/*    along with this program; if not, write to the Free Software              */
/*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA*/
/*******************************************************************************/
 
#include <stdio.h>
#include "rsccg_proto.h"

#define VHDL     0 
#define VLOG     1 
#define C        2 

extern unsigned char poly_table[] ;   // RS code table 
extern unsigned char log_poly_table[] ; 
extern unsigned int p_wsp ;           // RS Generator Polynomial

/* locals  */

void gen_rs_forney(FILE *stream, int base, int gf_m, int deg, int chk_sym, int hdl_type, int gf_field)
{
   int i, j, t, vlen ;
   unsigned char b_vector[16] ; // Binary bit vector

   vlen = vector_len(2*chk_sym) ;
   t = 2*chk_sym ;

   switch (hdl_type) {
       case C:
           fprintf(stream, "//--------------------------------------------------------------------//\n") ;
           fprintf(stream, "//                  Generated by RSTK                                 //\n") ;  
           fprintf(stream, "//                      rs_forney.c                                   //\n") ;  
           fprintf(stream, "//                                                                    //\n") ;  
           fprintf(stream, "//--------------------------------------------------------------------//\n") ;
           fprintf(stream, "/****************************************************************/\n") ;
           fprintf(stream, "/* Forney algorithm for error evaluator polynomial & correction */\n") ;
           fprintf(stream, "/****************************************************************/\n") ;
           fprintf(stream, "#include <stdio.h>\n") ;
           fprintf(stream, "#include \"proto_types.h\"\n") ;
           fprintf(stream, "extern unsigned char gf_alpha_tbl[] ;        // Alpha GF Feild (a^n)\n") ;
           fprintf(stream, "extern unsigned char gf_invalpha_tbl[] ;     // Iverse of Alpha GF Feild (a^n)\n") ;
           fprintf(stream, "extern unsigned char gf_powalpha_tbl[] ;     // Power of Alpha GF Feild (a^n)\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "void forney(v, syn, lam, n_roots, lam_root, corr_val, dbg ) \n") ;
           fprintf(stream, "int v ;\n") ;
           fprintf(stream, "unsigned char *syn, *lam ;\n") ;
           fprintf(stream, "int n_roots ;\n") ;
           fprintf(stream, "unsigned char *lam_root ;\n") ;
           fprintf(stream, "unsigned char *corr_val ;\n") ;
           fprintf(stream, "int dbg ;\n") ;
           fprintf(stream, "{\n") ;
           fprintf(stream, " int i, j, m, r ;\n") ;
           fprintf(stream, "unsigned char l[%d] ;\n", chk_sym+2) ;
           fprintf(stream, "unsigned char d_lam[%d] ;\n", chk_sym+2) ;
           fprintf(stream, "unsigned char s[%d] ;\n", 2*chk_sym) ;
           fprintf(stream, "unsigned char omega[%d] ;\n", chk_sym+2) ;
           fprintf(stream, "unsigned char norm, om, inv_k_lam, ei ;\n") ;
           fprintf(stream, "unsigned char  k_lam, k_om, z0, z1 ;\n") ;
           fprintf(stream, "unsigned char  roots[%d] ;\n", chk_sym) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "   for (i=0; i<%d; i++) {\n", 2*chk_sym) ;
           fprintf(stream, "     s[i] = 0 ;\n") ;
           fprintf(stream, "   }\n") ;
           fprintf(stream, "   for (i=0; i<%d; i++) {\n", chk_sym+2) ;
           fprintf(stream, "     l[i] = 0 ;\n") ;
           fprintf(stream, "     d_lam[i] = 0 ;\n") ;
           fprintf(stream, "     omega[i] = 0 ;\n") ;
           fprintf(stream, "   }\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "   for (i=0; i<chk_sym; i++) {\n") ;
           fprintf(stream, "     roots[i] = *lam_root++ ;\n") ;
           fprintf(stream, "     if (dbg == 1) printf(\"Lamda Roots = %%d\\n\",roots[i]) ;\n") ;
           fprintf(stream, "   }\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "   // Copy string\n") ;
           fprintf(stream, "   for (i=0; i<v; i++) {\n") ;
           fprintf(stream, "     l[i] = *lam++ ;\n") ;
           fprintf(stream, "     d_lam[i] = l[i] ;\n") ;
           fprintf(stream, "     s[i] = *syn++ ;\n") ;
           fprintf(stream, "     if (dbg == 1) printf(\"lam = %%2x\\n\", l[i]) ;\n") ;
           fprintf(stream, "   }\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "   //\n") ;
           fprintf(stream, "   // Generate derivative of Lambda by zeroing even powers of 2 in Lambda\n") ;
           fprintf(stream, "   // Note: L'(x) can be calculated as soon as L(x) has been computed by berl_mas()\n") ;
           fprintf(stream, "   //\n") ;
           fprintf(stream, "   for (i=0; i<v; i++) {\n") ;
           fprintf(stream, "     if (i%%2 == 0)\n") ; 
           fprintf(stream, "       d_lam[i] = 0 ;\n") ;
           fprintf(stream, "     if (dbg == 1) printf(\"d_lam = %%2x\\n\", d_lam[i]) ;\n") ;
           fprintf(stream, "   }\n") ;
           fprintf(stream, "   for (i=0; i<v; i++) { \n") ;
           fprintf(stream, "       d_lam[i] = d_lam[i+1] ;\n") ;
           fprintf(stream, "       if (dbg == 1) printf(\"shifted d_lam = %%2x\\n\", d_lam[i]) ;\n") ;
           fprintf(stream, "   }\n") ;
           fprintf(stream, "\n") ;   
           fprintf(stream, "   //\n") ;
           fprintf(stream, "   // Note: omega(x) can be calculated as soon as L(x) has been computed by berl_mas()\n") ;
           fprintf(stream, "   // \n") ; 
           fprintf(stream, "\n") ;
           fprintf(stream, "   omega[0] = s[0] ;\n") ;
           fprintf(stream, "   for (i=1; i< v; i++) {\n") ;
           fprintf(stream, "       omega[i] = s[i] ;\n") ;
           fprintf(stream, "           for (j=1; j<= i; j++) {\n") ;
           fprintf(stream, "              omega[i] = omega[i] ^ gfmul_ab(s[i-j], l[j]) ;\n") ;
           fprintf(stream, "           }\n") ;
           fprintf(stream, "   }\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "   for (i=0; i<v; i++)\n") ;
           fprintf(stream, "     if (dbg == 1) printf(\"omega[%%d] = %%2x\\n\", i, omega[i]) ;\n") ;
           fprintf(stream, "\n") ; 
           fprintf(stream, "\n") ;
           fprintf(stream, "   // \n") ;
           fprintf(stream, "   // Compute O(x)/L'(x)\n") ;
           fprintf(stream, "   //\n") ;
           fprintf(stream, "   for (r=0; r < n_roots; r++) {\n") ;
           fprintf(stream, "     k_lam = 0 ;\n") ;
           fprintf(stream, "     k_om = 0 ;\n") ;
           fprintf(stream, "     k_lam = k_lam ^ d_lam[0] ;\n") ;
           fprintf(stream, "     k_om  = k_om  ^ omega[0] ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "\n") ;    
           fprintf(stream, "     //\n") ;
           fprintf(stream, "     // Calculate k_om, k_lam for each root found by chien search\n") ;
           fprintf(stream, "     //\n") ;
           fprintf(stream, "     for (i=1; i<v; i++) {\n") ;
           fprintf(stream, "       z0 = gf_alpha_tbl[roots[r]] ;\n") ;
           fprintf(stream, "       z1 = 1 ;\n") ;
           fprintf(stream, "       for (m=1;m<=i; m++)\n") ;
           fprintf(stream, "          z1 = gfmul_ab(z0, z1) ;  // Power of root must be calculated to determine k_lam \n") ;
           fprintf(stream, "       k_lam = k_lam ^ gfmul_ab(d_lam[i], z1) ;\n") ;
           fprintf(stream, "       z0 = gf_alpha_tbl[roots[r]] ;\n") ;
           fprintf(stream, "       z1 = 1 ;\n") ;
           fprintf(stream, "       for (m=1;m<=i; m++)\n") ;
           fprintf(stream, "          z1 = gfmul_ab(z0, z1) ;  // Power of root must be calculated to determine k_om \n") ;
           fprintf(stream, "\n") ;    
           fprintf(stream, "       k_om  = k_om ^ gfmul_ab(omega[i], z1) ;\n") ;
           fprintf(stream, "     }\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "   //printf(\"Derivative of Lambda at root = %%2x\\n\", k_lam ) ;\n") ;
           fprintf(stream, "   //printf(\"Evaluation of Omega at root = %%2x\\n\", k_om ) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "   inv_k_lam = gf_invalpha_tbl[k_lam] ; // To devide O(x) by L'(x) get inverse of L'(x)\n") ;
           fprintf(stream, "\n") ;   
           if (base > 1)
               fprintf(stream, "   norm = gf_powalpha_tbl[roots[r]] ;  // Normalization factor for root to base\n") ;
           else
               fprintf(stream, "   norm = 0x01 ;  // Normalization factor for base = 0,1\n") ;
           fprintf(stream, "   om = gfmul_ab(k_om, inv_k_lam ) ; // calculate error value \n") ;
           fprintf(stream, "   ei = gfmul_ab(norm, om) ;         // and normalize to base\n") ; 
           fprintf(stream, "\n") ;
           fprintf(stream, "   //printf(\"norm = %%2x om = %%2x ei = %%2x\\n\", norm, om, ei) ;\n") ;
           fprintf(stream, "   *corr_val++ = ei ;\n") ;
           fprintf(stream, "   } \n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "}\n") ;
           break;

       case VLOG : 
           fprintf(stream, "//--------------------------------------------------------------------//\n") ;
           fprintf(stream, "//                  Generated by RSTK                                 //\n") ;  
           fprintf(stream, "//                      rs_forney.v                                   //\n") ;  
           fprintf(stream, "//                                                                    //\n") ;  
           fprintf(stream, "//--------------------------------------------------------------------//\n") ;
           fprintf(stream, "module forney( rst_n,\n") ;
           fprintf(stream, "           clk,\n") ;
           fprintf(stream, "           start_forney,\n") ;
           fprintf(stream, "           root_search_done,\n") ;
           for (i=0; i<2*chk_sym; i++)
               fprintf(stream, "           s%d,\n", i) ;
           for (i=0; i<2*chk_sym; i++)
               fprintf(stream, "           l%d,\n", i) ;
           for (i=0; i<chk_sym; i++)
               fprintf(stream, "           lroot%d,\n", i) ;
           fprintf(stream, "           n_roots,\n") ;
           for (i=0; i<chk_sym; i++)
               fprintf(stream, "           corr_err_val%d,\n", i) ;
           fprintf(stream, "           omega_dlam_rdy,\n") ;
           fprintf(stream, "           forney_done\n") ;
           fprintf(stream, "         ) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    input        rst_n ;\n") ;
           fprintf(stream, "    input        clk ;\n") ;
           fprintf(stream, "    input        init ;\n") ;
           fprintf(stream, "    input        start_forney ;\n") ;
           fprintf(stream, "    input        root_search_done ;\n") ;
           fprintf(stream, "\n") ;
           for (i=0; i<2*chk_sym; i++)
               fprintf(stream, "    input[%d:0]    s%d ;            // Syndrome 0\n", gf_m-1, i) ;
           fprintf(stream, "\n") ;
           for (i=0; i<2*chk_sym; i++)
               fprintf(stream, "    input[%d:0]    l%d ;            // L(x) coefiecient 0\n", gf_m-1, i) ;
           fprintf(stream, "\n") ;
           for (i=0; i< chk_sym; i++)
           fprintf(stream, "    input[%d:0]    lroot%d ;        // L(x) root 0\n", gf_m-1, i) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    input[%d:0]    n_roots ;       // Number of Roots\n", gf_m-1) ;
           fprintf(stream, "\n") ;  
           for (i=0; i< chk_sym; i++)
               fprintf(stream, "    output[%d:0]     corr_err_val%d ; // Correction error value %d\n", gf_m-1, i, i) ;
           fprintf(stream, "\n") ; 
           fprintf(stream, "    output          omega_dlam_rdy ; // Omega and derivative of L(x) ready\n") ;
           fprintf(stream, "    output          forney_done    ; // Forney Algorithm done\n") ;
           fprintf(stream, "\n") ; 
           fprintf(stream, "`include \"../gf_arith.v\"\n") ;
           fprintf(stream, "\n") ;			   
           fprintf(stream, "// Locals\n") ;
           fprintf(stream, "    wire[%d:0]    Om_x ;\n", gf_m-1) ;
           for (i=0; i<= chk_sym; i++)
           fprintf(stream, "    reg[%d:0]    Om_%d ;\n", gf_m-1, i) ;
           fprintf(stream, "    reg[%d:0]    Om_reg ;\n", gf_m-1) ;
           fprintf(stream, "\n") ; 
           fprintf(stream, "//  Forney Control SM\n") ;
           fprintf(stream, "    reg[3:0]    Forney_SM ;\n") ;
           fprintf(stream, "\n") ; 
           fprintf(stream, "    parameter   Forney_s0 = 0 ;\n") ;
           fprintf(stream, "    parameter   Forney_s1 = 1 ;\n") ;
           fprintf(stream, "    parameter   Forney_s2 = 2 ;\n") ;
           fprintf(stream, "    parameter   Forney_s3 = 3 ;\n") ;
           fprintf(stream, "    parameter   Forney_s4 = 4 ;\n") ;
           fprintf(stream, "    parameter   Forney_s5 = 5 ;\n") ;
           fprintf(stream, "    parameter   Forney_s6 = 6 ;\n") ;
           fprintf(stream, "    parameter   Forney_s7 = 7 ;\n") ;
           fprintf(stream, "    parameter   Forney_s8 = 8 ;\n") ;
           fprintf(stream, "    parameter   Forney_s9 = 9 ;\n") ;
           fprintf(stream, "    parameter   Forney_s10 = 10 ;\n") ;
           fprintf(stream, "\n") ;  
           fprintf(stream, "\n") ; 
           fprintf(stream, "//\n") ;
           for (i=0; i< 2*chk_sym; i++)
               fprintf(stream, "    parameter    syn_s%d    = %d ;\n", i, i) ;
           fprintf(stream, "\n") ;
           for (i=0; i< 2*chk_sym; i++)
               fprintf(stream, "    parameter    lam_l%d    = %d ;\n", i, i) ;
           for (i=0; i< 2*chk_sym; i++)
               fprintf(stream, "    parameter    Om_r%d    = %d ;\n", i, i) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "    wire[%d:0]    lam_mux ;\n", gf_m-1) ;
           fprintf(stream, "    wire[%d:0]    om_mux ;\n", gf_m-1) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    wire[%d:0]    K0 ;\n", gf_m-1) ;
           fprintf(stream, "    wire[%d:0]    K1 ;\n", gf_m-1) ;
           fprintf(stream, "\n") ;     
           fprintf(stream, "    wire[%d:0]   syn ;\n", gf_m-1) ;
           fprintf(stream, "    wire[%d:0]   syn_sel ;\n", vlen) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    wire[%d:0]   lam ;\n", gf_m-1) ;
           fprintf(stream, "    wire[%d:0]   lam_sel ;\n", vlen) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    reg[%d:0]    i ;\n", vlen) ;
           fprintf(stream, "    reg[%d:0]    j ;\n", vlen) ;
           fprintf(stream, "    reg[%d:0]    cnt ;\n", vlen) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    reg[%d:0]    n_root_cnt ;\n", vlen) ;
           fprintf(stream, "    reg[%d:0]    mux_poly_cnt  ;\n", vlen) ;
           fprintf(stream, "    reg         Sel_F  ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    wire[%d:0]    a_n ;\n", gf_m-1) ;
           fprintf(stream, "    reg[%d:0]    R_an ;\n", gf_m-1) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    wire[%d:0]   k_om ;\n", gf_m-1) ;
           fprintf(stream, "    reg[%d:0]    R_k_om ;\n", gf_m-1) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    wire[%d:0]   k_lam ;\n", gf_m-1) ;
           fprintf(stream, "    reg[%d:0]    R_k_lam ;\n", gf_m-1) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    reg[%d:0]    R_Om ;\n", gf_m-1) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    reg[%d:0]    Omega ;\n", gf_m-1) ;
           fprintf(stream, "\n") ;
           for (i=0; i< chk_sym; i++)
               fprintf(stream, "    reg[%d:0]    R_Ei_%d ;\n", gf_m-1, i) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    wire[%d:0]   om_lam_calc ;\n", gf_m-1) ;
           fprintf(stream, "    wire[%d:0]   inv_k_lam ;\n", gf_m-1) ;
           fprintf(stream, "    wire[%d:0]   norm ;\n", gf_m-1) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    reg         Decode_Done ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "//////////////////////////////////////////////////////////////////////////\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "//     ROM Tables for Forney Algorithm\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "///////////////////////////////////////////////////////////////////////////\n") ;
           fprintf(stream, "    wire[7:0]   inv_alpha_rom_adrs ;\n") ;
           fprintf(stream, "    wire[7:0]   gf_inv_alpha ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    wire[7:0]   gf_alpha_rom_adrs ;\n") ;
           fprintf(stream, "    wire[7:0]   gf_alpha ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    wire[7:0]   pow_alpha_rom_adrs ;\n") ;
           fprintf(stream, "    wire[7:0]   pow_alpha ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "// INSERT ROM HERE\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "// Forney Algorithm control\n") ; 
           fprintf(stream, "//\n") ;
           fprintf(stream, "always @(negedge rst_n or posedge clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n == 1'b0)\n") ;
           fprintf(stream, "        begin\n") ;
           fprintf(stream, "            Forney_SM <= Forney_s0  ;\n") ;
           fprintf(stream, "            cnt <= 0 ;\n") ;
           fprintf(stream, "            j <= %d ;\n", t-1) ;
           fprintf(stream, "            i <= %d ;\n", t-1) ;
           fprintf(stream, "            n_root_cnt <= 0 ;\n") ;
           fprintf(stream, "            mux_poly_cnt <= 0 ;\n") ;
           fprintf(stream, "            Sel_F <= 1'b0 ;\n") ;
           fprintf(stream, "        end\n") ;
           fprintf(stream, "     else\n") ;
           fprintf(stream, "         case (Forney_SM)\n") ;
           fprintf(stream, "             Forney_s0:\n") ;
           fprintf(stream, "                 begin\n") ;
           fprintf(stream, "                     if (start_forney == 1'b1)\n") ;
           fprintf(stream, "                         begin\n") ;
           fprintf(stream, "                             Forney_SM <= Forney_s1  ;\n") ;
           fprintf(stream, "                             j <= %d ;\n", t-1) ;
           fprintf(stream, "                             i <= %d ;\n", t-1) ;
           fprintf(stream, "                             cnt <= 0 ;\n") ;
           fprintf(stream, "                         end\n") ;
           fprintf(stream, "                     else \n") ;
           fprintf(stream, "                         Forney_SM <= Forney_s0  ;\n") ;
           fprintf(stream, "                 end\n") ;
           fprintf(stream, "             Forney_s1:\n") ;
           fprintf(stream, "                 begin\n") ;
           fprintf(stream, "                     if  (j > 0)\n") ;
           fprintf(stream, "                         begin\n") ;
           fprintf(stream, "                             Forney_SM <= Forney_s1  ;\n") ;
           fprintf(stream, "                             j <= j - 1 ;\n") ;
           fprintf(stream, "                         end\n") ;
           fprintf(stream, "                     else\n") ;
           fprintf(stream, "                         begin\n") ;
           fprintf(stream, "                             Forney_SM <= Forney_s2  ;\n") ;
           fprintf(stream, "                             cnt <= cnt + 1 ; \n") ;
           fprintf(stream, "                         end\n") ;
           fprintf(stream, "                 end\n") ;
           fprintf(stream, "             Forney_s2:\n") ;
           fprintf(stream, "                 begin\n") ;
           fprintf(stream, "                     if (cnt == %d)\n", 2*chk_sym) ;
           fprintf(stream, "                         begin\n") ;
           fprintf(stream, "                             Forney_SM <= Forney_s3 ;\n") ;
           fprintf(stream, "                         end\n") ; 
           fprintf(stream, "                     else\n") ;
           fprintf(stream, "                         begin\n") ;
           fprintf(stream, "                             Forney_SM <= Forney_s1  ;\n") ;
           fprintf(stream, "                             j <= i - 1 ;\n") ;
           fprintf(stream, "                             i <= i - 1 ;\n") ;
           fprintf(stream, "                         end \n") ;
           fprintf(stream, "                 end\n") ;
           fprintf(stream, "             Forney_s3:\n") ;
           fprintf(stream, "                 begin\n") ;
           fprintf(stream, "                     Forney_SM <= Forney_s4 ;\n") ;
           fprintf(stream, "                     n_root_cnt <= 0 ;\n") ;
           fprintf(stream, "                     mux_poly_cnt <= 0 ;\n") ;
           fprintf(stream, "                 end\n") ;
           fprintf(stream, "             Forney_s4:\n") ;
           fprintf(stream, "                 begin\n") ;
           fprintf(stream, "                     if (n_root_cnt == n_roots )\n") ;
           fprintf(stream, "                         Forney_SM <= Forney_s8  ;\n") ;
           fprintf(stream, "                     else\n") ;
           fprintf(stream, "                         Forney_SM <= Forney_s5  ;\n") ;
           fprintf(stream, "                 end\n") ;
           fprintf(stream, "             Forney_s5:\n") ;
           fprintf(stream, "                 begin\n") ;
           fprintf(stream, "                     if (mux_poly_cnt == %d)\n", chk_sym-1) ; 
           fprintf(stream, "                         begin\n") ;
           fprintf(stream, "                             Forney_SM <= Forney_s6  ;\n") ;
           fprintf(stream, "                         end\n") ;
           fprintf(stream, "                     else\n") ;
           fprintf(stream, "                         begin\n") ;
           fprintf(stream, "                             mux_poly_cnt <= mux_poly_cnt + 1 ;\n") ;
           fprintf(stream, "                             Forney_SM <= Forney_s5  ;\n") ;
           fprintf(stream, "                         end\n") ;
           fprintf(stream, "                 end\n") ;
           fprintf(stream, "             Forney_s6:\n") ;
           fprintf(stream, "                 begin\n") ;
           fprintf(stream, "                     Sel_F <= 1'b1 ;\n") ;
           fprintf(stream, "                     Forney_SM <= Forney_s7  ;\n") ;
           fprintf(stream, "                 end\n") ;
           fprintf(stream, "             Forney_s7:\n") ;
           fprintf(stream, "                 begin\n") ;
           fprintf(stream, "                     n_root_cnt <= n_root_cnt + 1 ;\n") ;
           fprintf(stream, "                     mux_poly_cnt <= 0 ;\n") ;
           fprintf(stream, "                     Sel_F <= 1'b0 ;\n") ;
           fprintf(stream, "                     Forney_SM <= Forney_s4  ;\n") ;
           fprintf(stream, "                 end\n") ;
           fprintf(stream, "             Forney_s8:\n") ;
           fprintf(stream, "                 begin\n") ;
           fprintf(stream, "                     Forney_SM <= Forney_s0  ;\n") ;
           fprintf(stream, "                 end\n") ;
           fprintf(stream, "         endcase\n") ;
           fprintf(stream, "end\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "// Generate Omega(x) polynomial\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "    assign syn_sel = i - j ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    assign syn = (syn_sel == syn_s0) ? s0 :\n") ;
           for (i=1; i<2*chk_sym; i++)
               fprintf(stream, "                 (syn_sel == syn_s%d) ? s%d :\n", i, i) ;
           fprintf(stream, "                 %d'b0 ;\n", gf_m) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    assign lam_sel = j ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    assign lam = (lam_sel == lam_l0) ? l0 :\n") ;
           for (i=0; i<2*chk_sym; i++)
               fprintf(stream, "                 (lam_sel == lam_l%d) ? l%d :\n", i, i) ;
           fprintf(stream, "                 %d'b0 ;\n", gf_m) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    assign Om_x = gfmul_ab(syn,lam) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "always @(negedge rst_n or posedge clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n == 1'b0)\n") ;
           fprintf(stream, "        begin\n") ;
           fprintf(stream, "           Om_reg <= %d'b0 ;\n", gf_m) ;
           fprintf(stream, "        end\n") ;
           fprintf(stream, "    else if (Forney_SM == Forney_s1)\n") ;
           fprintf(stream, "        begin \n") ;         
           fprintf(stream, "            Om_reg <= Om_reg ^ Om_x ;\n") ;
           fprintf(stream, "        end\n") ;
           fprintf(stream, "    else if (Forney_SM == Forney_s2)\n") ;
           fprintf(stream, "        begin\n") ;          
           fprintf(stream, "            Om_reg <= %d'b0 ;\n", gf_m) ;
           fprintf(stream, "        end\n") ;
           fprintf(stream, "end\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "always @(negedge rst_n or posedge clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n == 1'b0)\n") ;
           fprintf(stream, "        begin\n") ;
           for (i=0; i<2*chk_sym; i++)
               fprintf(stream, "           Om_%d <= %d'b0 ;\n", i, gf_m) ;
           fprintf(stream, "        end\n") ;
           fprintf(stream, "    else if (Forney_SM == Forney_s2)\n") ;
           fprintf(stream, "        case (cnt)\n") ;
           for (i=1; i<=2*chk_sym; i++) {
               fprintf(stream, "            %d: begin\n", i) ;
               fprintf(stream, "                   Om_%d <= Om_reg ;\n", 2*chk_sym-i) ;
               fprintf(stream, "               end\n") ;
           }
           fprintf(stream, "        endcase\n") ;
           fprintf(stream, "end\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "// Compute powers of a^n for each term in O(x) and L'(x)\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "always @(negedge rst_n or posedge clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n == 1'b0)\n") ;
           fprintf(stream, "        begin\n") ;
           fprintf(stream, "           R_an <= %d'b0 ;\n", gf_m) ;
           fprintf(stream, "        end\n") ;
           fprintf(stream, "    else if (Forney_SM == Forney_s3 || Forney_SM == Forney_s7)\n") ;
           fprintf(stream, "        begin\n") ;          
           fprintf(stream, "            R_an[0] <= 1'b1 ;\n") ;
           fprintf(stream, "            R_an[%d:1] <= %d'b0 ;\n", gf_m-1, gf_m-1) ;
           fprintf(stream, "        end\n") ;
           fprintf(stream, "    else if (Forney_SM == Forney_s5)\n") ;
           fprintf(stream, "        begin\n") ;          
           fprintf(stream, "            R_an <= a_n ;\n") ;
           fprintf(stream, "        end\n") ;
           fprintf(stream, "end\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    assign gf_alpha_rom_adrs = (n_root_cnt == 0) ? lroot0 :\n") ;
           for (i=i; i<chk_sym; i++)
               fprintf(stream, "                               (n_root_cnt == %d) ? lroot%d :\n", i, i) ;
           fprintf(stream, "                               %d'b0 ;\n", gf_m) ;
            fprintf(stream, "\n") ;            
           fprintf(stream, "    assign a_n = gfmul_ab(gf_alpha,R_an) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "// Compute Omega(root of lamda)\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "    assign om_mux = (mux_poly_cnt == Om_r0) ? Om_0 :\n") ;
           for (i=1; i<2*chk_sym; i++)
               fprintf(stream, "                    (mux_poly_cnt == Om_r%d) ? Om_%d :\n", i, i) ;
           fprintf(stream, "                    %d'b0 ;\n", gf_m) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    assign k_om = gfmul_ab(R_an, om_mux) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "// Compute  O(x)\n") ; 
           fprintf(stream, "//\n") ;
           fprintf(stream, "always @(negedge rst_n or posedge clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n == 1'b0)\n") ;
           fprintf(stream, "        begin\n") ;
           fprintf(stream, "           R_k_om <= %d'b0 ;\n", gf_m) ;
           fprintf(stream, "        end\n") ;
           fprintf(stream, "    else if (Forney_SM == Forney_s3 || Forney_SM == Forney_s7 )\n") ;
           fprintf(stream, "        begin\n") ;          
           fprintf(stream, "            R_k_om <= %d'b0 ;\n", gf_m) ;
           fprintf(stream, "        end\n") ;
           fprintf(stream, "    else if (Forney_SM == Forney_s5)\n") ;
           fprintf(stream, "            R_k_om <= R_k_om ^ k_om ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "end\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "// Compute deriative (root of lamda)\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "    // Derivative of lamda is created by zeroing even powers\n") ;
           fprintf(stream, "    assign lam_mux = (mux_poly_cnt == lam_l0) ? l1 :\n") ;
           for (i=2; i<=2*chk_sym; i++) {
               if (i%2 == 0)
                   fprintf(stream, "                     (mux_poly_cnt == lam_l%d) ? %d'b0 :\n", i-1, gf_m) ;
               else
                   fprintf(stream, "                     (mux_poly_cnt == lam_l%d) ? l%d :\n", i-1, i) ;
           }
           fprintf(stream, "                     %d'b0 ;\n", gf_m) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    assign k_lam = gfmul_ab(R_an, lam_mux) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "// Compute  L'(x)\n") ; 
           fprintf(stream, "//\n") ;
           fprintf(stream, "always @(negedge rst_n or posedge clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n == 1'b0)\n") ;
           fprintf(stream, "        begin\n") ;
           fprintf(stream, "           R_k_lam <= %d'b0 ;\n", gf_m) ;
           fprintf(stream, "        end\n") ;
           fprintf(stream, "    else if (Forney_SM == Forney_s3 || Forney_SM == Forney_s7 )\n") ;
           fprintf(stream, "        begin\n") ;          
           fprintf(stream, "            R_k_lam <= %d'b0 ;\n", gf_m) ;
           fprintf(stream, "        end\n") ;
           fprintf(stream, "    else if (Forney_SM == Forney_s5)\n") ;
           fprintf(stream, "            R_k_lam <= R_k_lam ^ k_lam ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "end\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "// Store  O(x)/L'(x) \n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "always @(negedge rst_n or posedge clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n == 1'b0)\n") ;
           fprintf(stream, "        begin\n") ;
           fprintf(stream, "           R_Om <= %d'b0 ;\n", gf_m) ;
           fprintf(stream, "        end\n") ;
           fprintf(stream, "    else if (Forney_SM == Forney_s3 || Forney_SM == Forney_s7 )\n") ;
           fprintf(stream, "        begin\n") ;          
           fprintf(stream, "            R_Om <= %d'b0 ;\n", gf_m) ;
           fprintf(stream, "        end\n") ;
           fprintf(stream, "    else if (Forney_SM == Forney_s6)\n") ;
           fprintf(stream, "            R_Om <= om_lam_calc ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "end\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    assign K0 = (Sel_F == 1'b0) ? R_k_om : norm ;\n") ;
           fprintf(stream, "    assign K1 = (Sel_F == 1'b0) ? inv_k_lam : R_Om ;\n") ;
           fprintf(stream, "\n") ;
		   if (base > 1) {
               fprintf(stream, "    assign pow_alpha_rom_adrs = gf_alpha_rom_adrs ;\n") ;
               fprintf(stream, "    assign norm = pow_alpha ;\n") ;
		   } else {
               fprintf(stream, "    assign norm = %d'b", gf_m) ;
			   for(i=0 ; i<gf_m-1; i++) fprintf(stream,"0") ;
			   fprintf(stream,"1;\n") ;
		   }
           fprintf(stream, "\n") ;
           fprintf(stream, "    assign inv_alpha_rom_adrs = R_k_lam ;\n") ;
           fprintf(stream, "    assign inv_k_lam = gf_inv_alpha ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    assign om_lam_calc = gfmul_ab(K0, K1) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "// Store Ei values to correct message\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "always @(negedge rst_n or posedge clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n == 1'b0)\n") ;
           fprintf(stream, "        begin\n") ;
           for (i=0; i<chk_sym; i++)
               fprintf(stream, "           R_Ei_%d <= %d'b0 ;\n", i, gf_m) ;
           fprintf(stream, "        end\n") ;
           fprintf(stream, "    else if (Forney_SM == Forney_s7)\n") ;
           fprintf(stream, "        case (n_root_cnt)\n") ;
           for (i=0; i<chk_sym; i++) {
               fprintf(stream, "            %d: begin\n", i) ;
               fprintf(stream, "                   R_Ei_%d <= om_lam_calc ;\n", i) ;
               fprintf(stream, "               end\n") ;
           }
           fprintf(stream, "         endcase\n") ;
           fprintf(stream, "end\n") ;
           fprintf(stream, "\n") ;
           for (i=0; i<chk_sym; i++)
               fprintf(stream, "    assign corr_err_val%d = R_Ei_%d ;\n", i, i) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "// Indicate correction values ready\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "always @(negedge rst_n or posedge clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n == 1'b0)\n") ;
           fprintf(stream, "        begin\n") ;
           fprintf(stream, "           Decode_Done <= 1'b0 ;\n") ;
           fprintf(stream, "        end\n") ;
           fprintf(stream, "    else if (Forney_SM == Forney_s8)\n") ;
           fprintf(stream, "           Decode_Done <= 1'b1 ;\n") ;
           fprintf(stream, "    else\n") ;
           fprintf(stream, "           Decode_Done <= 1'b0 ;\n") ;
           fprintf(stream, "end\n") ;
           fprintf(stream, "    assign forney_done = Decode_Done ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "endmodule\n") ; 
           break;

       case VHDL :
           fprintf(stream, "------------------------------------------------------------------------\n") ;
           fprintf(stream, "--                  Generated by RSTK                                 --\n") ;  
           fprintf(stream, "--                      rs_forney.vhd                                 --\n") ;  
           fprintf(stream, "--                                                                    --\n") ;  
           fprintf(stream, "------------------------------------------------------------------------\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "LIBRARY IEEE ;\n") ;
           fprintf(stream, "use IEEE.std_logic_1164.all ;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "library IEEE ;\n") ;
           fprintf(stream, "use     IEEE.std_logic_unsigned.all;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "library IEEE ;\n") ;
           fprintf(stream, "use     IEEE.NUMERIC_STD.all;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "library gf_arith ;\n") ;
           fprintf(stream, "use     gf_arith.gf_arith.all;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "library io_lib ;\n") ;
           fprintf(stream, "use     io_lib.io_utils.all;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "entity forney is port ( rst_n : IN std_logic;\n") ;
           fprintf(stream, "                        clk : IN std_logic;\n") ;
           fprintf(stream, "                        start_forney : IN std_logic;\n") ;
           fprintf(stream, "                        root_search_done : IN std_logic;\n") ;
           for (i=0; i< 2*chk_sym; i++)
               fprintf(stream, "                        s%d : IN std_logic_vector(%d downto 0);\n", i, gf_m-1) ;
           for (i=0; i< 2*chk_sym; i++)
              fprintf(stream, "                        l%d : IN std_logic_vector(%d downto 0);\n", i, gf_m-1) ;
           for (i=0; i< chk_sym; i++)
               fprintf(stream, "                        lroot%d : IN std_logic_vector(%d downto 0);\n", i, gf_m-1) ; 
           fprintf(stream, "                        n_roots : IN std_logic_vector(%d downto 0);\n", gf_m-1) ;
           for (i=0; i< chk_sym; i++)
               fprintf(stream, "                        corr_err_val%d : OUT std_logic_vector(%d downto 0);\n", i, gf_m-1) ;
           fprintf(stream, "                        omega_dlam_rdy : OUT std_logic;\n") ;
           fprintf(stream, "                        forney_done : OUT std_logic\n") ;
           fprintf(stream, "                        ) ;\n") ;
           fprintf(stream, "end forney ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "architecture rtl of forney is\n") ;
           fprintf(stream, "-- Locals\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    signal    Om_x : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           for (i=0; i< 2*chk_sym; i++)
               fprintf(stream, "    signal    Om_%d : std_logic_vector(%d downto 0);\n", i, gf_m-1) ;
           fprintf(stream, "    signal    Om_reg : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "--  Forney Control SM\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    type forney_states is (\n") ;
           fprintf(stream, "        Forney_s0,\n") ;
           fprintf(stream, "        Forney_s1,\n") ;
           fprintf(stream, "        Forney_s2,\n") ;
           fprintf(stream, "        Forney_s3,\n") ;
           fprintf(stream, "        Forney_s4,\n") ;
           fprintf(stream, "        Forney_s5,\n") ;
           fprintf(stream, "        Forney_s6,\n") ;
           fprintf(stream, "        Forney_s7,\n") ;
           fprintf(stream, "        Forney_s8,\n") ;
           fprintf(stream, "        Forney_s9,\n") ;
           fprintf(stream, "        Forney_s10  ) ;\n") ;
           fprintf(stream, "\n") ; 
           fprintf(stream, "    signal Forney_SM : forney_states;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "--\n") ;
           for (i=0; i< 2*chk_sym; i++) {
               fprintf(stream, "    constant    syn_s%d  : std_logic_vector(%d downto 0) := \"", i, vlen) ;
               conv2_bvec_str(i, vlen, b_vector) ;
               pr_vec_str(stream, vlen, b_vector) ;
               fprintf(stream, "\" ;\n") ;      
            }
           fprintf(stream, "\n") ;
           for (i=0; i< 2*chk_sym; i++) {
               fprintf(stream, "    constant    lam_l%d  : std_logic_vector(%d downto 0) := \"", i, vlen) ;
               conv2_bvec_str(i, vlen, b_vector) ;
               pr_vec_str(stream, vlen, b_vector) ;
               fprintf(stream, "\" ;\n") ;      
            }
           fprintf(stream, "\n") ;
           for (i=0; i< 2*chk_sym; i++) {
               fprintf(stream, "    constant    Om_r%d   : std_logic_vector(%d downto 0) := \"", i, vlen) ;
               conv2_bvec_str(i, vlen, b_vector) ;
               pr_vec_str(stream, vlen, b_vector) ;
               fprintf(stream, "\" ;\n") ;  
           }
           fprintf(stream, "--\n") ;
           fprintf(stream, "    signal lam_mux : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "    signal om_mux : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    signal K0 : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "    signal K1 : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "\n") ;     
           fprintf(stream, "    signal syn : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "    signal syn_sel : std_logic_vector(%d downto 0);\n", vlen) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    signal lam : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "    signal lam_sel: std_logic_vector(%d downto 0);\n", vlen) ;
           fprintf(stream, "    signal i : std_logic_vector(%d downto 0);\n", vlen) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    signal j : std_logic_vector(%d downto 0);\n", vlen) ;
           fprintf(stream, "    signal cnt : std_logic_vector(%d downto 0);\n", vlen) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    signal n_root_cnt : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "    signal mux_poly_cnt  : std_logic_vector(%d downto 0);\n", vlen) ;
           fprintf(stream, "    signal Sel_F  : std_logic;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    signal a_n : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "    signal R_an : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    signal k_om : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "    signal R_k_om : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    signal k_lam : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "    signal R_k_lam : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    signal R_Om : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    signal Omega : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           for (i=0; i < chk_sym; i++)
               fprintf(stream, "    signal R_Ei_%d : std_logic_vector(%d downto 0);\n", i, gf_m-1) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    signal om_lam_calc : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "    signal inv_k_lam : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "    signal norm : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    signal Decode_Done : std_logic ;\n") ;
           fprintf(stream, "--------------------------------------------------------------------------\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--     ROM Instanciation for Forney Algorithm\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--     This table is loaded into RAM during initialization\n") ;
           fprintf(stream, "--------------------------------------------------------------------------\n") ;
           fprintf(stream, "    signal gf_alpha_rom_adrs : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "    signal gf_alpha : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    signal inv_alpha_rom_adrs : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "    signal gf_inv_alpha : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           fprintf(stream, "\n") ;
		   if (base > 1) {
               fprintf(stream, "    signal pow_alpha_rom_adrs : std_logic_vector(%d downto 0);\n", gf_m-1) ;
               fprintf(stream, "    signal pow_alpha : std_logic_vector(%d downto 0);\n", gf_m-1) ;
           }
           fprintf(stream, "\n") ;
           fprintf(stream, "component spram_block is\n") ;
           fprintf(stream, "  generic ( RAM_NUM_WORDS : integer ;\n") ;
           fprintf(stream, "            RAM_WIDTH_A   : integer ;\n") ;
           fprintf(stream, "            RAM_WIDTH_D   : integer ;\n") ;
           fprintf(stream, "            RAM_FILE      : string \n") ;
           fprintf(stream, "           );\n") ;
           fprintf(stream, "  port (clk  : in std_logic;\n") ;
           fprintf(stream, "        we   : in std_logic;\n") ;
           fprintf(stream, "        addr : in std_logic_vector(RAM_WIDTH_A-1 downto 0);\n") ;
           fprintf(stream, "        di   : in std_logic_vector(RAM_WIDTH_D-1 downto 0);\n") ;
           fprintf(stream, "        do   : out std_logic_vector(RAM_WIDTH_D-1 downto 0)\n") ;
           fprintf(stream, "        );\n") ;
           fprintf(stream, "end component;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "--------------------------------------------------------------------------\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--     ROM Tables for Forney Algorithm\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--------------------------------------------------------------------------\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "-- GF ROM Table a^n\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "-- GF256 ROM table for a^n\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "  forney_alpha_rom: spram_block\n") ;
           fprintf(stream, "       generic map( RAM_NUM_WORDS => %d,\n", gf_field) ;
           fprintf(stream, "                    RAM_WIDTH_A => %d,\n", gf_m) ;
           fprintf(stream, "                    RAM_WIDTH_D => %d,\n", gf_m) ;
           fprintf(stream, "                    RAM_FILE => \"gf_alpha_rom.hex\")\n") ;
           fprintf(stream, "       port map(\n") ;
           fprintf(stream, "        clk => clk,\n") ;
           fprintf(stream, "        we => '0',\n") ;
           fprintf(stream, "        addr => gf_alpha_rom_adrs,\n") ;
           fprintf(stream, "        di => \"") ;
		   for (i=0; i< gf_m ; i++)
               fprintf(stream, "0") ;
		   fprintf(stream, "\",\n") ;
           fprintf(stream, "        do => gf_alpha\n") ;
           fprintf(stream, "        ) ;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "-- GF ROM Table Inverse a^n\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "  forney_inv_rom: spram_block\n") ;
           fprintf(stream, "       generic map( RAM_NUM_WORDS => %d,\n", gf_field) ;
           fprintf(stream, "                    RAM_WIDTH_A => %d,\n", gf_m) ;
           fprintf(stream, "                    RAM_WIDTH_D => %d,\n", gf_m) ;
           fprintf(stream, "                    RAM_FILE => \"gf_inv_rom.hex\")\n") ;
           fprintf(stream, "       port map(\n") ;
           fprintf(stream, "        clk => clk,\n") ;
           fprintf(stream, "        we => '0',\n") ;
           fprintf(stream, "        addr => inv_alpha_rom_adrs,\n") ;
           fprintf(stream, "        di => \"") ;
		   for (i=0; i< gf_m ; i++)
               fprintf(stream, "0") ;
		   fprintf(stream, "\",\n") ;
           fprintf(stream, "        do => gf_inv_alpha\n") ;
           fprintf(stream, "        ) ;\n") ;
		   if (base > 1) {
               fprintf(stream, "--\n") ;
               fprintf(stream, "-- GF Power ROM Table (a^n)^(B-1)\n") ;
               fprintf(stream, "--\n") ;
               fprintf(stream, "-- Generates the power of a^n to b-1\n") ;
               fprintf(stream, "--\n") ;
               fprintf(stream, "  forney_pow_rom: spram_block\n") ;
               fprintf(stream, "       generic map( RAM_NUM_WORDS => %d,\n", gf_field) ;
               fprintf(stream, "                    RAM_WIDTH_A => %d,\n", gf_m) ;
               fprintf(stream, "                    RAM_WIDTH_D => %d,\n", gf_m) ;
               fprintf(stream, "                    RAM_FILE => \"gf_pow_rom.hex\")\n") ;
               fprintf(stream, "       port map(\n") ;
               fprintf(stream, "        clk => clk,\n") ;
               fprintf(stream, "        we => '0',\n") ;
               fprintf(stream, "        addr => pow_alpha_rom_adrs,\n") ;
               fprintf(stream, "        di => \"") ;
		       for (i=0; i< gf_m ; i++)
                   fprintf(stream, "0") ;
		       fprintf(stream, "\",\n") ;
               fprintf(stream, "        do => pow_alpha\n") ;
               fprintf(stream, "        ) ;\n") ;
           }
           fprintf(stream, "\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "-- Forney Algorithm control \n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "process(rst_n, clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n = '0') then\n") ;
           fprintf(stream, "            Forney_SM <= Forney_s0  ;\n") ;
           fprintf(stream, "            cnt <= \"") ;
               conv2_bvec_str(0, vlen, b_vector) ;
               pr_vec_str(stream, vlen, b_vector) ;
               fprintf(stream, "\" ;\n") ;      
           fprintf(stream, "            j <= \"") ;
               conv2_bvec_str(t-1, vlen, b_vector) ;
               pr_vec_str(stream, vlen, b_vector) ;
               fprintf(stream, "\" ;\n") ;      
           fprintf(stream, "            i <= \"") ;
               conv2_bvec_str(t-1, vlen, b_vector) ;
               pr_vec_str(stream, vlen, b_vector) ;
               fprintf(stream, "\" ;\n") ;      
           fprintf(stream, "            n_root_cnt <= (OTHERS => '0') ;\n") ;
           fprintf(stream, "            mux_poly_cnt <=  (OTHERS => '0') ;\n") ;
           fprintf(stream, "            Sel_F <= '0' ;\n") ;
           fprintf(stream, "    elsif (clk'event and clk = '1')then\n") ;
           fprintf(stream, "         case Forney_SM is\n") ;
           fprintf(stream, "             when Forney_s0 =>\n") ;
           fprintf(stream, "                     if (start_forney = '1') then\n") ;
           fprintf(stream, "                             Forney_SM <= Forney_s1  ;\n") ;
           fprintf(stream, "                             j <= \"") ;
               conv2_bvec_str(t-1, vlen, b_vector) ;
               pr_vec_str(stream, vlen, b_vector) ;
               fprintf(stream, "\" ;\n") ;      
           fprintf(stream, "                             i <= \"") ;
               conv2_bvec_str(t-1, vlen, b_vector) ;
               pr_vec_str(stream, vlen, b_vector) ;
               fprintf(stream, "\" ;\n") ;      
           fprintf(stream, "                             cnt <= \"") ;
               conv2_bvec_str(0, vlen, b_vector) ;
               pr_vec_str(stream, vlen, b_vector) ;
               fprintf(stream, "\" ;\n") ;      
           fprintf(stream, "                     else\n") ; 
           fprintf(stream, "                         Forney_SM <= Forney_s0  ;\n") ;
           fprintf(stream, "                     end if ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "             when Forney_s1 =>\n") ;
           fprintf(stream, "                     if  (j > 0) then\n") ;
           fprintf(stream, "                             Forney_SM <= Forney_s1  ;\n") ;
           fprintf(stream, "                             j <= j - 1 ;\n") ;
           fprintf(stream, "                     else\n") ;
           fprintf(stream, "                             Forney_SM <= Forney_s2  ;\n") ;
           fprintf(stream, "                             cnt <= cnt + 1 ;\n") ; 
           fprintf(stream, "                     end if ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "             when Forney_s2 =>\n") ;
           fprintf(stream, "                     if (cnt = %d) then\n", 2*chk_sym) ;
           fprintf(stream, "                             Forney_SM <= Forney_s3  ;\n") ;
           fprintf(stream, "                     else\n") ;
           fprintf(stream, "                             Forney_SM <= Forney_s1  ;\n") ;
           fprintf(stream, "                             j <= i - 1 ;\n") ;
           fprintf(stream, "                             i <= i - 1 ;\n") ;
           fprintf(stream, "                     end if ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "             when Forney_s3 =>\n") ;
           fprintf(stream, "                     Forney_SM <= Forney_s4  ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "                     n_root_cnt <= (OTHERS => '0') ;\n") ;
           fprintf(stream, "                     mux_poly_cnt <= (OTHERS => '0') ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "             when Forney_s4 =>\n") ;
           fprintf(stream, "                     if (n_root_cnt = n_roots ) then\n") ;
           fprintf(stream, "                         Forney_SM <= Forney_s8  ;\n") ;
           fprintf(stream, "                     else\n") ;
           fprintf(stream, "                         Forney_SM <= Forney_s5  ;\n") ;
           fprintf(stream, "                     end if ;\n") ;
           fprintf(stream, "\n") ; 
           fprintf(stream, "             when Forney_s5 =>\n") ;
           fprintf(stream, "                     if (mux_poly_cnt = %d) then\n", chk_sym-1) ;
           fprintf(stream, "                             Forney_SM <= Forney_s6  ;\n") ;
           fprintf(stream, "                     else\n") ;
           fprintf(stream, "                             mux_poly_cnt <= mux_poly_cnt + 1 ;\n") ;
           fprintf(stream, "                             Forney_SM <= Forney_s5  ;\n") ;
           fprintf(stream, "                     end if ;\n") ;
           fprintf(stream, "\n") ; 
           fprintf(stream, "             when Forney_s6 =>\n") ;
           fprintf(stream, "                     Sel_F <= '1' ;\n") ;
           fprintf(stream, "                     Forney_SM <= Forney_s7  ;\n") ;
           fprintf(stream, "\n") ; 
           fprintf(stream, "             when Forney_s7 =>\n") ;
           fprintf(stream, "                     n_root_cnt <= n_root_cnt + 1 ;\n") ;
           fprintf(stream, "                     mux_poly_cnt <= \"") ;
               conv2_bvec_str(0, vlen, b_vector) ;
               pr_vec_str(stream, vlen, b_vector) ;
               fprintf(stream, "\" ;\n") ;      
           fprintf(stream, "                     Sel_F <= '0' ;\n") ;
           fprintf(stream, "                     Forney_SM <= Forney_s4  ;\n") ;
           fprintf(stream, "\n") ; 
           fprintf(stream, "             when Forney_s8 =>\n") ;
           fprintf(stream, "                     Forney_SM <= Forney_s0  ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "             when OTHERS =>\n") ;
           fprintf(stream, "                     Forney_SM <= Forney_s0  ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "         end case ;\n") ;
           fprintf(stream, "    end if ;\n") ;
           fprintf(stream, "end process ;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "-- Generate Omega(x) polynomial\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "    syn_sel <= i - j ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    with syn_sel select\n") ;
           fprintf(stream, "        syn <= s0 when  syn_s0,\n") ; 
           for (i=1; i<2*chk_sym; i++)
               fprintf(stream, "               s%d when  syn_s%d,\n", i, i) ;
           fprintf(stream, "                (OTHERS => '0')  when others ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    lam_sel <= j ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    with lam_sel select\n") ;
           fprintf(stream, "         lam <= l0 when lam_l0,\n") ;
           for (i=1; i< 2*chk_sym; i++)
               fprintf(stream, "                l%d when lam_l%d,\n", i, i) ;
           fprintf(stream, "                (OTHERS => '0')  when others ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "     Om_x <= gfmul_ab(syn,lam) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "process(rst_n, clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n = '0') then\n") ;
           fprintf(stream, "           Om_reg <= (OTHERS => '0') ;\n") ;
           fprintf(stream, "    elsif (clk'event and clk = '1')then\n") ;
           fprintf(stream, "        if (Forney_SM = Forney_s1) then\n") ;
           fprintf(stream, "            Om_reg <= Om_reg xor Om_x ;\n") ;
           fprintf(stream, "        elsif (Forney_SM = Forney_s2) then\n") ;
           fprintf(stream, "            Om_reg <= (OTHERS => '0') ;\n") ;
           fprintf(stream, "        end if ;\n") ;
           fprintf(stream, "    end if ;\n") ;
           fprintf(stream, "end process ;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "process(rst_n, clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n = '0') then\n") ;
           for (i=0; i<2*chk_sym; i++)
               fprintf(stream, "           Om_%d <= (OTHERS => '0') ;\n", i) ;
           fprintf(stream, "    elsif (clk'event and clk = '1')then\n") ;
           fprintf(stream, "      if (Forney_SM = Forney_s2) then\n") ;
           fprintf(stream, "        case cnt is\n") ;
           for (i=1; i<=2*chk_sym; i++) {
           fprintf(stream, "            when \"") ; 
               conv2_bvec_str(i, vlen, b_vector) ;
               pr_vec_str(stream, vlen, b_vector) ;
               fprintf(stream, "\" => \n") ;      
               fprintf(stream, "                   Om_%d <= Om_reg ;\n", 2*chk_sym-i) ;
               fprintf(stream, "\n") ;  
           }
           fprintf(stream, "            when OTHERS =>\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "        end case ;\n") ;
           fprintf(stream, "      end if ;\n") ;
           fprintf(stream, "   end if ;\n") ;
           fprintf(stream, "end process ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "-- Compute powers of a^n for each term in O(x) and L'(x)\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "process(rst_n, clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n = '0') then\n") ;
           fprintf(stream, "           R_an <= (OTHERS => '0') ;\n") ;
           fprintf(stream, "    elsif (clk'event and clk = '1')then\n") ;
           fprintf(stream, "        if (Forney_SM = Forney_s3 or Forney_SM = Forney_s7) then\n") ;
           fprintf(stream, "            R_an(0) <= '1' ;\n") ;
           fprintf(stream, "            R_an(%d downto 1) <= (OTHERS => '0') ;\n", gf_m-1) ;
           fprintf(stream, "        elsif (Forney_SM = Forney_s5) then\n") ;
           fprintf(stream, "            R_an <= a_n ;\n") ;
           fprintf(stream, "        end if ;\n") ;
           fprintf(stream, "    end if ;\n") ;
           fprintf(stream, "end process ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "     gf_alpha_rom_adrs <= lroot0 when (n_root_cnt = 0) else\n") ;
           for (i=1; i<chk_sym; i++)
               fprintf(stream, "                          lroot%d when (n_root_cnt = %d) else\n", i, i) ;
           fprintf(stream, "                          (OTHERS => '0') ;\n") ;
           fprintf(stream, "\n") ;             
           fprintf(stream, "     a_n <= gfmul_ab(gf_alpha,R_an) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "-- Compute Omega(root of lamda)\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "     with mux_poly_cnt select\n") ; 
           fprintf(stream, "          om_mux <= Om_0 when Om_r0,\n") ;
           for (i=1; i< 2*chk_sym; i++)
               fprintf(stream, "                    Om_%d when Om_r%d,\n", i, i) ;
           fprintf(stream, "                     (OTHERS => '0') when others ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    k_om <= gfmul_ab(R_an, om_mux) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "-- Compute  O(x)\n") ; 
           fprintf(stream, "--\n") ;
           fprintf(stream, "process(rst_n, clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n = '0') then\n") ;
           fprintf(stream, "           R_k_om <= (OTHERS => '0') ;\n") ;
           fprintf(stream, "    elsif (clk'event and clk = '1')then\n") ;
           fprintf(stream, "        if (Forney_SM = Forney_s3 or Forney_SM = Forney_s7 ) then\n") ;
           fprintf(stream, "            R_k_om <= (OTHERS => '0') ;\n") ;
           fprintf(stream, "        elsif (Forney_SM = Forney_s5) then\n") ;
           fprintf(stream, "            R_k_om <= R_k_om xor k_om ;\n") ;
           fprintf(stream, "        end if ;\n") ;
           fprintf(stream, "    end if ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "end process ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "-- Compute derivative (root of lamda)\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "    -- Derivative of lamda is created by zeroing even powers\n") ;
           fprintf(stream, "      lam_mux <= l1 when (mux_poly_cnt = lam_l0) else\n") ;
           for (i=2; i<2*chk_sym; i++) {
               if (i%2 == 0)
                   fprintf(stream, "                 (OTHERS => '0') when (mux_poly_cnt = lam_l%d) else\n", i-1) ;
               else
                   fprintf(stream, "                 l%i when (mux_poly_cnt = lam_l%d) else\n", i, i-1) ;
           }
           fprintf(stream, "                 (OTHERS => '0') ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "     k_lam <= gfmul_ab(R_an, lam_mux) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "-- Compute  L'(x) \n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "process(rst_n, clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n = '0') then\n") ;
           fprintf(stream, "           R_k_lam <= (OTHERS => '0') ;\n") ;
           fprintf(stream, "    elsif (clk'event and clk = '1')then\n") ;
           fprintf(stream, "        if (Forney_SM = Forney_s3 or Forney_SM = Forney_s7 ) then\n") ;
           fprintf(stream, "            R_k_lam <= (OTHERS => '0') ;\n") ;
           fprintf(stream, "        elsif (Forney_SM = Forney_s5) then\n") ;
           fprintf(stream, "            R_k_lam <= R_k_lam xor k_lam ;\n") ;
           fprintf(stream, "        end if ;\n") ;
           fprintf(stream, "    end if ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "end process ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "-- Store  O(x)/L'(x)\n") ; 
           fprintf(stream, "--\n") ;
           fprintf(stream, "process(rst_n, clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n = '0') then\n") ;
           fprintf(stream, "           R_Om <= (OTHERS => '0') ;\n") ;
           fprintf(stream, "    elsif (clk'event and clk = '1')then\n") ;
           fprintf(stream, "        if (Forney_SM = Forney_s3 or Forney_SM = Forney_s7 ) then\n") ;
           fprintf(stream, "            R_Om <= (OTHERS => '0') ;\n") ;
           fprintf(stream, "        elsif (Forney_SM = Forney_s6) then\n") ;
           fprintf(stream, "            R_Om <= om_lam_calc ;\n") ;
           fprintf(stream, "        end if ;\n") ;
           fprintf(stream, "    end if ;\n") ;
           fprintf(stream, "end process ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    K0 <=  R_k_om when (Sel_F = '0') else norm ;\n") ;
           fprintf(stream, "    K1 <=  inv_k_lam when (Sel_F = '0') else R_Om ;\n") ;
           fprintf(stream, "\n") ;
           if (base > 1) {
               fprintf(stream, "    pow_alpha_rom_adrs <= gf_alpha_rom_adrs ;\n") ;
               fprintf(stream, "    norm <= pow_alpha ;\n") ;
		   } else {
               fprintf(stream, "    norm <= \"") ;
			   for(i=0; i<gf_m-1; i++) fprintf(stream,"0") ;
			   fprintf(stream,"1\" ;\n") ;
		   }
           fprintf(stream, "\n") ;
           fprintf(stream, "    inv_alpha_rom_adrs <= R_k_lam ;\n") ;
           fprintf(stream, "    inv_k_lam <= gf_inv_alpha ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    om_lam_calc <= gfmul_ab(K0, K1) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "-- Store Ei values to correct message\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "process(rst_n, clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n = '0') then\n") ;
           for (i=0; i<chk_sym; i++)
               fprintf(stream, "           R_Ei_%d <= (OTHERS => '0') ;\n", i) ;
           fprintf(stream, "    elsif (clk'event and clk = '1')then\n") ;
           fprintf(stream, "        if (Forney_SM = Forney_s7) then\n") ;
           fprintf(stream, "            case n_root_cnt is\n") ;
           for (i=0; i < chk_sym; i++) {
               fprintf(stream, "                when \"") ;
               for (j=0; j <gf_m; j++) fprintf(stream, "0") ;
               fprintf(stream, "\" =>\n") ;      
               fprintf(stream, "                   R_Ei_%d <= om_lam_calc ;\n", i) ;
               fprintf(stream, "\n") ;
           }
           fprintf(stream, "                when OTHERS =>\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "            end case ;\n") ;
           fprintf(stream, "        end if ;\n") ;
           fprintf(stream, "    end if ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "end process ;\n") ;
           for (i=0; i<chk_sym; i++)
               fprintf(stream, "    corr_err_val%d <= R_Ei_%d ;\n", i, i) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "-- Indicate correction values ready\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "process(rst_n, clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n = '0') then\n") ;
           fprintf(stream, "           Decode_Done <= '0' ;\n") ;
           fprintf(stream, "    elsif (clk'event and clk = '1')then\n") ;
           fprintf(stream, "        if (Forney_SM = Forney_s8) then\n") ;
           fprintf(stream, "           Decode_Done <= '1' ;\n") ;
           fprintf(stream, "        else\n") ;
           fprintf(stream, "           Decode_Done <= '0' ;\n") ;
           fprintf(stream, "        end if ;\n") ;
           fprintf(stream, "    end if ;\n") ;
           fprintf(stream, "end process ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    forney_done <= Decode_Done ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "end ;\n") ;
           fprintf(stream, "\n") ;
           break ;
       default :
           break ;
   }
}

