/*******************************************************************************/
/*                                                                             */
/*                                                                             */
/*                      gen_rs_rom.c                                           */
/*                                                                             */
/*                                                                             */
/*                                                                             */
/*  Decription:                                                                */
/*                                                                             */
/*       Generate ROM images for FPGAs                                         */
/*                                                                             */
/*                                                                             */
/*                                                                             */
/*    Copyright (C) 2003  John S. Willy (jswilly@pacbell.net)                  */ 
/*                                                                             */
/*    This program is free software; you can redistribute it and/or modify     */
/*    it under the terms of the GNU General Public License as published by     */
/*    the Free Software Foundation; either version 2 of the License, or        */
/*    (at your option) any later version.                                      */
/*                                                                             */
/*    This program is distributed in the hope that it will be useful,          */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*    GNU General Public License for more details.                             */
/*                                                                             */
/*    You should have received a copy of the GNU General Public License        */
/*    along with this program; if not, write to the Free Software              */
/*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA*/
/*******************************************************************************/
 
#include <stdio.h>
#include "rsccg_proto.h"
//
//
#define VHDL     0 
#define VLOG     1 
#define C        2 

#define X_UCF    0 
#define X_COE    1 
#define ALT_MIF  2 
//
// Work space arrays
//
extern char poly_table[256] ;      // a^n table         
extern char inv_poly_table[256] ;  // 1/a^n table      
extern char pow_poly_table[256] ;  // (a^n)^basis table 
extern char log_poly_table[256] ;  // n = log(a^n)     
//
// Generate ROM code & Image data
//
void gen_rs_rom_data(FILE *stream, unsigned char *tbl, unsigned char *ref, int mem_type,  int gf_field_size)
{

   int n, l_cnt, n_lines ;

    // Create memory initialization files for simulation or FPGA compiler
    switch (mem_type) {
        case C:
            n = 0 ;
			n_lines = gf_field_size/16 ;

            fprintf (stream, "// Reed Solomon GF table\n") ;
            fprintf (stream,"unsigned char ") ;
            fprintf (stream,"%s[%d] = {\n", ref, gf_field_size) ;

            while (n_lines > 0){

	            if (n_lines > 1) {
	                for (l_cnt=0; l_cnt<16; l_cnt++) fprintf(stream,"0x%.2x,",tbl[n+l_cnt]);
	                    fprintf(stream,"\n") ;
		                n_lines -= 1 ;
						n += 16 ;
				} else if (n_lines == 1) {
	            for (l_cnt=0; l_cnt<15; l_cnt++) fprintf(stream,"0x%.2x,",tbl[n+l_cnt]);
		            fprintf(stream,"0x%.2x} ;",tbl[n+l_cnt]);
		            n_lines -= 1 ;
					n += 16 ;
		        }
			}
            fprintf(stream, "\n") ;
            break ;

        case VHDL:
            if (mem_type == VHDL) {
	            for (n=0; n< gf_field_size; n++) {
	                fprintf(stream, "16#") ;
                    fprintf(stream,"%x",tbl[n]);
	                fprintf(stream, "#\n") ;

				}
			}
		     break ;

		case VLOG:

            if (mem_type == VLOG) {
	            for (n=0; n< gf_field_size; n++) {
                    fprintf(stream,"%x,",tbl[n]);
                    fprintf(stream, "\n") ;
				}
             break ;
			}
	}
}
//
// Generate simulation or synthesis models for simulation or FPGA compiler
//
void gen_rs_rom_rtl(FILE *stream, int mem_type)
{
    switch (mem_type) {
        case C:
			break ;

		case VHDL:
            fprintf(stream, "------------------------------------------------------------------------------------\n") ;
            fprintf(stream, "--\n") ;	 
            fprintf(stream, "--\n") ;
            fprintf(stream, "-- File: sim_spram.vhd\n") ;
            fprintf(stream, "--\n") ;
            fprintf(stream, "--\n") ;
            fprintf(stream, "------------------------------------------------------------------------------------\n") ;
            fprintf(stream, "--\n") ;
            fprintf(stream, "library ieee;\n") ;
            fprintf(stream, "use ieee.std_logic_1164.all;\n") ;
            fprintf(stream, "--\n") ;
            fprintf(stream, "library ieee;\n") ;
            fprintf(stream, "use ieee.std_logic_unsigned.all;\n") ;
            fprintf(stream, "--\n") ;
            fprintf(stream, "library ieee;\n") ;
            fprintf(stream, "use std.textio.all;\n") ;
            fprintf(stream, "--\n") ;
            fprintf(stream, "library arithmetic ;\n") ;
            fprintf(stream, "use arithmetic.std_logic_arith.all ;\n") ;
            fprintf(stream, "--\n") ;
            fprintf(stream, "library io_lib ;\n") ;
            fprintf(stream, "use io_lib.io_utils.all;\n") ;
            fprintf(stream, "--\n") ;
            fprintf(stream, "entity spram_block is\n") ;
            fprintf(stream, "  generic ( RAM_NUM_WORDS : integer ;\n") ;
            fprintf(stream, "            RAM_WIDTH_A   : integer ;\n") ;
            fprintf(stream, "            RAM_WIDTH_D   : integer ;\n") ;
            fprintf(stream, "			  RAM_FILE      : string \n") ;
            fprintf(stream, "		   );\n") ;
            fprintf(stream, "  port (clk  : in std_logic;\n") ;
            fprintf(stream, "        we   : in std_logic;\n") ;
            fprintf(stream, "        addr : in std_logic_vector(RAM_WIDTH_A-1 downto 0);\n") ;
            fprintf(stream, "        di   : in std_logic_vector(RAM_WIDTH_D-1 downto 0);\n") ;
            fprintf(stream, "        do   : out std_logic_vector(RAM_WIDTH_D-1 downto 0)\n") ;
            fprintf(stream, "		);\n") ;
            fprintf(stream, "end spram_block;\n") ;
            fprintf(stream, "--\n") ;
            fprintf(stream, "architecture syn of spram_block is\n") ;
            fprintf(stream, "\n") ;
            fprintf(stream, "  attribute ram_style : string ;\n") ;
            fprintf(stream, "\n") ;
            fprintf(stream, "  type ram_type is array (0 to RAM_NUM_WORDS-1)\n") ;
            fprintf(stream, "        of std_logic_vector (RAM_WIDTH_D-1 downto 0);\n") ;
            fprintf(stream, "\n") ;
            fprintf(stream, "  signal RAM : ram_type;\n") ;
            fprintf(stream, "  attribute ram_style of RAM : signal is \"block\";\n") ;
            fprintf(stream, "\n") ;
            fprintf(stream, "  signal read_adr : std_logic_vector(RAM_WIDTH_A-1 downto 0) ;\n") ;
            fprintf(stream, "\n") ;
            fprintf(stream, "  signal ram_a    : integer := 0 ;\n") ;
            fprintf(stream, "--  signal r_data   : std_logic_vector(RAM_WITDH-1 downto 0) ;\n") ;
            fprintf(stream, "\n") ;
            fprintf(stream, "  signal mem_init : integer := 0 ;\n") ;
            fprintf(stream, "\n") ;
            fprintf(stream, "begin\n") ;
            fprintf(stream, "\n") ;
            fprintf(stream, "--process (clk)\n") ;
            fprintf(stream, "--  variable outline  : line;\n") ;
            fprintf(stream, "--begin\n") ;
            fprintf(stream, "--    if (clk'event and clk = '1') then\n") ;
            fprintf(stream, "--      if (we = '1') then\n") ;
            fprintf(stream, "--        RAM(conv_integer(addr)) <= di;\n") ;
            fprintf(stream, "--            write_string(outline, \"what???\");\n") ;
            fprintf(stream, "--      end if;\n") ;
            fprintf(stream, "--      read_adr <= addr ;\n") ;
            fprintf(stream, "--    end if;\n") ;
            fprintf(stream, "--end process;\n") ;
            fprintf(stream, "--\n") ;
            fprintf(stream, "----   do <= RAM(conv_integer(read_adr)) ;\n") ;
            fprintf(stream, "--\n") ;
            fprintf(stream, "-- Xilinx RAM simulation model\n") ;
            fprintf(stream, "-- For use with VHDL simulator.\n") ;
            fprintf(stream, "--\n") ;
            fprintf(stream, "-- Note the use of variables in the procedure during initialization.\n") ;
            fprintf(stream, "-- This is required since the <= operator is used for concurrent assignment\n") ;
            fprintf(stream, "-- and the variable operator := is scheduled for immediate assignment.\n") ;
            fprintf(stream, "-- If the variable is needed outside of the scope of the process\n") ;
            fprintf(stream, "-- it is assigned using the <= operator.\n") ;
            fprintf(stream, "-- \n") ;
            fprintf(stream, "process(clk)\n") ;
            fprintf(stream, "  variable val      : integer;\n") ;
            fprintf(stream, "  file     myfile   : text IS  RAM_FILE;\n") ;
            fprintf(stream, "  variable inline   : line;\n") ;
            fprintf(stream, "  variable outline  : line;\n") ;
            fprintf(stream, "  variable mem_init_intl : integer := 0 ;\n") ;
            fprintf(stream, "  variable ram_a_intl    : integer := 0 ;\n") ;
            fprintf(stream, "  variable r_data_intl   : std_logic_vector(RAM_WIDTH_D-1 downto 0) ;\n") ;
            fprintf(stream, "begin\n") ;
            fprintf(stream, "   if (mem_init_intl = 0 and not endfile(myfile)) then\n") ;
            fprintf(stream, "            readline(myfile, inline);\n") ;
            fprintf(stream, "\n") ;
            fprintf(stream, "            write(outline, inline.ALL);\n") ;
            fprintf(stream, "\n") ;
            fprintf(stream, "            read_based(inline, val);\n") ;
            fprintf(stream, "            write_string(outline, \" hex: \");\n") ;
            fprintf(stream, "            write(outline, val, base => hex);\n") ;
            fprintf(stream, "            writeline(output, outline);\n") ;
            fprintf(stream, "\n") ;
            fprintf(stream, "			r_data_intl := conv_std_logic_vector(val,RAM_WIDTH_D) ;\n") ;
            fprintf(stream, "\n") ;
            fprintf(stream, "			RAM(ram_a_intl) <= r_data_intl ;\n") ;
            fprintf(stream, "\n") ; 
            fprintf(stream, "            ram_a_intl := ram_a_intl + 1 ;\n") ;
            fprintf(stream, "\n") ;
            fprintf(stream, "			ram_a <= ram_a_intl ;\n") ;
            fprintf(stream, "            if (ram_a_intl = RAM_NUM_WORDS) then \n") ;
            fprintf(stream, "			    mem_init_intl := 1 ;\n") ;
            fprintf(stream, "			end if ;\n") ;
            fprintf(stream, "			mem_init <= mem_init_intl ;\n") ;
            fprintf(stream, "--\n") ;  
            fprintf(stream, "--  Memory model is defined here after the initialization\n") ;
            fprintf(stream, "--\n") ;
            fprintf(stream, "    elsif (clk'event and clk = '1') then\n") ;
            fprintf(stream, "      if (we = '1') then\n") ;
            fprintf(stream, "        RAM(conv_integer(addr)) <= di;\n") ;
            fprintf(stream, "       end if;\n") ;
            fprintf(stream, "      read_adr <= addr after 1 ns ;\n") ;
            fprintf(stream, "    end if;\n") ;
            fprintf(stream, "\n") ;
            fprintf(stream, "end process;\n") ;
            fprintf(stream, "\n") ;
            fprintf(stream, "   do <= RAM(conv_integer(read_adr)) after 1 ns ;\n") ;
            fprintf(stream, "\n") ;
            fprintf(stream, "end syn;\n") ;
            fprintf(stream, "\n") ;
			break ;
	 }
}

//
// Generate ROM initialization files for Altera or Xilinx.
//     Three initialization files type are created:
//          Xilinx UCF initialization, used if COREGEN is not available
//          Xilinx COE file initialization, used if COREGEN for RAM is availabe
//          Altera MIF file initialization, used for Altera RAM 
//
void gen_rom_init(FILE *stream, unsigned char *tbl, unsigned char *ref, int mem_init, int gf_field_size)
{
    int i, j ;
//
// Xilinx UCF initialization
//
	if( mem_init == X_UCF) {
        fprintf(stream,"#\n") ;
        fprintf(stream,"#UCF RAM Initialization - %s\n", ref) ;
        fprintf(stream,"#\n") ;
	    for(i=0; i<8 ; i++) {
            fprintf(stream,"INST instance_label/mram_ram_inst_ramb_0\n") ;
		    fprintf(stream,"INIT_%.2x=\"", i) ;
		    for (j=0; j<32; j++) fprintf(stream,"%.2x",tbl[((32*(i+1))-1)-j]) ;
		    fprintf(stream,"\" ;\n") ;
		}

	    for(i=8; i<16 ; i++) {
		    fprintf(stream,"INIT_%.2x=\"", i) ;
		    for (j=0; j<32; j++) fprintf(stream,"00") ;
		    fprintf(stream,"\" ;\n") ;
		}
		return ;
	}

//
// Altera MIF file Initialization
//
	if( mem_init == ALT_MIF) {
	    fprintf(stream,"DEPTH = 256 ;\n") ;
	    fprintf(stream,"WIDTH = 8 ;\n") ;
	    fprintf(stream,"ADDRESS_RADIX = DEC ;\n") ;
	    fprintf(stream,"DATA_RADIX = HEX ;\n") ;
	    fprintf(stream,"CONTENT\n") ;
	    fprintf(stream,"    BEGIN\n") ;
	        for(i=0; i<gf_field_size; i++) fprintf(stream,"%d : %2x ;\n", i, tbl[i]) ;
	        for(i=gf_field_size; i<256; i++) fprintf(stream,"%d : 00 ;\n", i) ;

	            fprintf(stream,"END ;\n") ;
		return ;
	}
	    
	
//
// Xilinx COE File Initialization
//
	if( mem_init == X_COE) {
        fprintf(stream,"coefficients %s\n", ref) ;
        fprintf(stream,"Component_Name=<your component label>\n") ;
        fprintf(stream,"Depth = 512\n") ;
        fprintf(stream,"Data_Width = 8\n") ;
        fprintf(stream,"Radix = 16\n") ;
        fprintf(stream,"Default_Data = 00\n") ;
  	    fprintf(stream,"MEMORY_INITIALIZATION_VECTOR = \n") ;
	    for(i=0; i<gf_field_size; i++) fprintf(stream,"%.2x,\n", tbl[i]) ;
	    for(i=gf_field_size; i<512; i++) fprintf(stream,"%d : 00 ;\n", i) ;

        fprintf(stream,"%.2x ;\n", tbl[i]) ;
		return ;
	}
	return ;
}

