/*******************************************************************************/
/*                                                                             */
/*            gen_syndec.c                                                     */
/*                                                                             */
/*    Generate RS Receiver Syndrome Check                                      */
/*                                                                             */
/*                                                                             */
/*                                                                             */
/* Copyright (C) 2003 John S. Willy                                            */
/*                                                                             */
/*                                                                             */
/*                                                                             */
/*                                                                             */
/*    Copyright (C) 2003  John S. Willy (jswilly@pacbell.net)                  */ 
/*                                                                             */
/*    This program is free software; you can redistribute it and/or modify     */
/*    it under the terms of the GNU General Public License as published by     */
/*    the Free Software Foundation; either version 2 of the License, or        */
/*    (at your option) any later version.                                      */
/*                                                                             */
/*    This program is distributed in the hope that it will be useful,          */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*    GNU General Public License for more details.                             */
/*                                                                             */
/*    You should have received a copy of the GNU General Public License        */
/*    along with this program; if not, write to the Free Software              */
/*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA*/
/*******************************************************************************/
 
#include <stdio.h>

#define VHDL     0 
#define VLOG     1 
#define C        2 

extern unsigned char poly_table[] ;   // RS code table 
extern unsigned char log_poly_table[] ; 
extern unsigned int p_wsp ;           // RS Generator Polynomial

/* locals  */

void gen_rs_syndec(FILE *stream, int base, int gf_m, int deg, int hdl_type)
{
    int i ;

   switch (hdl_type) {
       case C:
           fprintf(stream, "//--------------------------------------------------------------------//\n") ;
           fprintf(stream, "//                  Generated by RSTK                                 //\n") ;  
           fprintf(stream, "//                      rs_syndec.c                                   //\n") ;  
           fprintf(stream, "//                                                                    //\n") ;  
           fprintf(stream, "//--------------------------------------------------------------------//\n") ;
           fprintf(stream, "#include <stdio.h>\n") ;
           fprintf(stream, "#include \"proto_types.h\"\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "extern unsigned char msg_in[] ;\n") ;
           fprintf(stream, "extern unsigned char enc_msg_out[] ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "void dec_syn_calc( n,\n") ;
           for(i=1; i <= deg; i++)
               fprintf(stream, "          s%d,\n", i) ;
           fprintf(stream, "syn_zero )\n") ;
           fprintf(stream, "int  n ;\n") ; 
           fprintf(stream, "unsigned char *s1,\n") ;           
           for(i=2; i < deg; i++)
               fprintf(stream, "          *s%d, \n", i) ;
           fprintf(stream, "          *s%d ;\n", i) ;
           fprintf(stream, "unsigned char *syn_zero ;\n") ;
           fprintf(stream, "{\n") ;
           fprintf(stream, "int i ;\n") ;
           fprintf(stream, "unsigned char s1_r,\n") ;
           for(i=2; i < deg; i++)
               fprintf(stream, "          s%d_r, \n", i) ;
           fprintf(stream, "          s%d_r ;\n", i) ;
           fprintf(stream, "unsigned char d_in ;\n") ;
           for(i=1; i <= deg; i++)
               fprintf(stream, "   s%d_r = 0 ;\n", i) ;

           fprintf(stream, "  for (i=0; i<n; i++) {\n") ;
           fprintf(stream, "    d_in = enc_msg_out[i] ;\n") ;
           for(i=0; i < deg; i++) {
                 fprintf(stream, "    s%d_r = gfmul_%d(s%d_r) ^ d_in ;\n", i+1, base+i, i+1) ;
           }
           fprintf(stream, "  }\n") ;
           fprintf(stream, "        // Output Syndromes\n") ;
           for(i=1; i <= deg; i++) 
               fprintf(stream, "    *s%d = s%d_r ;\n",i,i) ;

           fprintf(stream, "    /* If  (S1+S2+...S5) equals zero then no error occured */\n") ;
           fprintf(stream, "  *syn_zero = s1_r | \n") ;
           for(i=2; i < deg; i++)
           fprintf(stream, "              s%d_r |\n", i) ;
           fprintf(stream, "              s%d_r ;\n", i) ;
           fprintf(stream, "}\n") ;
           break;

       case VLOG : 
           fprintf(stream, "//--------------------------------------------------------------------//\n") ;
           fprintf(stream, "//                  Generated by RSTK                                 //\n") ;  
           fprintf(stream, "//                      rs_syndec.v                                   //\n") ;  
           fprintf(stream, "//                                                                    //\n") ;  
           fprintf(stream, "//--------------------------------------------------------------------//\n") ;
           fprintf(stream, "// Syndrome calculation:\n") ;

           fprintf(stream, "module syn_calc( rst_n,\n") ;
           fprintf(stream, "           init,\n") ;
           fprintf(stream, "           clk,\n") ;
           fprintf(stream, "           enb_dec,\n") ;
           fprintf(stream, "           d_in,\n") ;
           for(i=1; i <= deg; i++)
               fprintf(stream, "           s%d,\n", i) ;
           fprintf(stream, "         ) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    input        rst_n ;\n") ;
           fprintf(stream, "    input        clk ;\n") ;
           fprintf(stream, "    input        init ;\n") ;
           fprintf(stream, "    input        enb_dec ;\n") ;
           fprintf(stream, "    input[%d:0]    d_in ;\n", gf_m-1) ;
           for(i=1; i <= deg; i++)
               fprintf(stream, "    output[%d:0]    s%d ;\n", gf_m-1, i) ;
           fprintf(stream, "    output          syn_zero ;\n") ;
           fprintf(stream, "\n") ;        
           fprintf(stream, "\n") ;
           fprintf(stream, "// Locals\n") ;
           for(i=1; i <= deg; i++)
               fprintf(stream, "    reg[%d:0]    s%d_r ;\n", gf_m-1, i) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "always @(negedge rst_n or posedge clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n == 1'b0)\n") ;
           fprintf(stream, "    begin\n") ;
           for(i=1; i <= deg; i++)
           fprintf(stream, "        s%d_r <= 0 ;\n", i) ;
           fprintf(stream, "        end\n") ;
           fprintf(stream, "    else if (init == 1'b1)\n") ; 
           fprintf(stream, "    begin\n") ;
           for(i=1; i <= deg; i++)
           fprintf(stream, "           s%d_r <= 0 ;\n", i) ;
           fprintf(stream, "        end\n") ;
           fprintf(stream, "    else if (enb_dec == 1'b1 )\n") ;
           for(i=0; i <= deg-1; i++) 
           fprintf(stream, "           s%d_r <= gfmul_%d(s%d_r) ^ d_in ;\n", i+1, base+i, i) ;          fprintf(stream, "    begin\n") ;
           fprintf(stream, "    end\n") ;
           fprintf(stream, "end\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    // Output Syndromes\n") ;
           for(i=1; i <= deg; i++)
           fprintf(stream, "    assign #1 s%d = s%d_r ;\n", i, i) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    // If  (S1+S2+...S5) equals zero then no error occured \n") ;
           fprintf(stream, "    assign #1 syn_zero = (s1_r  |\n") ;
           for(i=2; i <= deg-1; i++)
           fprintf(stream, "                         s%d_r |\n", i) ;
           fprintf(stream, "                         s%d_r );\n", i) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "endmodule \n") ;
           break;

       case VHDL :
           fprintf(stream, "------------------------------------------------------------------------\n") ;
           fprintf(stream, "--                  Generated by RSTK                                 --\n") ;  
           fprintf(stream, "--                      rs_encode.vhd                                  --\n") ;  
           fprintf(stream, "--                                                                    --\n") ;  
           fprintf(stream, "------------------------------------------------------------------------\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "LIBRARY IEEE ;\n") ;
           fprintf(stream, "use IEEE.std_logic_1164.all ;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "library IEEE ;\n") ;
           fprintf(stream, "use     IEEE.std_logic_unsigned.all;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "library IEEE ;\n") ;
           fprintf(stream, "use     IEEE.NUMERIC_STD.all;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "library gf_arith ;\n") ;
           fprintf(stream, "use     gf_arith.gf_arith.all;\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "--\n") ;
           fprintf(stream, "entity dec_syn_calc is port( rst_n   : IN std_logic ;\n") ;
           fprintf(stream, "                             init    : IN std_logic ;\n") ;
           fprintf(stream, "                             clk     : IN std_logic ;\n") ;
           fprintf(stream, "                             enb_dec : IN std_logic ;\n") ;
           fprintf(stream, "                             d_in : IN std_logic_vector(%d downto 0) ;\n", gf_m-1) ;
           for(i=1; i <= deg; i++)
               fprintf(stream, "                             s%d   : OUT std_logic_vector(%d downto 0) ;\n", i, gf_m-1) ;
           fprintf(stream, "                             syn_zero : OUT std_logic \n") ;
           fprintf(stream, "                             ) ;\n") ;
           fprintf(stream, "end dec_syn_calc ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "architecture rtl of dec_syn_calc is\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "-- Local signals\n") ;
           fprintf(stream, "\n") ;
           for(i=1; i <= deg; i++)  
               fprintf(stream, "    signal    s%d_r : std_logic_vector(%d downto 0);\n", i, gf_m-1) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "process(rst_n ,clk)\n") ;
           fprintf(stream, "begin\n") ;
           fprintf(stream, "    if (rst_n = '0') then\n") ;
           for(i=1; i <= deg; i++)  
               fprintf(stream, "           s%d_r <= (OTHERS => '0') ;\n", i) ;
           fprintf(stream, "    elsif (clk'event and clk ='1') then\n") ;
           fprintf(stream, "        if (init = '1') then\n") ;
           for(i=1; i <= deg; i++)
               fprintf(stream, "           s%d_r <= (OTHERS => '0') ;\n", i) ;
           fprintf(stream, "        elsif (enb_dec = '1') then\n") ;
           for(i=0; i <= deg-1; i++) 
               fprintf(stream, "           s%d_r <= gfmul_%d(s%d_r) xor d_in ;\n", i+1, base+i, i+1) ;
           fprintf(stream, "        end if ;\n") ;
           fprintf(stream, "    end if ;\n") ;
           fprintf(stream, "end process ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    -- Output Syndromes\n") ;
           fprintf(stream, "\n") ;
           for(i=1; i <= deg; i++)
               fprintf(stream, "   s%d <= s%d_r ;\n", i, i) ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    -- If  (S1+S2+...S5) equals zero then no error occured \n") ;
           fprintf(stream, "  syn_zero <= '1' when (s1_r=0 and \n") ;
           for(i=2; i < deg; i++)
               fprintf(stream, "                        s%d_r=0 and\n", i) ;
           fprintf(stream, "                        s%d_r=0 )  \n", i) ;           
           fprintf(stream, "                  else '0';\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "END ;\n") ;
           fprintf(stream, "\n") ;
           break ;
       default :
           break ;
   }
}

