/*******************************************************************************/
/*                                                                             */
/*            gf_arith.c                                                       */
/*                                                                             */
/*                                                                             */
/* Copyright (C) 2003 John S. Willy                                            */
/*                                                                             */
/*                                                                             */
/*   gf_mul(a,b)  Multiply A*B                                                 */ 
/*   gf_add(a,b)  Add A+B                                                      */
/*                                                                             */
/*  Comments:                                                                  */
/*      Currently multiply and add are up to 8 bits only                       */
/*                                                                             */
/*    Copyright (C) 2003  John S. Willy (jswilly@pacbell.net)                  */ 
/*                                                                             */
/*    This program is free software; you can redistribute it and/or modify     */
/*    it under the terms of the GNU General Public License as published by     */
/*    the Free Software Foundation; either version 2 of the License, or        */
/*    (at your option) any later version.                                      */
/*                                                                             */
/*    This program is distributed in the hope that it will be useful,          */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*    GNU General Public License for more details.                             */
/*                                                                             */
/*    You should have received a copy of the GNU General Public License        */
/*    along with this program; if not, write to the Free Software              */
/*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA*/
/*******************************************************************************/
 
#include <stdio.h>
#define   N_BITS    8 ;

/* Multiply values  */
unsigned short p ;
unsigned short q ;
//
//
// gf_mul(a,b, gf_poly)
//
// Multiply two feild values a*b using feild generator polynomial
// represented as an integer 
//
// inputs:
//      a   : Field value A
//      b   : Field value B
//   gf_poly: Field generator polynomial repesented as an interger
//      gf_m: Feild size (2^n, i.e GF(2^8) gf_m=8
//
// returns:
//      a*b
//     
unsigned char gf_mul(unsigned char a, unsigned char b, int gf_poly,  int gf_m)
{

int i ;
unsigned short m, bit_mask ;

  m = (unsigned int)a ;
  q = 0 ;
  p = 0 ;
  bit_mask = 1 ; 
  //printf("gf_m = %d\n", gf_m) ;

  // Multiply a * b
  for (i=0; i < gf_m; i++) {
     if (b & 0x01) q = q ^ m ;
	 m = m << 1 ;
	 b = b >> 1 ;
//     printf("q = %.x \n", q) ;	 
  }

//  printf("multiply q = %.x \n", q) ;

// 
// Shift until a '1' is found then subtract polynomial
// Do this until length of 16 - polynomial length is true
// For poly = 391 << 7 this is at most 7 shifts
//
  p = gf_poly << (gf_m-1) ;
  bit_mask = bit_mask << (2*gf_m-1) ;
  //printf("bit_mask = %x\n", bit_mask ) ;

  i =0 ;
  while (i < gf_m) {
      if ((q & bit_mask) == 0) {
		  q = q << 1 ;
	  	  //printf("test for 0 bit in q = %.x \n",  q) ;
		  i = i + 1 ;
	  }
	  else {
		  q = q ^ p ;
	      //printf("i = %d q = %.x \n", i, q) ;
	  }
  }  

  // Normalize back to gf_m bits 
  q = q >> gf_m ;

//  printf("q = %.x \n", q) ;
  return( (unsigned char)q );
}

//
// gf_add(a,b)
//     Add a + b 
//
unsigned char gf_add(unsigned char a, unsigned char b)
{
    return (a ^ b) ;
}

