/*******************************************************************************/
/*                                                                             */
/*            poly_math.c                                                      */
/*                                                                             */
/*    Polynomial add/multiply                                                  */
/*                                                                             */
/*                                                                             */
/* Copyright (C) 2003 John S. Willy                                            */
/*                                                                             */
/*                                                                             */
/*  input:                                                                     */
/*      Pointer to polynomial 1, Degree Polynomial 1                           */
/*      Pointer to polynomial 2, Degree Polynomial 2                           */
/*  output:                                                                    */
/*      Polynomial 1 = P1 * P2                                                 */
/*                                                                             */
/*  returns:                                                                   */
/*      Degree of Polynomial 1                                                 */
/*                                                                             */
/*    Copyright (C) 2003  John S. Willy (jswilly@pacbell.net)                  */
/*                                                                             */
/*    This program is free software; you can redistribute it and/or modify     */
/*    it under the terms of the GNU General Public License as published by     */
/*    the Free Software Foundation; either version 2 of the License, or        */
/*    (at your option) any later version.                                      */
/*                                                                             */
/*    This program is distributed in the hope that it will be useful,          */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*    GNU General Public License for more details.                             */
/*                                                                             */
/*    You should have received a copy of the GNU General Public License        */
/*    along with this program; if not, write to the Free Software              */
/*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA*/
/*******************************************************************************/
/**************************************************************/
#include <stdio.h>
#include "rsccg_proto.h"
//
// Externals
//
extern unsigned char poly_table[] ;        // Alpha GF Feild (a^n)
extern unsigned char log_poly_table[] ;    // Log(alpha), this table look up returns alpha from n
extern unsigned int p_wsp[256] ;           // Polynomial work space
extern unsigned int p1[2] ;                // Initial polynomial is a^0 + 1*x^1

//
// Workspace
//
unsigned int tmp1[64] ;
unsigned int tmp2[64] ;

int gf_poly_mul(unsigned int *p1, int deg_p1, unsigned int *p2, int deg_p2, int gf_poly, int gf_m)
{

int i ;
int term ;
int deg ;
int new_deg ;  

  for (i=0; i<= 63; i++) {
    tmp1[i] = 0 ;
    tmp2[i] = 0 ;
  }

  new_deg = deg_p1 + deg_p2 ;

  term = 0 ;

  while (term <= deg_p1) {
     i = 0 ;
	 deg = 0 ;
	 while (i <= deg_p2) {
		 tmp1[i+term] = gf_mul((unsigned char)p2[i],(unsigned char)p1[term], gf_poly, gf_m) ;
		 i = i + 1 ;
		 deg = deg + 1 ;
     }

	 i = 0 ;
	 while (i <= deg) {
		 tmp2[i+term] = gf_add((unsigned char)tmp1[i+term], (unsigned char)tmp2[i+term]) ;
	     i = i + 1 ;
	 }
     tmp1[term] = 0 ;
	 term = term  + 1 ;
	 deg = 0 ;
	}

  for(i=0; i<= new_deg; i++) {
    p1[i] = tmp2[i] ;
  }
  return(new_deg) ;
}
//
//
//  Expand code generator polynomial (x-a^base)(x-a^base+1)...
//
//
int expand_gen_poly(int gf_field_poly, int num_chk_sym, int base, int gf_m, int gf_field)
{
int alpha ;
int r_deg ;
int n_terms ;
int tmp_base, i, j;

	tmp_base = base ; // Keep base and use tmp_base as work variable

	n_terms = 2*num_chk_sym - 1 ;

	for (i=0; i<256; i++) p_wsp[i] = 0 ;

	p_wsp[1] = 1 ;
	p_wsp[0] = poly_table[tmp_base] ;
	tmp_base = tmp_base + 1 ;
	p1[0] = poly_table[tmp_base] ; 
    r_deg = 1 ;

	for (j=0; j<n_terms; j++) {
	    r_deg = gf_poly_mul(p_wsp, r_deg, p1, 1, gf_field_poly, gf_m) ;
		tmp_base = tmp_base + 1 ;
		p1[0] = poly_table[tmp_base] ;
    }
	//printf("R_DEG = %d\n", r_deg) ;
	for (i=0; i <= r_deg; i++) {
		alpha = log_poly_table[p_wsp[i]] % (gf_field-1) ;
        printf("g(%d) = a^%d  pwsp[%d] = %x\n ", i, alpha, i, p_wsp[i]) ;
	}
	printf("\n") ;

	return (r_deg) ;
}


