/*******************************************************************************/
/*                                                                             */
/*            rs_c_util.c                                                      */
/*                                                                             */
/*    Polynomial add/multiply utilities for 'C' RS code Gen                    */
/*                                                                             */
/*                                                                             */
/*                                                                             */
/*                                                                             */
/*    Copyright (C) 2003  John S. Willy (jswilly@pacbell.net)                  */
/*                                                                             */
/*    This program is free software; you can redistribute it and/or modify     */
/*    it under the terms of the GNU General Public License as published by     */
/*    the Free Software Foundation; either version 2 of the License, or        */
/*    (at your option) any later version.                                      */
/*                                                                             */
/*    This program is distributed in the hope that it will be useful,          */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*    GNU General Public License for more details.                             */
/*                                                                             */
/*    You should have received a copy of the GNU General Public License        */
/*    along with this program; if not, write to the Free Software              */
/*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA*/
/*******************************************************************************/
#include <stdio.h>
#include <string.h>

void gen_rs_cutil(FILE *stream)
{
           fprintf(stream, "#include <stdio.h>\n") ;
           fprintf(stream, "#include \"proto_types.h\"\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "// Externals\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "extern unsigned char gf_alpha_tbl[] ;        // Alpha GF Feild (a^n)\n") ;
           fprintf(stream, "extern unsigned char gf_invalpha_tbl[] ;     // Iverse of Alpha GF Feild (a^n)\n") ;
           fprintf(stream, "#define MAX_POLY_ORDER   64 \n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "// Workspace\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "unsigned char poly_p1[64] ;\n") ;
           fprintf(stream, "unsigned char poly_p2[64] ;\n") ;
           fprintf(stream, "unsigned char poly_p3[64] ;\n") ;
           fprintf(stream, "unsigned int tmp1[64] ;\n") ;
           fprintf(stream, "unsigned int tmp2[64] ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "void gf_poly_mul(unsigned char *p1, int deg_p1, unsigned char *p2, int deg_p2, unsigned char *p3, int *deg_p3)\n") ;
           fprintf(stream, "{\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "int i ;\n") ;
           fprintf(stream, "int term ;\n") ;
           fprintf(stream, "int deg ;\n") ;
           fprintf(stream, "int new_deg ;\n") ;  
           fprintf(stream, "\n") ;
           fprintf(stream, "  for (i=0; i<MAX_POLY_ORDER; i++) {\n") ;
           fprintf(stream, "    tmp1[i] = 0 ;\n") ;
           fprintf(stream, "    tmp2[i] = 0 ;\n") ;
           fprintf(stream, "  }\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "  new_deg = deg_p1 + deg_p2 ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "  term = 0 ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "  while (term <= deg_p1) {\n") ;
           fprintf(stream, "     i = 0 ;\n") ;
           fprintf(stream, "	 deg = 0 ;\n") ;
           fprintf(stream, "	 while (i <= deg_p2) {\n") ;
           fprintf(stream, "		 tmp1[i+term] = gfmul_ab((unsigned char)p2[i],(unsigned char)p1[term]) ;\n") ;
           fprintf(stream, "		 i = i + 1 ;\n") ;
           fprintf(stream, "		 deg = deg + 1 ;\n") ;
           fprintf(stream, "     }\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "	 i = 0 ;\n") ;
           fprintf(stream, "	 while (i <= deg) {\n") ;
           fprintf(stream, "		 tmp2[i+term] = gf_add((unsigned char)tmp1[i+term], (unsigned char)tmp2[i+term]) ;\n") ;
           fprintf(stream, "	     i = i + 1 ;\n") ;
           fprintf(stream, "	 }\n") ;
           fprintf(stream, "     tmp1[term] = 0 ;\n") ;
           fprintf(stream, "	 term = term  + 1 ;\n") ;
           fprintf(stream, "	 deg = 0 ;\n") ;
           fprintf(stream, "	}\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "  for(i=0; i<= new_deg; i++) {\n") ;
           fprintf(stream, "    p3[i] = tmp2[i] ;\n") ;
           fprintf(stream, "  }\n") ;
           fprintf(stream, "*deg_p3 = new_deg ;\n") ;
           fprintf(stream, "}\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "/* Multiply polynomial by a Constant */\n") ;
           fprintf(stream, "void gf_poly_conmul( unsigned char K, unsigned char *p1, int deg1, unsigned char *p2, int deg2, int *deg )\n") ;
           fprintf(stream, "{\n") ;
           fprintf(stream, "unsigned char *t1  ;\n") ;
           fprintf(stream, "int i ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "  t1 = p1 ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "  for (i=0; i<=deg1; i++) {\n") ;
           fprintf(stream, "    poly_p1[i] = *t1++ ;\n") ;
           fprintf(stream, "  }\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "  for (i=0; i<=deg1; i++) {\n") ;
           fprintf(stream, "    poly_p1[i] =  gfmul_ab(K, poly_p1[i]) ;\n") ;
           fprintf(stream, "  }\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "  *deg = deg1 ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "  for(i=0; i<= 9; i++) {\n") ;
           fprintf(stream, "    *p2++ = poly_p1[i] ;\n") ;
           fprintf(stream, "  }\n") ;
           fprintf(stream, "\n") ; 
           fprintf(stream, "}\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "// Add two polynomials\n") ; 
           fprintf(stream, "\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "void gf_poly_add( unsigned char *p1, int deg1, unsigned char *p2, int deg2, unsigned char *p3, int *deg )\n") ;
           fprintf(stream, "{\n") ;
           fprintf(stream, "unsigned char *t1, *t2 ;\n") ;
           fprintf(stream, "int i, j;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "  t1 = p1 ;\n") ;
           fprintf(stream, "  t2 = p2 ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "  for (i=0; i<=deg1; i++) {\n") ;
           fprintf(stream, "    poly_p1[i] = *t1++ ;\n") ;
           fprintf(stream, "  }\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "  for (i=0; i<=deg2; i++) {\n") ;
           fprintf(stream, "    poly_p2[i] = *t2++ ;\n") ;
           fprintf(stream, "  }\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "  for (i=0; i<MAX_POLY_ORDER; i++) {\n") ;
           fprintf(stream, "    poly_p3[i] = 0 ;\n") ;
           fprintf(stream, "  }\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "  if (deg1 >= deg2) \n") ;
           fprintf(stream, "    j = deg1 ;\n") ;
           fprintf(stream, "  else\n") ;
           fprintf(stream, "    j = deg2 ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "  for (i=0; i<=j; i++) {\n") ;
           fprintf(stream, "    poly_p3[i] = poly_p1[i] ^ poly_p2[i] ;\n") ;
           fprintf(stream, "  }\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "  *deg = j ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "  for(i=0; i<j; i++) {\n") ;
           fprintf(stream, "    *p3++ = poly_p3[i] ;\n") ;
           fprintf(stream, "  }\n") ;
           fprintf(stream, "\n") ;  
           fprintf(stream, "}\n") ;
           fprintf(stream, "\n") ;

           fprintf(stream, "\n") ;
           fprintf(stream, "//\n") ;
           fprintf(stream, "// gf_add(a,b)\n") ;
           fprintf(stream, "//     Add a + b\n") ; 
           fprintf(stream, "//\n") ;
           fprintf(stream, "unsigned char gf_add(unsigned char a, unsigned char b)\n") ;
           fprintf(stream, "{\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "    return (a ^ b) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "}\n") ;

           fprintf(stream, "/* Find Inverse of a (1/a) */\n") ;
           fprintf(stream, "unsigned char gf_inverse( a )\n") ;
           fprintf(stream, "unsigned char a ;\n") ;
           fprintf(stream, "{\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "return(gf_invalpha_tbl[a]) ;\n") ;
           fprintf(stream, "\n") ;
           fprintf(stream, "}\n") ;

}

//
// integer to binary length
//
int vector_len( int n)
{
	int len ;

    len = 0 ;
	while (n/2 != 0) {
		n = n/2 ;
		len += 1 ;
	}
	return (len) ;
}
//
// Convert Integer to binary vector string
//
void conv2_bvec_str(int n, int vlen, char *s)
{
	int i ;


    for (i=0; i<= vlen; i++ )
		s[i] = '0' ;

	i = vlen;

	while (n/2 != 0) {
		if (n%2)
			s[i] = '1' ;
		else
			s[i] = '0' ;
		n = n/2 ;
	    i -= 1 ;
	}
	if (n%2)
		s[i] = '1' ;
	else
		s[i] = '0' ;
	
}
//
// Print vector string
//
void pr_vec_str(FILE *stream, int vlen, char *s) 
{
	int i ;

	for(i=0; i <= vlen; i++) 
	    fprintf(stream, "%c", s[i]) ;
}
//
// Convert binary string to integer
//
int bstr_to_int(char *s)
{
    int i, j, n, str_len ;

    str_len = strlen(s) ;
    n = 0 ;
	j = 1 ;
	for(i=str_len-1; i>=0;i--) {
	  if (s[i] == '1' || s[i] == '0') {       
	      if (s[i] == '1') n = n | j ;
	  	  j = j << 1 ;
	  } else {
		  printf("Error @ index %d - Polynomial value is not binary : %c\n", i, s[i]) ;
		  return(n) ;
	  }
	}
 	return(n) ;
}
//
// Reverse string
//
void rev_str(char *s)
{
	int i, str_len ;
    char r[32] ;

    str_len = strlen(s) ;
	for (i=0; i<32; i++) r[i] = 0 ;
	for (i=0; i<str_len; i++) r[i] = s[str_len-i-1] ;
	for (i=0; i<str_len; i++) s[i] = r[i] ;
}



